###############################################################################
#    Plotting sequence-structure logo for SSMART                              #
#    Sequence-structure motif identification for RNA-binding proteins         #
#    25 Jan 2017                                                              #
#    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           #
#                                                                             #
#    SSMART is free software: you can redistribute it and/or modify           #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.    #
#                                                                             #
#    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   #
#                                                                             #
###############################################################################

# redirect normal output to file
sink("R.verbose", append=TRUE, split=FALSE)

if(!suppressMessages(require(grid))){
    install.packages("grid")
    library(grid)
}

args = commandArgs()

calculateic <- function(pwm, background) {
  if( length(background) != nrow(pwm) ) {
    stop("background length is not equal to pwm row length")
  }
  ic <- 0
  for(i in 1:ncol(pwm)) {
    tx <- 0
    for(j in 1:nrow(pwm)) {
      tx <- tx + pwm[j,i]*log2(pwm[j,i] / background[j])
    }
    ic[i] <- tx
  }
  ic
}

## get information content profile from PWM
pwm2ic<-function(pwm) {
    npos<-ncol(pwm)
    ic<-numeric(length=npos)
    for (i in 1:npos) {
        ic[i]<-2 + sum(sapply(pwm[, i], function(x) { 
            if (x > 0) { x*log2(x) } else { 0 }
        }))
    }    
    ic
}

# colapse 8 row pwm to 4 row pwm
colapse_pwm <- function(x, new_rows=4)
{
	npos <- ncol(x)
	for (i in 1:new_rows)
		for (j in 1:npos)
			x[i,j] = x[i,j]+x[i+new_rows,j]
	x <- x[1:4,]
	return(x)
}

# compute and return the unpaired probabilities =sum of the first 4 rows
get_struct_profile <- function(x, unp_rows=4)
{
	npos <- ncol(x)
	unp_prob <- array(0, dim=c(1,npos))
	for (i in 1:unp_rows)
		for (j in 1:npos)
			unp_prob[1,j] = unp_prob[1,j] + x[i,j]
	return(unp_prob)
}

# create and plot pwm
logoB <- function (pwm, str_profile, ic.scale = TRUE, xaxis = TRUE, yaxis = TRUE, xfontsize = 15, yfontsize = 15, plus_one=0) 
{
	# check pwm
	if (any(abs(1 - apply(pwm, 2, sum)) > 0.01)) 
		stop("Columns of PWM must add up to 1.0")

	# if pwm has 8 columns, colapse it to 4 and create a probability profile for structure
	if (str_profile) 
	{
		unpaired_prob = get_struct_profile(pwm)
		pwm = colapse_pwm(pwm)
	}

	# generate the letters of the logo
    chars <- c("A", "C", "G", "U")
    letters <- list(x = NULL, y = NULL, id = NULL, fill = NULL)
    npos <- ncol(pwm)
    if (ic.scale) {
        ylim <- 2.0
        ylab <- "Information content"
        facs <- pwm2ic(pwm)
    }
    else {
        ylim <- 1
        ylab <- "Probability"
        facs <- rep(1, npos)
    }
    wt <- 1
    x.pos <- 0
    for (j in 1:npos) {
        column <- pwm[, j]
        hts <- 0.95 * column * facs[j]
        letterOrder <- order(hts)
        y.pos <- 0
        for (i in 1:4) {
            letter <- chars[letterOrder[i]]
            ht <- hts[letterOrder[i]]
            if (ht > 0) 
                letters <- addLetter(letters, letter, x.pos, 
                  y.pos, ht, wt)
            y.pos <- y.pos + ht + 0.01
        }
        x.pos <- x.pos + wt
    }
    
    if (str_profile)
    {
		# generate the rectangles of the profile
		bars <- list(x = NULL, y = NULL, ht = NULL, wt = NULL)
		bar_width <- 1/npos
		x.pos <- bar_width/2
		for (j in 1:npos) 
		{
			up <- unpaired_prob[1,j]
			pp <- 1- up
			bar_height <- up/2
			y.pos <- 0.5 + bar_height/2
			bars <- addBar(bars, x.pos,y.pos,bar_height,bar_width)
			bar_height <- pp/2
			y.pos <- 0.5 - bar_height/2		
			bars <- addBar(bars, x.pos,y.pos,bar_height,bar_width)
			x.pos <- x.pos + bar_width
		}
	    #print(unpaired_prob)
	    #print(bars)
    }

    
    # PLOT
    grid.newpage()
	bottomMargin = ifelse(xaxis, 2 + xfontsize/3.5, 2)
	leftMargin = ifelse(yaxis, 2 + yfontsize/3.5, 2)
    if (str_profile)
    {
		vp <- viewport(width = 1, height=1)
		plotViewport(c(bottomMargin, leftMargin, 2, 2))
		pushViewport(vp)
		# TOP image - logo
		vp1 <- viewport(x=0.57, y=0.75, width=0.8, height=0.45, xscale=c(0,npos), yscale=c(0,2.0))
		pushViewport(vp1)
		grid.polygon(x = unit(letters$x, "native"), y = unit(letters$y,"native"), id = letters$id, gp = gpar(fill = letters$fill,col = "transparent"))
		if (yaxis) {
			grid.yaxis(gp = gpar(fontsize = yfontsize))
			grid.text(ylab, x = unit(-3, "lines"), rot = 90, gp = gpar(fontsize = yfontsize))
		}	
		# BOTTOM image - profile
		upViewport()
		vp2 <- viewport(x=0.57, y=0.35, width=0.8, height=0.25, xscale=c(0,npos), yscale=c(-1,1))
		pushViewport(vp2)
		#grid.rect(gp=gpar(col="blue"))
		for (j in 1:(npos*2))
			grid.rect(x=bars$x[j], y=bars$y[j], width=bars$wt[j], height=bars$ht[j], gp=gpar(col="black", fill ="grey"))
		ylab <- "<- P   U ->"
	}
	else
	{
		pushViewport(plotViewport(c(bottomMargin, leftMargin, 2, 2)))
		pushViewport(dataViewport(0:ncol(pwm), seq(0,2.0,by=0.5), name = "vp1"))
		grid.polygon(x = unit(letters$x, "native"), y = unit(letters$y,"native"), id = letters$id, gp = gpar(fill = letters$fill,col = "transparent"))
	}
	if (xaxis) {
	  labels <- 0
	  ats <- 0
	  if(plus_one > 0) {
		total_tics <- floor(ncol(pwm) / 10)
		labels[1] <- "+1"
		labels[2] <- ""
		labels[3] <- ""
		ats[1] <- plus_one
		ats[2] <- 1
		ats[3] <- ncol(pwm)
		i <- plus_one + 10
		j <- 4
		while( i < ncol(pwm)-1 ) {
		  labels[j] <- paste("+",abs(i - plus_one),sep="")
		  ats[j] <- i - 1
		  i <- i + 10
		  j <- j + 1
		}
		i <- plus_one - 10
		while( i > 1 ) {
		  labels[j] <- paste("-",abs(i - plus_one),sep="")
		  ats[j] <- i
		  j <- j + 1
		  i <- i - 10
		}
		ats <- ats - 0.5
	  }
	  else {
		labels <- 1:ncol(pwm)
		ats <- seq(0.5,ncol(pwm)-0.5)
	  }
		grid.xaxis(at = ats, label = labels, gp = gpar(fontsize = xfontsize))
		grid.text("Position", y = unit(-3, "lines"), gp = gpar(fontsize = xfontsize))
	}
	if (yaxis) {
		grid.yaxis(gp = gpar(fontsize = yfontsize))
		grid.text(ylab, x = unit(-3, "lines"), rot = 90, gp = gpar(fontsize = yfontsize))
	}
	popViewport()
	popViewport()
    par(ask = FALSE)
}


# add a new bar object to the list of bars
addBar <- function(bars,x.pos,y.pos,ht,wt)
{
  bars$x <- c(bars$x,x.pos)
  bars$y <- c(bars$y,y.pos)
  bars$ht <- c(bars$ht,ht)
  bars$wt <- c(bars$wt,wt)
  bars
} 

## A
 letterA <- function(x.pos,y.pos,ht,wt,id=NULL){
   
   x <- c(0,4,6,10,8,6.8,3.2,2,0,3.6,5,6.4,3.6)
  y <- c(0,10,10,0,0,3,3,0,0,4,7.5,4,4)
  x <- 0.1*x
  y <- 0.1*y

  x <- x.pos + wt*x
  y <- y.pos + ht*y

  if (is.null(id)){
	id <- c(rep(1,9),rep(2,4))
  }else{
	id <- c(rep(id,9),rep(id+1,4))
  }
	
  fill <- c("green","white")
	
  list(x=x,y=y,id=id,fill=fill)
}

## T
letterT <- function(x.pos,y.pos,ht,wt,id=NULL){
  
  x <- c(0,10,10,6,6,4,4,0)
  y <- c(10,10,9,9,0,0,9,9)
 x <- 0.1*x
  y <- 0.1*y

  x <- x.pos + wt*x
  y <- y.pos + ht*y

  if (is.null(id)){
	id <- rep(1,8)
  }else{
	id <- rep(id,8)
  }
  
  fill <- "red"
	
  list(x=x,y=y,id=id,fill=fill)
}

## U
letterU <- function(x.pos,y.pos,ht,wt,id=NULL){
	# draw the lower half elipse
      angle <- seq(pi/2,pi,length=100)
      x.l <- 0.5 + 0.5*sin(angle)
      y.l <- 0.5 + 0.5*cos(angle)
      
      x <- c(1-x.l,rev(x.l))
      y <- c(y.l,rev(y.l))
      
      x.i <- 0.5 +0.35*sin(angle)
      y.i <- 0.5 +0.35*cos(angle)
      x.i <- x.i[y.i<=max(y.l)]
      y.i <- y.i[y.i<=max(y.l)]
      y.i[1] <- max(y.l)
           
      x1 <- c(1-x.i,rev(x.i))
      y1 <- c(y.i,rev(y.i))
     
      x <- c(x,rev(x1))
      y <- c(y,rev(y1))
      
	# draw the right and left rectangular part    
     h <- 0.5	# starting height
     w <- 0.15 #width
     r1 <- max(x.l)   # right margin
     x.right <- c(r1-w,r1-w,r1,r1)
     y.right <- c(h,1,1,h)
	 x.left <- c(0,0,w,w)
     y.left <- c(h,1,1,h)
     
   
     if (is.null(id)){
       id <- c(rep(1,length(x)),rep(2,length(x.left)),rep(3,length(x.right)))
     }else{
       id <- c(rep(id,length(x)),rep(id+1,length(x.left)),rep(id+3,length(x.right)))
     }
   
    x <- c(rev(x),x.left,x.right)
     y <- c(rev(y),y.left,y.right)
       
     x <- x.pos + wt*x
     y <- y.pos + ht*y
     
     fill <- c("red","red","red")

        
      list(x=x,y=y,id=id,fill=fill)
    }

## C
letterC <- function(x.pos,y.pos,ht,wt,id=NULL){
  angle1 <- seq(0.3+pi/2,pi,length=100)
  angle2 <- seq(pi,1.5*pi,length=100)
  x.l1 <- 0.5 + 0.5*sin(angle1)
  y.l1 <- 0.5 + 0.5*cos(angle1)
  x.l2 <- 0.5 + 0.5*sin(angle2)
  y.l2 <- 0.5 + 0.5*cos(angle2)
  
  x.l <- c(x.l1,x.l2)
  y.l <- c(y.l1,y.l2)
  
  x <- c(x.l,rev(x.l))
  y <- c(y.l,1-rev(y.l))
  
  x.i1 <- 0.5 +0.35*sin(angle1)
  y.i1 <- 0.5 +0.35*cos(angle1)
  x.i1 <- x.i1[y.i1<=max(y.l1)]
  y.i1 <- y.i1[y.i1<=max(y.l1)]
  y.i1[1] <- max(y.l1)
  
  x.i2 <- 0.5 +0.35*sin(angle2)
  y.i2 <- 0.5 +0.35*cos(angle2)
  
  x.i <- c(x.i1,x.i2)
  y.i <- c(y.i1,y.i2)
  
  x1 <- c(x.i,rev(x.i))
  y1 <- c(y.i,1-rev(y.i))
  
  x <- c(x,rev(x1))
  y <- c(y,rev(y1))
  
  x <- x.pos + wt*x
  y <- y.pos + ht*y

  if (is.null(id)){
	id <- rep(1,length(x))
  }else{
	id <- rep(id,length(x))
  }
  
  fill <- "blue"
	
  list(x=x,y=y,id=id,fill=fill)
}


## G
letterG <- function(x.pos,y.pos,ht,wt,id=NULL){
 angle1 <- seq(0.3+pi/2,pi,length=100)
 angle2 <- seq(pi,1.5*pi,length=100)
 x.l1 <- 0.5 + 0.5*sin(angle1)
 y.l1 <- 0.5 + 0.5*cos(angle1)
 x.l2 <- 0.5 + 0.5*sin(angle2)
 y.l2 <- 0.5 + 0.5*cos(angle2)
 
 x.l <- c(x.l1,x.l2)
 y.l <- c(y.l1,y.l2)
 
 x <- c(x.l,rev(x.l))
 y <- c(y.l,1-rev(y.l))
 
 x.i1 <- 0.5 +0.35*sin(angle1)
 y.i1 <- 0.5 +0.35*cos(angle1)
 x.i1 <- x.i1[y.i1<=max(y.l1)]
 y.i1 <- y.i1[y.i1<=max(y.l1)]
 y.i1[1] <- max(y.l1)
 
 x.i2 <- 0.5 +0.35*sin(angle2)
 y.i2 <- 0.5 +0.35*cos(angle2)
 
 x.i <- c(x.i1,x.i2)
 y.i <- c(y.i1,y.i2)
 
 x1 <- c(x.i,rev(x.i))
 y1 <- c(y.i,1-rev(y.i))
 
 x <- c(x,rev(x1))
 y <- c(y,rev(y1))
 
 h1 <- max(y.l1)
 r1 <- max(x.l1)
 
 h1 <- 0.4
 x.add <- c(r1,0.5,0.5,r1-0.2,r1-0.2,r1,r1)
 y.add <- c(h1,h1,h1-0.1,h1-0.1,0,0,h1)

 

 if (is.null(id)){
   id <- c(rep(1,length(x)),rep(2,length(x.add)))
 }else{
   id <- c(rep(id,length(x)),rep(id+1,length(x.add)))
 }

x <- c(rev(x),x.add)
 y <- c(rev(y),y.add)
   
 x <- x.pos + wt*x
 y <- y.pos + ht*y
 
 
 fill <- c("orange","orange")
   
 list(x=x,y=y,id=id,fill=fill)

}

# add a new letter object to the list of letters
addLetter <- function(letters,which,x.pos,y.pos,ht,wt){
  if (which == "A"){
    letter <- letterA(x.pos,y.pos,ht,wt)
  }else if (which == "C"){
    letter <- letterC(x.pos,y.pos,ht,wt)    
  }else if (which == "G"){
    letter <- letterG(x.pos,y.pos,ht,wt)
  }else if (which == "T"){
    letter <- letterT(x.pos,y.pos,ht,wt)
  }else if (which == "U"){
    letter <- letterU(x.pos,y.pos,ht,wt)  
  }else{
    stop("which must be one of A,C,G,T,U")
  }
 
  letters$x <- c(letters$x,letter$x)
  letters$y <- c(letters$y,letter$y)
  lastID <- ifelse(is.null(letters$id),0,max(letters$id))
  letters$id <- c(letters$id,lastID+letter$id)
  letters$fill <- c(letters$fill,letter$fill)
  letters
} 


# read data
data = read.table(args[5], sep=",", header=FALSE)
x    = as.matrix(data)

# if pwm has 8 columns, colapse it to 4 and create a probability profile for structure
str_profile = FALSE
ew = 10
pw = 14
eh = ew/2
ph = pw/2
if (nrow(x) == 8) 
{
	str_profile = TRUE
	eh = 7
	ph = 10
}

# generate output file
outf <- substr(args[5],1,nchar(args[5])-4)
if (args[6]=="eps") {
	postscript(file=paste(outf, "_logo.eps", sep=""), paper="special", width = ew, height = eh, horizontal=FALSE)
}else if (args[6]=="png") {
	png(paste(outf, "_logo.png", sep=""), width = pw, height = ph, units = "cm", res=160, pointsize = 6)
}else if (args[6]=="pdf") {
	pdf(paste(outf, "_logo.pdf", sep=""), width=pw/2, height=ph/2, pagecentre=TRUE)
}else{
	stop("Last argument must be one of eps,png,pdf")
}
	
# plot logo	and structure profile
logoB(x, str_profile)

dev.off()
