###############################################################################
#    Plotting similarity between evolved k-mers for SSMART                    #
#    Sequence-structure motif identification for RNA-binding proteins         #
#    20 Jul 2017                                                              #
#    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           #
#                                                                             #
#    SSMART is free software: you can redistribute it and/or modify           #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.    #
#                                                                             #
#    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   #
#                                                                             #
###############################################################################

# redirect normal (and err) output to file
sink("R.verbose", append=TRUE, split=FALSE)

if(!suppressMessages(require(qgraph))){
    install.packages("qgraph")
    library(qgraph)
}
if(!suppressMessages(require(gplots))){
    install.packages("gplots")
    library(gplots)
}

if(!suppressMessages(require(NMF))){
    install.packages("NMF")
    library(NMF)
}

args = commandArgs()
# args[5] = input file
# args[6] = output type {eps, png, pdf}
# args[7] = plot title
# args[8] = number of motifs

max_out <- 10 #maximum number of motifs ploted with different colors

# Reading in data and transform it into matrix format
data = read.table(args[5], header = TRUE, quote = "\"'", dec = ".")
rnames <- data[,1]                            # assign labels in column 1 to "rnames"
sscores <- data[,2]							  # scores are in column 2
sclusters <- data[,3]						  # clusters are in column 3
mat_data <- data.matrix(data[,4:ncol(data)])  # transform column 4- into a matrix
rownames(mat_data) <- rnames                  # assign row names	
mmax <- max(sscores) 
mmin <- min(sscores)			  # value for converting scores to an interval
#  scale a range [min,max] to [a,b]
# (b-a)(x - min)/(max - min) +a
scaled_scores <- sscores
for (i in 1:length(sscores))
{
	scaled_scores[i] <- 4*(sscores[i]-mmin)/(mmax-mmin) + 1 # scale to [1,5]
}
#convert the clusters to the accepted format for groups
sgroups <- list() # a list of lists with indices for each motif cluster
motifs <- max(sclusters) # number of motifs
# group only top x motifs
out_motifs <- min(as.numeric(args[8]),motifs)
if (out_motifs > max_out)
	out_motifs <- max_out
assigned_motifs <- 0
for (mi in 1:out_motifs)
{
	mname <- paste("Motif",mi,sep="")
	# count the cluster members and create list
	mn <- sum(sclusters == mi)
	assigned_motifs <- assigned_motifs + mn
	sgroups[[mname]] <- rep(-1, mn)		
	j=1
	# put indices to list
	for (i in 1:length(sclusters))
		if (sclusters[i]==mi)
		{
			sgroups[[mname]][j] <- i
			j <- j+1
		}
}
# assigne all the rest to a single group
if (assigned_motifs < length(sclusters))
{
	# count the cluster members and create list
	sgroups[["Others"]] <- rep(-1, length(sclusters)-assigned_motifs)		
	j=1
	# put indices to list
	for (i in 1:length(sclusters))
		if (sclusters[i]> out_motifs)
		{
			sgroups[["Others"]][j] <- i
			sclusters[i] <- out_motifs+1 #change the sclusters for heatmap annotation
			j <- j+1
		}
	out_motifs <- out_motifs+1
}
#print(sgroups)

# colors
#colorblind_colors = c("#E69F00", "#56B4E9", "#009E73", "#F0E442", "#0072B2", "#D55E00", "#CC79A7")
gray_colors = c("#aaaaaa", "#888888")
colorblind_colors = c("#ba4c40",
						"#6872d7",
						"#62a351",
						"#a8803f",
						"#ba4d7e",
						"#45c097",
						"#844ba0",
						"#ca812e",
						"#788ed6",
						"#a5af3f") # 10 colors from http://tools.medialab.sciences-po.fr/iwanthue/
extra_color = "#e9e1ad" # ecc577
network_colors = c(colorblind_colors[1:out_motifs-1],extra_color)


# save plot 
if (args[6]=="eps") {
	plot_height = 8
	postscript(file=paste(args[5], ".eps", sep=""), paper="special", width = 1.7*plot_height, height = plot_height, horizontal=FALSE)
}else if (args[6]=="png") {
	plot_height = 14
	png(paste(args[5], ".png", sep=""), width = 1.7*plot_height, height = plot_height, units = "cm", res=300, pointsize = 8)
}else if (args[6]=="pdf") {
	plot_height = 8
	pdf(paste(args[5], ".pdf", sep=""), width = 1.7*plot_height, height = plot_height, pagecentre=TRUE)
}else{
	stop("Last argument must be one of eps,png,pdf")
}

# layout
par(mar=c(0,0,0,0),xpd = NA)
layout(matrix(c(1,1,2,3,4,5),ncol=2,byrow = TRUE),heights=c(0.5,8,1), widths=c(5,4))


plot.new()
text(0.5,0.5,args[7],cex=2,font=2)

# plot heatmap
motif_ann = data.frame(Motifs = factor(sclusters, labels = names(sgroups)))
aheatmap(mat_data, annCol = motif_ann, annRow = motif_ann, annColors=list(Motifs=network_colors))

# plot network
Q <- qgraph(mat_data, posCol = gray_colors, color=network_colors, minimum = 0.75, cut = 0.95, vsize = 8, labels=rnames, border.width=scaled_scores, groups=sgroups, layout = "spring", legend = FALSE)

# description
plot.new()
text(0.5,0.9,"Heatmap of the similarity scores between all unique evolved k-mers.", font=2, cex=1.2)
text(0.5,0.7,paste("The rows and columns represent k-mers and are color-coded to match the top",out_motifs-1,"motifs and all the rest."), cex=1.2)
plot.new()
text(0.5,0.9,"Network graph of all unique evolved k-mers.", font=2, cex=1.2)
text(0.5,0.5,"The nodes represent the k-mers (the contour thickness correlates with its SSMART score).\nThe edges represent the similarity score between k-mers:\nthin light gray lines correspond to a score > 0.7, while thick dark grey lines to a score > 0.9", cex=1.2)

# close plot
dev.off()
