#!/usr/bin/env python

###############################################################################
#    Predicting secondary structure for SSMART                                #
#    Sequence-structure motif identification for RNA-binding proteins         #
#    using RNAprofiling (E. Rogers, C. E. Heitsch, NAR 2014)                  #
#    20 Feb 2017                                                              #
#    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           #
#                                                                             #
#    This is free software: you can redistribute it and/or modify             #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.    #
#                                                                             #
#    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   #
#                                                                             #
###############################################################################

import os, random, time, multiprocessing, shutil
from argparse import ArgumentParser, FileType
from subprocess import Popen

param_folder = 'data' # data/GTparams/

time_limit = 300 # time in seconds = 5min (for each sequence)

def get_results(input_file, rnaprofiling_path, output_path_prefix, random_key, freq_cutoff, ncpus, ensemble):    
    # transform all paths into absolute paths
    input_file = os.path.abspath(input_file)
    output_path_prefix = os.path.abspath(output_path_prefix)
    rnaprofiling_path = os.path.abspath(rnaprofiling_path)    
    # create the result_path if it do not exists
    output_path = output_path_prefix[:output_path_prefix.rfind('/')]
    if not os.path.isdir(output_path):
        os.makedirs(output_path)
    # create the samples_path if it do not exists
    samples_path = os.path.join(output_path,'samples')
    if not os.path.isdir(samples_path):
        os.makedirs(samples_path)
    # copy param folder to running path
    if not os.path.isdir(os.path.join(samples_path,param_folder)):
        shutil.copytree(os.path.join(rnaprofiling_path,param_folder), os.path.join(samples_path,param_folder))
    # read fasta file
    print 'Reading input data ...'
    f=open(input_file,'r')
    fasta = SimpleFastaParser(f)
    f.close()
    # get the random key if needed and not provided
    if ensemble and not random_key:
        random_key = get_random_key(samples_path, fasta[0])
    if not random_key:
        ensemble = False # .sample files not find
        # make it short - only 3 characters - because the fileneme is restricted to 15 characters
        random_key = ''.join(random.choice('0123456789abcdefghijklmnoprstuvxzwABCDEFGHIJKLMNOPQRSTUVWXYZ') for i in range(3))
    # some data
    nseq = len(fasta)
    skipped = 0
    file_ext = input_file.rfind('.fa')
    if file_ext==-1:
        file_prefix = input_file[input_file.rfind('/')+1:]
    else:
        file_prefix = input_file[input_file.rfind('/')+1:file_ext]
    # for each sequence in fasta, write a txt file and run RNAprofiling, then read results    
    print 'Running RNAprofile on '+str(nseq)+' sequences from the fasta file '+file_prefix+' with random prefix '+random_key+'...'
    # getting number of CPUs usable  
    if ncpus == 0:
        ncpus = multiprocessing.cpu_count()
    else:
        ncpus = min(ncpus, multiprocessing.cpu_count())
    # write partial results to files
    partial_seqs = 100
    if nseq>10000:
        partial_seqs = 1000
    # create summary result files
    new_fasta = ''
    structures = ''
    profiles = 'Start_main\tStop_main\tHelices\tFrequencies\n'
    # write new fasta
    f=open(output_path_prefix+'.fasta','w')
    f.write(new_fasta) 
    f.close()
    # write structures 
    f=open(output_path_prefix+'.struct','w')
    f.write(structures) 
    f.close()
    # write profiles for all seqs
    f=open(output_path_prefix+'.profiles','w')
    f.write(profiles) 
    f.close()
    # fold blocks of 100 or 1000 sequences and then write the results to files
    for partial_run in range(nseq/partial_seqs+1):
        # multithreading
        pool = multiprocessing.Pool(ncpus)
        p_res = [] # list of results from parallel runs as tuple: (seq_idx, result)
        for seq_idx in range(partial_run*partial_seqs,min((partial_run+1)*partial_seqs, nseq)):            
            # get data
            sequence = fasta[seq_idx]
            end_pos = sequence[0].find('|')
            if end_pos==-1:
                file_name = random_key+'.'+sequence[0]
            else:
                file_name = random_key+'.'+sequence[0][:end_pos]
            if len(file_name)>15:
                file_name = file_name[:15]
            # run job
            res = pool.apply_async(func=job_profiling, args=(rnaprofiling_path, samples_path, file_name, sequence, freq_cutoff, ensemble))
            p_res.append((seq_idx,res)) # res = [paired_prob, profile_str, new_fasta_seq]        
        pool.close()
        pool.join()
        # create partial results
        new_fasta = ''
        structures = ''
        profiles = ''  
        for seq_idx, res in p_res:
            [paired_prob, profile_str, new_fasta_seq] = res.get()
            if paired_prob:
                new_fasta += new_fasta_seq            
                profiles += profile_str
                for prob in paired_prob:
                    structures += str(round(prob,3))+'\t'
                structures += '\n'
            else:
                skipped +=1           
        # write partial results
        # write new fasta
        f=open(output_path_prefix+'.fasta','a')
        f.write(new_fasta) 
        f.close()
        # write structures 
        f=open(output_path_prefix+'.struct','a')
        f.write(structures) 
        f.close()
        # write profiles for all seqs
        f=open(output_path_prefix+'.profiles','a')
        f.write(profiles) 
        f.close()
    print 'Finished '+str(seq_idx+1)+' / '+str(nseq)
    print 'Skipped '+str(skipped)+' / '+str(nseq)
    return True

def get_random_key(path, fasta_seq):
    seq_name= fasta_seq[0][:fasta_seq[0].find('|')]
    if len(seq_name)>11:
        seq_name= seq_name[:11]
    file_names = os.listdir(path)
    for file_name in file_names:
        if file_name.find(seq_name)>0:
            return file_name[:file_name.find('.')]
    return False

def get_paired_prob(sequence, helices, frequencies, total_structures):                        
    # get structure in the form of paired probabililies
    lseq = len(sequence)
    paired_prob = [0 for idx in range(lseq)]
    exception=False
    # traverse helices in reverse order: more frequent helices last
    for idx in range(len(frequencies)-1,-1,-1):
        for i in range(helices[idx][2]):
            try:               
                paired_prob[helices[idx][0]-1+i] = frequencies[idx]/float(total_structures)
                paired_prob[helices[idx][1]-1-i] = frequencies[idx]/float(total_structures)
            except:
                exception = True
    return [paired_prob, exception]

def job_profiling(rnaprofiling_path, output_path, file_name, sequence, freq_cutoff, ensemble):
    # runs RNAprofile on one sequence - if the time limit passes, it quits without results
    # returns [paired_prob, profile_str, new_fasta_seq]
    file_prefix = os.path.join(output_path,file_name)
    # write file 
    write_seq_file(output_path, file_name+'.txt', sequence[1])
    # verbose file for rnaprofiling output
    fo = open(file_prefix+'.verbose','w')
    # command
    my_env=os.environ.copy()
    cmd = os.path.join(rnaprofiling_path,'RNAprofile')+' -h '+str(freq_cutoff)+' -g -v -t 1 -o '+file_name
    if ensemble and os.path.isfile(file_prefix+'.samples'):
        cmd += ' -e '+file_name+'.samples'
    cmd += ' '+file_name+'.txt'
    # run command
    start = time.time()
    end = start + time_limit
    interval = .25
    p = Popen(cmd, shell=True, cwd=output_path, stdout=fo, env=my_env)
    not_terminated = True
    while not_terminated:
        if p.poll() is None: # p is running
            if time.time() >= end:
                # Terminate the process recursively
                p.kill()
                not_terminated = False                        
        else: # p finished successfully
            os.system('rm '+file_prefix+'.txt') # delete text file with sequence
            fo.close()
            # read helices
            [helices,frequencies, total_structures] = read_feat_helices('',file_prefix+'.verbose')     
            # delete file
            os.remove(file_prefix+'.verbose')
            # process data           
            if len(helices)>0: # structure exist                
                # get start and stop index for uppercase - in 0-based representation
                start = min([sequence[1].find(base) for base in ['A','C','G','T'] if base in sequence[1]])
                stop = max([sequence[1].rfind(base) for base in ['A','C','G','T']])
                # get structure as paired probabilities
                [paired_prob, exception] = get_paired_prob(sequence[1], helices, frequencies, total_structures)
                if not exception:  
                    # construct new sequence + struct info
                    new_fasta_seq = '>'+sequence[0]+'\n'
                    profile_str = str(start+1)+'\t'+str(stop+1)+'\t'+str(helices)+'\t'+str(frequencies)+'\n'                    
                    for idx in range(start, stop+1): #range(lseq):
                        if paired_prob[idx]<0.5:
                            new_fasta_seq += sequence[1][idx].upper()
                        else:
                            new_fasta_seq += sequence[1][idx].lower()
                    new_fasta_seq += '\n' 
                    return [paired_prob, profile_str, new_fasta_seq]         
            return [False, False, False]        
        time.sleep(interval)
    fo.close() 
    return [False, False, False]


def write_seq_file(path, file_name, seq):
    f = open(os.path.join(path, file_name),'w')
    f.write(seq)
    f.close()
    
def read_feat_helices(path,file_name):
    # read info from a verbose file, with useful info on "Featured helix" rows:
#     ...
#     Threshold to find frequent helices: 50.0%
#     Number of structures processed: 1000
#     Featured helix 1: 40 95 7 with freq 999
#     Featured helix 2: 48 66 7 with freq 998
#     Featured helix 3: 22 124 7 with freq 985
#     Featured helix 4: 32 115 10 with freq 985
#     Featured helix 5: 76 88 3 with freq 985
#     Featured helix 6: 96 109 4 with freq 972
#     Featured helix 7: 46 69 3 with freq 902
#     Featured helix 8: 6 17 3 with freq 725
#     Coverage by featured helix classes: 0.854
#     Total number of featured helix classes: 8
#     ...
    helices = []
    frequencies = []
    no_structures = 0
    with open(os.path.join(path, file_name),'r') as f:
        for line in f:
            if line.startswith('Number of structures processed'):
                no_structures =  int(line.split(': ')[1])
            if line.startswith('Featured helix'):    
                h = line[line.find(':')+2:line.find('with')-1].split(' ')            
                helices.append([int(item) for item in h])
                frequencies.append(int(line[line.find('freq')+5:]))
    return [helices, frequencies, no_structures]


def SimpleFastaParser(handle):
    """Generator function to iterator over Fasta records (as string tuples).

    For each record a tuple of two strings is returned, the FASTA title
    line (without the leading '>' character), and the sequence (with any
    whitespace removed). The title line is not divided up into an
    identifier (the first word) and comment or description.

    >>> for values in SimpleFastaParser(open("Fasta/dups.fasta")):
    ...     print(values)
    ('alpha', 'ACGTA')
    ('beta', 'CGTC')
    ('gamma', 'CCGCC')
    ('alpha (again - this is a duplicate entry to test the indexing code)', 'ACGTA')
    ('delta', 'CGCGC')

    """
    #Skip any text before the first record (e.g. blank lines, comments)
    while True:
        line = handle.readline()
        if line == "":
            return  # Premature end of file, or just empty?
        if line[0] == ">":
            break

    result = []
    while True:
        if line[0] != ">":
            raise ValueError(
                "Records in Fasta files should start with '>' character")
        title = line[1:].rstrip()
        lines = []
        line = handle.readline()
        while True:
            if not line:
                break
            if line[0] == ">":
                break
            lines.append(line.rstrip())
            line = handle.readline()

        #Remove trailing whitespace, and any internal spaces
        #(and any embedded \r which are possible in mangled files
        #when not opened in universal read lines mode)
        result.append([title, "".join(lines).replace(" ", "").replace("\r", "")])

        if not line:
            return result # StopIteration

    assert False, "Should not reach this line"   
    
########### PARSER ##################
parser = ArgumentParser(description='Reads a fasta input file and runs RNAprofiling on each sequence separately. Sequences can have a "main" part in uppercase, and extension in lowercase. Outputs all helices info in .profiles and the predicted structure for the main part in 2 files: a .fasta file with ssmart encoding (uppercase and lowercase nucleotides - lowercase corresponds to helix), and a .struct file with paired probabilities.')
parser.add_argument('-i', '--input', action='store', dest="input", required=True, type=FileType('r'), help="Path and name for the input fasta file.")
parser.add_argument('-p', '--profiling', action='store', dest="rnaprofiling", required=True, help="Path for RNAprofiling.")
parser.add_argument('-c', '--cutoff', action='store', dest="cutoff", type=int, default=0, help="Cutoff for the frequencies of helices (percentage). Supported values: between 0 and 100. Default 0.")
parser.add_argument('-n', '--ncpus', action='store', dest="ncpus", type=int, default=0, help="Number of workers (CPUs) to run in parallel, if omitted it will be set to the number of processors in the system.")
parser.add_argument('-o', '--output', action='store', dest="out", required=True, help="Path and prefix for the output files (prefix.fasta, prefix.profile and prefix.struct files will be written in path). If the directories specified do not exist, they will be created.")
parser.add_argument('-r', '--randomkey', action='store', dest="random_key", default=False, help="A string representing the 3 character random key for the run. To be used only when a run was killed and the completed profiling are only read.")
parser.add_argument('-e', '--ensemble', action='store_true', dest="ensemble", default=False, help="If this option is used, gtboltzmann style structure files are also available as input, one for each sequence.")

args = parser.parse_args()
if args.cutoff>=0 and args.cutoff<=100:
    get_results(input_file=args.input.name, rnaprofiling_path=args.rnaprofiling ,output_path_prefix=args.out,random_key=args.random_key, freq_cutoff=args.cutoff, ncpus=args.ncpus, ensemble=args.ensemble)
else:
    print 'Error: argument -c/--cutoff has supported values between 0 and 100!'
