/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */



#include "IUPACalphabet.h"

const char *all_IUPAC_nucleotides = "ACGTWSMKRYBDHVNacgtwsmkrybdhvn\0";
const char *my_IUPAC_nucleotides = "ACGTWSMKRYNacgtwsmkryn\0"; // without 3 letters degeneracy
const char primary_IUPAC_nucleotides[8] = {'A', 'G', 'C', 'T', 'a', 'g', 'c', 't'};
const int no_primary_IUPAC_nucleotides = 8;
const char pssm_rows_TRANSFAC[8] = {'A','C','G','T','a','c','g','t'};
//const int no_pssm_rows_TRANSFAC = 8;

int find_index(char a[], int num_elements, char value)
//returns the index of a char in a char array - used to check if a nucleotide is in a sequence
{
   int i;
   for (i=0; i<num_elements; i++)
   {
	 if (a[i] == value)
	 {
	    return i;  /* it was found */
	 }
   }
   return -1;  /* if it was not found */
}

char *get_rev_compl(char *str)
{
	int i, len = strlen(str);
	int k=len-1;
	char *rev_str=malloc(sizeof(char) * len);

	for(i = 0; i < len; i++)
	{
		rev_str[i] = get_complement(str[k]);
		k--;
	}

	return rev_str;
}

/*IUPAC ambiguity codes
---------------------
M  	A or C
R  	A or G
W 	A or T
S 	C or G
Y 	C or T
K 	G or T
V 	A or C or G
H 	A or C or T
D 	A or G or T
B 	C or G or T
N 	A or C or G or T
*/
/*IUPAC Complementarity
---------------------
A	T
T	A
U	A
G	C
C	G
Y	R
R	Y
S	S*
W	W*
K	M
M	K
B	V
D	H
H	D
V	B
N	N
*/



char *get_encoding(char symbol)
// converts a symbol into 1,2,3,4... symbols that it represents
{
	switch(symbol){
		// 1 letter symbol (lower+upper)
		case 'A':
		case 'C':
		case 'G':
		case 'T':
		case 'a':
		case 'c':
		case 'g':
		case 't':
		{// return same symbol as char array
			char res[2];
			res[0] = symbol;
			res[1] = '\0';
			return strcpy(malloc(sizeof(char) * 2), res);
		}
		// 2 letter symbol (upper)
		case 'W':
			return "AT\0";
		case 'S':
			return "CG\0";
		case 'M':
			return "AC\0";
		case 'K':
			return "GT\0";
		case 'R':
			return "AG\0";
		case 'Y':
			return "CT\0";
		// 3 letter symbol (upper)
		case 'B':
			return "CGT\0";
		case 'D':
			return "AGT\0";
		case 'H':
			return "ACT\0";
		case 'V':
			return "ACG\0";
		// 4 letter symbol (upper)
		case 'N':
			return "ACGT\0";
		// 2,3,4 letter symbols for lowercase
		case 'w':
			return "at\0";
		case 's':
			return "cg\0";
		case 'm':
			return "ac\0";
		case 'k':
			return "gt\0";
		case 'r':
			return "ag\0";
		case 'y':
			return "ct\0";
		case 'b':
			return "cgt\0";
		case 'd':
			return "agt\0";
		case 'h':
			return "act\0";
		case 'v':
			return "acg\0";
		case 'n':
			return "acgt\0";
	}
	// if none of the above, return same symbol as char array
	char res[2];
	res[0] = symbol;
	res[1] = '\0';
	return strcpy(malloc(sizeof(char) * 2), res);
}

// returns one degenerate symbol corresponding to the specified pair
// used in pssm_to_IUPAC and so also (lower, upper) pairs of same letter are accepted
char get_degenerate_symbol(char symbol1, char symbol2)
{
	if(symbol1==symbol2)
		return symbol1;
	switch(symbol1)
	{
		case 'A':{
			switch(symbol2){
				case 'C': return 'M';
				case 'G': return 'R';
				case 'T': return 'W';
				case 'a': return 'A';
			}
			return 'N';
		}
		case 'C':{
			switch(symbol2){
				case 'A': return 'M';
				case 'G': return 'S';
				case 'T': return 'Y';
				case 'c': return 'C';
			}
			return 'N';
		}
		case 'G':{
			switch(symbol2){
				case 'A': return 'R';
				case 'C': return 'S';
				case 'T': return 'K';
				case 'g': return 'G';
			}
			return 'N';
		}
		case 'T':{
			switch(symbol2){
				case 'A': return 'W';
				case 'C': return 'Y';
				case 'G': return 'K';
				case 't': return 'T';
			}
			return 'N';
		}
		case 'a':{
			switch(symbol2){
				case 'c': return 'm';
				case 'g': return 'r';
				case 't': return 'w';
				case 'A': return 'a';
			}
			return 'n';
		}
		case 'c':{
			switch(symbol2){
				case 'a': return 'm';
				case 'g': return 's';
				case 't': return 'y';
				case 'C': return 'c';
			}
			return 'n';
		}
		case 'g':{
			switch(symbol2){
				case 'a': return 'r';
				case 'c': return 's';
				case 't': return 'k';
				case 'G': return 'g';
			}
			return 'n';
		}
		case 't':{
			switch(symbol2){
				case 'a': return 'w';
				case 'c': return 'y';
				case 'g': return 'k';
				case 'T': return 't';
			}
			return 'n';
		}
	}
}

short int get_degeneracy(char symbol)
// returns the number of letters coded by a symbol
// 1 for ACGTacgt, 2 for MRYK..., 3 for BDHV.., 4 for Nn
{
	switch(symbol){
		// 1 letter symbol (lower+upper)
		case 'A':
		case 'C':
		case 'G':
		case 'T':
		case 'a':
		case 'c':
		case 'g':
		case 't':
			return 1;
		// 2 letter symbol (lower+upper)
		case 'W':
		case 'S':
		case 'M':
		case 'K':
		case 'R':
		case 'Y':
		case 'w':
		case 's':
		case 'm':
		case 'k':
		case 'r':
		case 'y':
			return 2;
		// 3 letter symbol (lower+upper)
		case 'B':
		case 'D':
		case 'H':
		case 'V':
		case 'b':
		case 'd':
		case 'h':
		case 'v':
			return 3;
		// 4 letter symbol (lower+upper)
		case 'N':
		case 'n':
			return 4;
		default:
			return 0;
	}
}

char *get_symbols(short int degen)
// returns all symbols that have a specified degeneracy
{
	short int i, idx = 0, alphabet_len = (short int)strlen(all_IUPAC_nucleotides);
	char res[alphabet_len];
	if (degen>0 && degen <5)
	{
		for(i=0;i<alphabet_len;i++)
			if (get_degeneracy(all_IUPAC_nucleotides[i]) == degen)
			{
				res[idx] = all_IUPAC_nucleotides[i];
				idx++;
			}
	}
	res[idx] = '\0';
	return strcpy(malloc(sizeof(char) * alphabet_len), res);
}

char get_complement(char symbol)
// returns the complement of a symbol
{
	switch(symbol){
		// 1 letter symbol (upper)
		case 'A':
			return 'T';
		case 'C':
			return 'G';
		case 'G':
			return 'C';
		case 'T':
			return 'A';
		// 2 letter symbol (upper)
		case 'W':
			return 'W';
		case 'S':
			return 'S';
		case 'M':
			return 'K';
		case 'K':
			return 'M';
		case 'R':
			return 'Y';
		case 'Y':
			return 'R';
		// 3 letter symbol (upper)
		case 'B':
			return 'V';
		case 'D':
			return 'H';
		case 'H':
			return 'D';
		case 'V':
			return 'B';
		// 4 letter symbol (upper)
		case 'N':
			return 'N';
		// 1,2,3,4 letter symbols for lowercase
		case 'a':
			return 't';
		case 'c':
			return 'g';
		case 'g':
			return 'c';
		case 't':
			return 'a';
		case 'w':
			return 'w';
		case 's':
			return 's';
		case 'm':
			return 'k';
		case 'k':
			return 'm';
		case 'r':
			return 'y';
		case 'y':
			return 'r';
		case 'b':
			return 'v';
		case 'd':
			return 'h';
		case 'h':
			return 'd';
		case 'v':
			return 'b';
		case 'n':
			return 'n';
		default:
			return symbol;
	}
}

char get_struct_complement(char symbol)
// returns the structural complement of a symbol => uppercase if lowercase and viceversa
{
	if (isupper(symbol))
		return tolower(symbol);
	if (islower(symbol))
		return toupper(symbol);
	return symbol;
}

int get_intersection(char symbol1, char symbol2)
//returns the number of symbols in common
{
	if (strchr(my_IUPAC_nucleotides,symbol1) != NULL && strchr(my_IUPAC_nucleotides,symbol2) != NULL)
		{ // if both symbols are in my alphabet
			int i,j,intersection=0;
			int ds1 = get_degeneracy(symbol1);
			int ds2 = get_degeneracy(symbol2);
			char* s1 = get_encoding(symbol1);
			char* s2 = get_encoding(symbol2);
			for(i=0;i<ds1;i++)
				for(j=0;j<ds2;j++)
					if (s1[i]==s2[j])
						intersection++;
			return intersection;
		}
	return 0;
}

double get_relation_score(char symbol1, char symbol2)
// returns the relationship score for encoding of the two symbols:
// 1 if identic, 1/degeneracy if one is encoded by the other, 0 otherwise
{
	if (strchr(my_IUPAC_nucleotides,symbol1) != NULL && strchr(my_IUPAC_nucleotides,symbol2) != NULL)
	{ // if both symbols are in my alphabet
		int ds1 = get_degeneracy(symbol1);
		int ds2 = get_degeneracy(symbol2);
		if (symbol1 == symbol2) // same symbol
			return 1;
		if (ds1==1 && ds2>1) { // symbol1 included in symbol2
			if (strchr(get_encoding(symbol2), symbol1) != NULL)
				return 1.0/ds2;
		}
		if (ds1>1 && ds2==1) { //symbol2 included in symbol1
			if (strchr(get_encoding(symbol1), symbol2) != NULL)
				return 1.0/ds1;
		}
		// all other cases
		return 0;
	}
	return 0;
}

char *increase_degeneracy(char symbol)
// returns all possible new values for an increase in degeneracy (works for degen 1,2,3)
// adds first the same symbol in the other structural context
{
	char result[5];
	short int i, idx = 0, alphabet_len = (short int)strlen(all_IUPAC_nucleotides);
	char *encode;
	// adds first the same symbol in the other structural context
	result[idx] = get_struct_complement(symbol);
	idx++;
	// adds the actual sequence degeneracy
	switch(get_degeneracy(symbol)){
		// 1 letter symbol
		case 1:
			// find all symbols that encode 2 letters that contain "symbol"
			for(i=0;i<alphabet_len;i++)
				if (get_degeneracy(all_IUPAC_nucleotides[i]) == 2)
					if (strchr(get_encoding(all_IUPAC_nucleotides[i]),symbol) != NULL)
					{
						result[idx] = all_IUPAC_nucleotides[i];
						idx++;
					}
			break;
		// 2 letters symbol - behaviour for 3 letter symbol
		case 2:
/*
			// find all symbols that encode 3 letters that contain both nucleotides encoded by "symbol"
			for(i=0;i<alphabet_len;i++)
			{
				if (get_degeneracy(all_IUPAC_nucleotides[i]) == 3)
				{
					encode=get_encoding(symbol);
					if (strchr(get_encoding(all_IUPAC_nucleotides[i]),encode[0]) != NULL && strchr(get_encoding(all_IUPAC_nucleotides[i]),encode[1]) != NULL)
					{
						result[idx] = all_IUPAC_nucleotides[i];
						idx++;
					}
				}
			}
			break;
		// 3 letters symbol
		case 3:
*/
			if (symbol >= 65 && symbol <= 90) {// is upper
				result[idx] = 'N';
			} else {
				result[idx] = 'n';
			}
			idx++;
			break;
	}
	result[idx] = '\0';
	return strcpy(malloc(sizeof(char) * 5), result);
}


char *decrease_degeneracy(char symbol)
// returns all possible new values for an decrease in degeneracy (works for degen 2 and 4)
{
	char result[12];
	// adds first the same symbol in the other structural context
	result[0] = get_struct_complement(symbol);
	// adds the actual sequence degeneracy
	if(get_degeneracy(symbol)==2){
		// 2 letters symbol
		strcpy(result+1, get_encoding(symbol));
	}
	if (symbol=='N')
		strcpy(result+1, "ACGTWSMKRY\0");
	if (symbol=='n')
		strcpy(result+1, "acgtwsmkry\0");
	return strcpy(malloc(sizeof(char) * 12), result);
}

void test_alphabet()
{
	printf("Encoding of K: %s\n", get_encoding('K'));
	printf("Encoding of b: %s\n", get_encoding('b'));
	printf("Encoding of A: %s\n", get_encoding('A'));
	printf("Degeneracy of n: %d\n", get_degeneracy('n'));
	printf("Degeneracy of Y: %d\n", get_degeneracy('Y'));
	printf("Complement of A: %c\n", get_complement('A'));

	short int i, alphabet_len = (short int)strlen(all_IUPAC_nucleotides);
	for(i=0;i<alphabet_len;i++)
		printf("increase_degeneracy of %c: %s\n", all_IUPAC_nucleotides[i], increase_degeneracy(all_IUPAC_nucleotides[i]));
	printf("increase_degeneracy of x: %s\n", increase_degeneracy('x'));
	for(i=0;i<alphabet_len;i++)
		printf("decrease_degeneracy of %c: %s\n", all_IUPAC_nucleotides[i], decrease_degeneracy(all_IUPAC_nucleotides[i]));
	printf("decrease_degeneracy of x: %s\n", decrease_degeneracy('x'));

	for(i=0;i<6;i++)
		printf("Symbols with degeneracy %d: %s\n",i, get_symbols(i));

	printf("Intersection of A, A: %d\n", get_intersection('A','A'));
	printf("Intersection of a, k: %d\n", get_intersection('a','k'));
	printf("Intersection of A, W: %d\n", get_intersection('A','W'));
	printf("Intersection of A, D: %d\n", get_intersection('A','D'));
	printf("Intersection of A, N: %d\n", get_intersection('A','N'));
	printf("Intersection of m, r: %d\n", get_intersection('m','r'));
	printf("Intersection of M, N: %d\n", get_intersection('M','N'));
}
