/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#include "PSSM.h"
#include "similarity_metrics.h"
#include "IUPACalphabet.h"

extern const char *my_IUPAC_nucleotides;
extern const char primary_IUPAC_nucleotides[8];
extern const int no_primary_IUPAC_nucleotides;
extern char pssm_rows_TRANSFAC[8];

/*IUPAC ambiguity codes
---------------------
M  	A or C
R  	A or G
W 	A or T
S 	C or G
Y 	C or T
K 	G or T
V 	A or C or G
H 	A or C or T
D 	A or G or T
B 	C or G or T
N 	A or C or G or T
*/

// assume rows are in following order: A, G, C, T
gsl_matrix *read_mat(char *file)
{
//	printf("%s\n", file);
 	FILE *f = fopen(file, "r");
 	if(f == NULL)
	{
 		return 0;
	}

	int rows, cols;
	fscanf(f, "%i\t%i", &rows, &cols);
	gsl_matrix *mat = gsl_matrix_alloc(rows, cols);
 	int res = gsl_matrix_fscanf(f, mat);
  	fclose(f);
 	if(res == GSL_EFAILED)
  	{
 //		printf("failed reading pssm from file '%s'\n", file);
  		return 0;
  	}
  	return mat;
}

void write_mat(gsl_matrix *mat, char *file)
{
	FILE *f = fopen(file, "w");
	int i,j;
//	fprintf(f, "%i\t%i\n", mat->size1, mat->size2);
	for(i=0;i<mat->size1;i++)
	{
	   	for(j=0;j<mat->size2;j++)
	   	{
	   		if(j != mat->size2 - 1)
	   			fprintf(f, "%-15.4g", gsl_matrix_get(mat, i, j));
	   		else
	   			fprintf(f, "%-15.4g\n", gsl_matrix_get(mat, i, j));
	   	}
 	}
 	fclose(f);
}


gsl_matrix *colapse_pssm(gsl_matrix *pssm, char **struct_logo)
// convert 8 a line pssm to 4 line pssm
// return the structure logo as string in struct_logo
{
	// get new number of rows
	int i,j, new_size1;
	new_size1 = (pssm->size1)/2;
	// create a new matrix object
	gsl_matrix *new_pssm;
	new_pssm = gsl_matrix_alloc(new_size1, pssm->size2);

	// create a char * for struct_logo
	char *my_logo;
	MALLOC(my_logo, sizeof(char)*(pssm->size2+1))

	// parse the matrix on columns
	double upper, lower, upper_sum;
	for (j=0;j<new_pssm->size2;j++)
	{
		upper_sum = 0;
		for (i=0;i<new_pssm->size1;i++)
		{
//			printf("%i x %i, (new_size1=%i): %f + %f \n", i, j, new_size1, gsl_matrix_get(pssm, i, j), gsl_matrix_get(pssm, i+new_size1, j));
			upper = gsl_matrix_get(pssm, i, j);
			lower = gsl_matrix_get(pssm, i+new_size1, j);
			upper_sum += upper;
			gsl_matrix_set(new_pssm, i, j, upper + lower);
		}
		my_logo[j] = (upper_sum) > (0.5) ? 'U' : 'P';
	}
	my_logo[j] = '\0';
	*struct_logo = my_logo;
	return new_pssm;
}


void output_pssm_csv_format(gsl_matrix *pssm, char *file)
{
	int row, column;
	FILE *f = open_file(file, "w");
  	for(row=0;row<pssm->size1;row++)
	{
		for(column=0;column<pssm->size2;column++)
		{
			char letter = pssm_rows_TRANSFAC[row];
			if(column == pssm->size2 - 1)
				fprintf(f, "%lf\n", get_pos(pssm, column, letter) );
			else
				fprintf(f, "%lf,", get_pos(pssm, column, letter) );
		}
	}
	fclose(f);
}

int next_line(FILE *fp, char **line)
{
	*line = 0; int line_length = 0;
	int num_read = getdelim(line, &line_length, '\n', fp);
	if((*line)[(num_read - 1)*sizeof(char)] == '\n')
		(*line)[(num_read - 1)*sizeof(char)] = '\0';
	return num_read;
}

void discard_leading_lines(FILE *fp, int num_line_to_discard)
{
//	printf("discarding:\n");
	char *line;
	int i;
	for(i=0;i<num_line_to_discard;i++)
	{
		int read = next_line(fp, &line);
//		printf("%i: %s\n", read, line);
		FREE(line);
	}
//	printf("done\n");
}

int find_num_cols_TRASNFAC_pssm(char *pssm_TRANSFAC_file)
{
//	printf("find_num_cols_TRASNFAC_pssm\n");
	FILE *fp = open_file(pssm_TRANSFAC_file, "r");
 	if(fp == NULL)
 		return 0;

	char *line; ssize_t read;
	int column_count = 0;
	int dum1 = -1; char dum2 = 'I'; int dum1_prev = -2;
	bool start_flag = false;
	while((read = next_line(fp, &line)) != -1)
	{
//		printf("%i: %s\n", read, line);
//		printf("column_count = %i\n", column_count);
		double counts[4];
		if(column_count < 10)
			sscanf(line, "0%i\t%lf\t%lf\t%lf\t%lf\t%c", &dum1, &counts[0], &counts[1], &counts[2], &counts[3], &dum2);
		else
			sscanf(line, "%i\t%lf\t%lf\t%lf\t%lf\t%c", &dum1, &counts[0], &counts[1], &counts[2], &counts[3], &dum2);
//		printf("dum1 = %i\tcounts[3] = %lf\tdum2 = %c\n", dum1, counts[3], dum2);//getchar();

		if(dum1 != -1)
			start_flag = true;
		if(dum1 == dum1_prev && start_flag == true)
			break;
		dum1_prev = dum1;

		if(start_flag == true)
			column_count++;
		FREE(line);

	}
	FREE(line);
	fclose(fp);
//	printf("done find_num_cols_TRASNFAC_pssm\n");
	return column_count;
}

void read_pssm_TRANSFAC_like_format_counts(gsl_matrix *pssm_out, char *file, double pseudocount)
{
	FILE *f = open_file(file, "r");
	int i;

	char *line; ssize_t read;
	int column = 0;
	int dum1 = -1; char dum2 = 'I', dum2_prev = 'I'; int dum1_prev = -2;
	bool start_flag = false;
	while((read = next_line(f, &line)) != -1)
	{
//		printf("%i: %s\n", read, line);
		double counts[4];
		if(column < 10)
			sscanf(line, "0%i\t%lf\t%lf\t%lf\t%lf\t%c", &dum1,&counts[0],&counts[1],&counts[2],&counts[3], &dum2);
		else
			sscanf(line, "%i\t%lf\t%lf\t%lf\t%lf\t%c", &dum1,&counts[0],&counts[1],&counts[2],&counts[3], &dum2);
//		printf("dummy = %i\tdum2 = %c\n", dum1, dum2);//getchar();
		if(dum1 != -1)
			start_flag = true;

		if(dum1 == dum1_prev && start_flag == true)
			break;
		dum1_prev = dum1;
		dum2_prev = dum2;

		if(start_flag == true)
		{
//			printf("count = %i\n", count);
			double sum = 0;
			for(i=0;i<4;i++)
				sum += counts[i];
			sum += 4*pseudocount;
			for(i=0;i<4;i++)
			{
				char letter = pssm_rows_TRANSFAC[i];
				set_pos(pssm_out, column, letter, (counts[i] + pseudocount) / sum);
			}	
			column++;
		}
		FREE(line);
	}
	FREE(line);
  	fclose(f);
}


void read_pssm_TRANSFAC_like_format_freq(gsl_matrix *pssm_out, char *file)
{
	FILE *f = open_file(file, "r");
	int i;

	char *line; ssize_t read;
	int column = 0;
	int dum1 = -1; char dum2 = 'I'; int dum1_prev = -2;
	bool start_flag = false;
	while((read = next_line(f, &line)) != -1)
	{
//		printf("%i: %s\n", read, line);
		double freq[4];
		if(column < 10)
			sscanf(line, "0%i\t%lf\t%lf\t%lf\t%lf\t%c", &dum1,&freq[0],&freq[1],&freq[2],&freq[3],&dum2);
		else
			sscanf(line, "%i\t%lf\t%lf\t%lf\t%lf\t%c", &dum1,&freq[0],&freq[1],&freq[2],&freq[3],&dum2);
//		printf("dummy = %i\tdum2 = %c\n", dum1, dum2);getchar();
		if(dum1 != -1)
			start_flag = true;
		if(start_flag == true)
		{
			for(i=0;i<4;i++)
			{
				char letter = pssm_rows_TRANSFAC[i];
				set_pos(pssm_out, column, letter, freq[i]);
			}	
			column++;
		}
		FREE(line);
		if(dum1 == dum1_prev && dum1 != -1)
			break;
		dum1_prev = dum1;
	}
	FREE(line);
//	printf("done: num columns = %i\n", column);
// 	gsl_matrix_view mview_tmp = gsl_matrix_submatrix(tmp, 0, 0, 4, column);
// 	gsl_matrix_memcpy(pssm_out, &mview_tmp.matrix);
// 	gsl_matrix_free(tmp);
  	fclose(f);
//	print_pssm(pssm_out);
//	printf("done all\n"); getchar();
}

extern species species_type_global;


void write_pssm_TRANSFAC_like_format_freq(gsl_matrix *pssm, char *file, char *itype)
{
	char file_csv_out[500];
	sprintf(file_csv_out, "%s.txt", file);

/*/////////////////////////********************************
	// convert 8 line pssm to 4 line pssm for logo plot
	gsl_matrix *colapsed_pssm = colapse_pssm(pssm, &struct_logo);

	printf("Normal PSSM:\n");
	print_pssm(pssm);
	printf("Collapsed PSSM:\n");
	print_pssm(colapsed_pssm);
	printf("struct_logo: %s\n", struct_logo);

	output_pssm_csv_format(colapsed_pssm, file_csv_out);
//////////////////////////********************************/

	output_pssm_csv_format(pssm, file_csv_out);
//	printf("file_csv_out: %s\n", file_csv_out);

	char species[50];
	if(species_type_global == human)
		sprintf(species, "human");	
	else if(species_type_global == yeast)
		sprintf(species, "yeast");
	else if(species_type_global == arabidopsis)
		sprintf(species, "arabidopsis");
	else
		sprintf(species, "unspecified_species");


	// write R file script
	char filename_R_script[500];
	FILE *fp_test;
	if( (fp_test = fopen("./generate_logo.R", "r")) == NULL)
	{
		printf("PSSM logo generation file 'generate_logo.R' not found in current running directory!\n");
		return;
	}
	else
		fclose(fp_test);

	sprintf(filename_R_script, "./generate_logo.R");
	char tmp_file[500];

	while(file_exist(filename_R_script) == false)
	{
		strcpy(tmp_file, filename_R_script);
		sprintf(filename_R_script, "./.%s", tmp_file);
	}

	char command[500];
	sprintf(command, "R < %s --slave --vanilla --args %s %s", filename_R_script, file_csv_out, itype);
	//sprintf(command, "R < %s --slave --vanilla --args %s %s %s", filename_R_script, file_csv_out, species, struct_logo);
	//printf("!!!!!!!command: %s\n", command);

	system(command);
	unlink(file_csv_out);
	//FREE(struct_logo);


	char motif[pssm->size2 + 1];motif[pssm->size2] = '\0';
	pssm_to_IUPAC(pssm, motif);

	FILE *f = open_file(file, "w");
	int i,j;
//	char pssm_rows_TRANSFAC[4] = {'A','C','G','T'};
	char species_str[50];
	switch(species_type_global) 
	{
		case human: strcpy(species_str, "human"); break;
		case yeast: strcpy(species_str, "yeast"); break;
		case arabidopsis: strcpy(species_str, "arabidopsis"); break;
		default: strcpy(species_str, "unspecified"); break;
	}
   	fprintf(f, "# %s; Species: %s\n", motif, species_str);
   	fprintf(f, "PO\tA\tC\tG\tT\ta\tc\tg\tt\n");
	for(i=0;i<pssm->size2;i++)
	{
		if(i < 10)
			fprintf(f, "0%i", i+1);
		else
			fprintf(f, "%i", i+1);
	   	for(j=0;j<pssm->size1;j++)
		{
			int column = i;
			char letter = pssm_rows_TRANSFAC[j];
   			fprintf(f, "\t%-10.4lf", get_pos(pssm, column, letter) );
		}
//		fprintf(f, "N	\n");
		fprintf(f, "%c	\n", motif[i]);
 	}
 	fclose(f);
}




gsl_matrix *generate_PSSM(int length)
{
	gsl_matrix *pssm = gsl_matrix_alloc(no_primary_IUPAC_nucleotides, length);
	double *freq, denom;
	MALLOC(freq, sizeof(double)*no_primary_IUPAC_nucleotides);
	int i,j;
	double *ind_arr_double[8] = {0,1,2,3,4,5,6,7};
	int ind_arr_int[no_primary_IUPAC_nucleotides];
	gsl_permutation *per = gsl_permutation_alloc(no_primary_IUPAC_nucleotides);
	for(j=0;j<length;j++)
	{
		permute1(ind_arr_double, no_primary_IUPAC_nucleotides, per);
		for(i=0;i<no_primary_IUPAC_nucleotides;i++)
			ind_arr_int[i] = (int)ind_arr_double[i];
		for(i=0;i<no_primary_IUPAC_nucleotides;i++)
			ran_uniform(freq+ind_arr_int[i], 1, i, i*5);
		denom = sum(freq, no_primary_IUPAC_nucleotides);
		for(i=0;i<no_primary_IUPAC_nucleotides;i++)
			gsl_matrix_set(pssm, i, j, freq[i] / denom);
	}
	gsl_permutation_free(per);
	FREE(freq);
	return pssm;
}

// assume rows are in following order: A, G, C, T
char draw_col_entry_pssm(gsl_vector_view vv)
{
	double ran = ran_uniform_single_value();
	if(ran < gsl_vector_get(&vv.vector, 0))
		return 'A';
	else if(ran < gsl_vector_get(&vv.vector, 0)+gsl_vector_get(&vv.vector, 1))
		return 'G';
	else if(ran < gsl_vector_get(&vv.vector, 0)+gsl_vector_get(&vv.vector, 1)+gsl_vector_get(&vv.vector, 2))
		return 'C';
	else
		return 'T';
}

char *draw_from_pssm(gsl_matrix *pssm)
{
	int i;
  	char *draw;
  	MALLOC(draw, sizeof(char)*(pssm->size2+1));
  	draw[pssm->size2]='\0';
  	for (i = 0; i < pssm->size2; i++)
  	{
  		draw[i] = draw_col_entry_pssm(gsl_matrix_column(pssm, i));
  	}
  	return draw;
}


void set_pos(gsl_matrix *pssm, int column, char letter, double value)
{
	if(value > 1 || value < 0)
	{
	//	printf("ERROR! file PSSM.c | function 'set_pos' | message: value outside of range [0,1]: %g\n", value);
	}
	// row order: A, G, C, T, a, g, c, t
	int idx;
	idx = find_index(primary_IUPAC_nucleotides, no_primary_IUPAC_nucleotides, letter);
	if (idx>=0)
		gsl_matrix_set(pssm, idx, column, value);
}

double get_pos(gsl_matrix *pssm, int column, char letter)
{
	// row order: A, G, C, T, a, g, c, t
	int idx;
	idx = find_index(primary_IUPAC_nucleotides, no_primary_IUPAC_nucleotides, letter);
	if (idx>=0)
		return gsl_matrix_get(pssm, idx, column);
	return -1;
}

// convert pssm to consensus sequence using rules derived from
/* Cavener, Nucleic Acids Res. 1987 Feb 25;15(4):1353-61.  Comparison of the consensus sequence flanking translational start sites in Drosophila and vertebrates.
"If the frequency of a single nucleotide at a specific position is greater than 50% and greater than twice the number of the second most frequent nucleotide it is assigned as the consensus nucleotide.
If the sum of the frequencies of two nucleotides is greater than 75% (but neither meet the criteria for a single nucleotide assignment) they are assigned as co-consensus nucleotides.
If no single nucleotide or pair of nucleotides meet the criteria of consensus nucleotide(s) the letter N Is assigned to that position."
*/
void pssm_to_IUPAC(gsl_matrix *pssm, char *motif)
{
	// row order: A, G, C, T, a, g, c, t
//	print_pssm(pssm);
	int column,i;
	bool not_set;
	double val[no_primary_IUPAC_nucleotides];
	int max_idx, next_to_max_idx;
	for(column=0;column<pssm->size2;column++)
	{
		not_set = true;
		max_idx = 0;
		next_to_max_idx = 0;
		// get the values and indices for top two
		for(i=0; i<no_primary_IUPAC_nucleotides; i++)
		{
			val[i] = get_pos(pssm, column, primary_IUPAC_nucleotides[i]);
			if(val[i]>val[next_to_max_idx])
				next_to_max_idx = i;
			if(val[i]>val[max_idx])
			{
				next_to_max_idx = max_idx;
				max_idx = i;
			}
		}
		if(max_idx==0) //need to get the next to max index if max on the first position
		{
			next_to_max_idx = 1;
			for(i=1; i<no_primary_IUPAC_nucleotides; i++)
				if(val[i]>val[next_to_max_idx])
					next_to_max_idx = i;
		}
		// if top value is greater than 0.5 on a row, and greater than twice the next value, set it at this position in the motif
		if((val[max_idx]>0.5)&&(val[max_idx]>2*val[next_to_max_idx]))
		{
			motif[column] = primary_IUPAC_nucleotides[max_idx];
			not_set = false;
		}
		// if not found, see if the sum of top 2 values is greater than 0.75 and put a degenerate symbol
		if (not_set)
		{
			if(val[max_idx]+val[next_to_max_idx] > 0.75)
			{
				motif[column] = get_degenerate_symbol(primary_IUPAC_nucleotides[max_idx],primary_IUPAC_nucleotides[next_to_max_idx]);
				not_set = false;
			}
		}
		// if still not found, then put 'N' or 'n'
		if (not_set)
		{
			double value_N=0, value_n=0;
			for(i=0; i<no_primary_IUPAC_nucleotides; i++)
				if (i<no_primary_IUPAC_nucleotides/2)
					value_N += val[i];
				else
					value_n += val[i];
			if (value_N>=value_n)
				motif[column] = 'N';
			else
				motif[column] = 'n';
		}
//		printf("column %i, max %c %8.4f, next_to_max %c  %8.4f => %c\n", column, primary_IUPAC_nucleotides[max_idx], val[max_idx], primary_IUPAC_nucleotides[next_to_max_idx], val[next_to_max_idx], motif[column]);
	}
//	printf("motif = %s\n", motif);
}

void IUPAC_to_pssm(char *motif, gsl_matrix *pssm)
{
	int column,i,degeneracy;
	char *encoding;
	int length = strlen(motif);
//	gsl_matrix_set_zero(pssm);

	// setting to default values in case of exact letter, where the remaining positions are filled with 0.006
	double non_zero = 0.006, rest;
	for(column=0;column<length;column++)
	{
		// fill everything with the small non-zero value
		for(i=0;i<no_primary_IUPAC_nucleotides;i++)
			set_pos(pssm, column, primary_IUPAC_nucleotides[i], non_zero);
		//add values to the pssm, depending of the motif letter
		degeneracy = get_degeneracy(motif[column]);
		if (degeneracy==0)
		{
			printf("error input consensus motif: contains illegal symbol '%c'!\n", motif[column]);
			exit(1);
		}
		if (degeneracy==1)
		{
			rest = 1-non_zero*(no_primary_IUPAC_nucleotides-1);
			set_pos(pssm, column, motif[column], rest);
		}
		else
		{
			encoding = get_encoding(motif[column]);
			rest = 1-non_zero*(no_primary_IUPAC_nucleotides-degeneracy);
			for (i=0; i<degeneracy; i++)
				set_pos(pssm, column, encoding[i], rest/degeneracy);
		}
	}
}

void test_pssm_to_IUPAC()
{
	int length = 5;
	char motif[5];
	char m[] = {"AYTAAY"};
	printf("%s\n", m);
	gsl_matrix *pssm = gsl_matrix_alloc(4, strlen(m));
	IUPAC_to_pssm(m, pssm);
	print_pssm(pssm);
	pssm_to_IUPAC(pssm, motif);
	printf("%s\n", motif);
	exit(0);
}

gsl_matrix *get_rev_compl_pssm(gsl_matrix *pssm)
{
	int column,i;
	gsl_matrix *rev_compl_pssm = gsl_matrix_alloc(no_primary_IUPAC_nucleotides, pssm->size2);
	for(column=0;column<pssm->size2;column++)
	{
		for (i=0; i<no_primary_IUPAC_nucleotides; i++)
			set_pos(rev_compl_pssm, pssm->size2-column-1, get_complement(primary_IUPAC_nucleotides[i]), get_pos(pssm, column, primary_IUPAC_nucleotides[i]));
	}
	return rev_compl_pssm;
}

#include "random_gen.h"

void test_rev_compl_pssm()
{
	int i;
	int length = 8;
	gsl_matrix *pssm = generate_PSSM(length);
	print_pssm(pssm);
	gsl_matrix *rev_compl_pssm = get_rev_compl_pssm(pssm);
	print_pssm(rev_compl_pssm);
	gsl_matrix_free(pssm);
	gsl_matrix_free(rev_compl_pssm);
	clean_up_ran_gen_static();
//	getchar();
}


void test_pssm()
{
	int i;
	int length = 8;
	gsl_matrix *pssm = generate_PSSM(length);
	print_pssm(pssm);

	char *file = "pssm1";
	write_mat(pssm, file);
	gsl_matrix *pssm_read = read_mat(file);
	print_pssm(pssm_read);

	char *motif;
	for(i=0;i<10;i++)
	{
		motif = draw_from_pssm(pssm_read);
		printf("%s\n", motif);
		FREE(motif);
	}
	gsl_matrix_free(pssm);
	getchar();
}

void print_pssm(gsl_matrix *pssm)
{
	int i,j;
	for(i=0;i<pssm->size1;i++)
	{
	   	printf("%c\t", primary_IUPAC_nucleotides[i]);
	   	for(j=0;j<pssm->size2;j++)
	   	{
	   		if(j != pssm->size2 - 1)
	   			printf("%-10.4lf\t", gsl_matrix_get(pssm, i, j));
	   		else
	   			printf("%-10.4lf\n", gsl_matrix_get(pssm, i, j));
	   	}
 	}
}
//
//void output_pssm_(gsl_matrix *pssm, file_name)
//{
//	int i,j;
//	FILE *fp = open_file(file_name, "w");
//	char pssm_rows[4] = {'A','G','C','T'};
//	for(i=0;i<pssm->size1;i++)
//	{
//	   	fprintf(fp, "%c\t", pssm_rows[i]);
//	   	for(j=0;j<pssm->size2;j++)
//	   	{
//	   		if(j != pssm->size2 - 1)
//	   			fprintf(fp, "%-10.4lf\t", gsl_matrix_get(pssm, i, j));
//	   		else
//	   			fprintf(fp, "%-10.4lf\n", gsl_matrix_get(pssm, i, j));
//	   	}
// 	}
//}


double sum(double *arr, int num)
{
	double sum = 0;
	int i;
	for(i=0;i<num;i++)
		sum += arr[i];
	return sum;
}
