/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */


// Read in each set of input files (binding evidence & sequence) and run motif discovery

#include "analysis.h"

extern gsl_matrix *covariates, *covariates_original, *covariates_shuffled;
extern bool is_regression_initialized;
extern const int no_primary_IUPAC_nucleotides;

double *top_random_scores;
int random_index;


void allocate_gene_level_scores(gene_level_scores *s)
{
	MALLOC(s->scores, sizeof(double) * s->num);
	MALLOC(s->index, sizeof(int) * s->num);
	MALLOC(s->sort_index, sizeof(score_index) * s->num);
	MALLOC(s->seq_ids, sizeof(char *) * s->num);
}

void free_gene_level_scores(gene_level_scores *s)
{
	FREE(s->seq_ids);
	FREE(s->sort_index);
	FREE(s->index);
	FREE(s->scores);
	free_file_lines(&s->raw_input, s->num);
}

void load_gene_level_scores(gene_level_scores *s, char *gene_level_scores_file)
{
	int i, num_lines;
	read_file_lines(gene_level_scores_file, &s->raw_input, &s->num);
	allocate_gene_level_scores(s);
	char *delim = "\t";
	for(i=0;i<s->num;i++)
	{
//		printf("%i\n", i);fflush(stdout);
		s->seq_ids[i] = strtok(s->raw_input[i], delim);
		s->scores[i]  = atof(strtok(NULL, delim));
		s->index[i]   = i;
		s->sort_index[i].index = i;
		s->sort_index[i].score = s->scores[i];
	}
}

double get_percentile(double percentile, double *seq_level_stats, int num_seq_scores)
{
	int i, first_non_zero_index;
	for(i=num_seq_scores-1;i>=0;i--)
	{
		if(seq_level_stats[i] > ZERO)
			{first_non_zero_index = i;break;}
	}
	return seq_level_stats[(int)(percentile * first_non_zero_index)];
}

/*
extern double *global_means, *global_vars, *global_mean_sum_squares;
extern double *global_medians, *global_IQRs;
extern double *means_KS, *vars_KS;
extern int *indexes, *random_gene_set;
*/

void init_geneset_scoring_params(geneset_set_scoring_params *scoring_params, int num_seq)
{
	int i;
	MALLOC(scoring_params->location_estimates, sizeof(double)*num_seq);
	MALLOC(scoring_params->scale_estimates, sizeof(double)*num_seq);
	MALLOC(scoring_params->mean_sum_squares_estimates, sizeof(double)*num_seq);
	MALLOC(scoring_params->random_gene_set, sizeof(int)*num_seq);
	MALLOC(scoring_params->indexes, sizeof(int)*num_seq);
	for(i=0;i<num_seq;i++)
		scoring_params->indexes[i]=i;
//	printf("done 'init_geneset_scoring_params'\n");getchar();
}


void free_geneset_scoring_params(geneset_set_scoring_params *scoring_params)
{
	FREE(scoring_params->location_estimates);
	FREE(scoring_params->scale_estimates);
	FREE(scoring_params->mean_sum_squares_estimates);
	FREE(scoring_params->random_gene_set);
	FREE(scoring_params->indexes);
}


int zero_threshold_index = -1;


double global_sample_mean, global_sample_var;

void read_seq_details_fixed_length(char *file, params *p)
{
	int i;
	gene_level_scores s;
	load_gene_level_scores(&s, file);

	p->d.num_seq_level_stats = s.num;
//	p->d.seq.seq_ids = 0;
	MALLOC(p->d.seq.seq_ids, sizeof(char *)*(p->d.num_seq_level_stats));
	MALLOC(p->d.seq_level_stats, sizeof(double)*(p->d.num_seq_level_stats));
	for(i=0;i<p->d.num_seq_level_stats;i++)
	{
		MALLOC(p->d.seq.seq_ids[i], sizeof(char)* strlen(s.seq_ids[i]) + 1);
		strcpy(p->d.seq.seq_ids[i], s.seq_ids[i]);
		p->d.seq_level_stats[i] = s.scores[i];
//		if(i % 50 == 0)
//			printf("%s(%s)\t%lf\n", s.seq_ids[i], p->d.seq.seq_ids[i], p->d.seq_level_stats[i]);
	}
	free_gene_level_scores(&s);
//	getchar();
	
	// make scores on top of the list to be smallest negative
	for(i=0;i<p->d.num_seq_level_stats;i++)
	{

		p->d.seq_level_stats[i] = -p->d.seq_level_stats[i];
		if(p->d.seq_level_stats[i] == 0 && zero_threshold_index == -1 || i == p->d.num_seq_level_stats - 1)
			zero_threshold_index = i;
	}

	if((int)(p->scoring_params.min_gene_set_size_percentage_threshold * p->d.num_seq_level_stats) > p->scoring_params.min_gene_set_size_absolute_threshold)
		p->scoring_params.min_gene_set_size_absolute_threshold = (int)(p->scoring_params.min_gene_set_size_percentage_threshold * p->d.num_seq_level_stats);
//	printf("min_gene_set_size_absolute_threshold = %i\n", p->scoring_params.min_gene_set_size_absolute_threshold);
	p->scoring_params.max_gene_set_size_absolute_threshold = (int)(p->scoring_params.max_gene_set_size_percentage_threshold * p->d.num_seq_level_stats);
//	printf("max_gene_set_size_absolute_threshold = %i\n", p->scoring_params.max_gene_set_size_absolute_threshold);
	p->d.is_initialized = true;

	init_geneset_scoring_params(&p->scoring_params, p->d.num_seq_level_stats);

	preprocess_thresholds_no_kNN(p->d.seq_level_stats, p->d.num_seq_level_stats, p->scoring_params.min_gene_set_size_absolute_threshold, &p->scoring_params);

	size_t stride = 1;
	double scores_local[p->d.num_seq_level_stats];
	for(i=0;i<p->d.num_seq_level_stats;i++)
	    	scores_local[i] = p->d.seq_level_stats[i];
	gsl_sort(scores_local, stride, p->d.num_seq_level_stats);
	p->d.median_seq_level_stats = gsl_stats_median_from_sorted_data(scores_local, stride, p->d.num_seq_level_stats);
	p->d.lower_percentile = 0.25;
	p->d.upper_precentile = 0.75;
	double global_IPR_lower = gsl_stats_quantile_from_sorted_data(p->d.seq_level_stats, stride, p->d.num_seq_level_stats, p->d.lower_percentile);
	double global_IPR_upper = gsl_stats_quantile_from_sorted_data(p->d.seq_level_stats, stride, p->d.num_seq_level_stats, p->d.upper_precentile);
	p->d.IPR_seq_level_stats = global_IPR_upper - global_IPR_lower;

	p->d.mean_seq_level_stats = gsl_stats_mean(p->d.seq_level_stats, stride, p->d.num_seq_level_stats);

	double mean_sum_squared = 0;
	for(i=0;i<p->d.num_seq_level_stats;i++)
		mean_sum_squared += (p->d.seq_level_stats[i])*(p->d.seq_level_stats[i]);
	mean_sum_squared = mean_sum_squared / p->d.num_seq_level_stats;
	p->d.var_seq_level_stats = mean_sum_squared - p->d.mean_seq_level_stats*p->d.mean_seq_level_stats;

	
}


void build_file_name(char **dest, char *base, char *append)
{
	MALLOC(*dest, sizeof(char)*(strlen(base)+strlen(append)+2));
	strcpy(*dest, base);
	if(strlen(append) > 0)
		strcat(*dest, "_");
	strcat(*dest, append);
}

// inputfile FORMAT: gene_id<TAB>gene_level_score<newline>...
void read_inputfile(dataset_file_names *d_files, char *inputfile_list_file)
{
	int i;
	FILE *fp = open_file(inputfile_list_file, "r");
	char *lineptr = 0; int n = 0; int index = 0;
	for(i=0;i<d_files->num_datasets;i++)
	{
		if(d_files->num_datasets == index)
		{
			printf("ERROR: 'analysis.c', 'read_inputfile', more input files than datasets, d_files->num_datasets!\n");
			exit(1);
		}

		// read a line from the input file
		getline(&lineptr, &n, fp);

		if(lineptr[strlen(lineptr) - 1] == '\n')
			lineptr[strlen(lineptr) - 1]= '\0';
		d_files->d[index++].inputfile = lineptr;
		lineptr = 0;
	}
	fclose(fp);
}




dataset *get_next_dataset(dataset_file_names *datasets)
{
	if(datasets->current_index < datasets->num_datasets)
	{
		datasets->current_index++;
		return &(datasets->d[datasets->current_index-1]);
	}
	else
		return NULL;
}


void free_data(data *d)
{
	int i;
	for(i=0;i<d->num_seq_level_stats;i++)
		FREE(d->seq.seq_ids[i]);
	FREE(d->seq_level_stats);
	FREE(d->seq.seq_ids);
	d->seq.is_initialized = false;
}


void free_sequence_data(params *parameters)
{
	int i;
	destroy_index(&parameters->d.seq.index);
	parameters->d.seq.is_initialized = false;
	if(parameters->use_conservation == true)
	{
		for(i=0;i<parameters->d.num_orth_sequences;i++)
		{
			destroy_index(&parameters->d.orth_sequences[i].index);
			parameters->d.orth_sequences[i].is_initialized = false;
		}
		FREE(parameters->d.orth_sequences);
	}
	free_data(&(parameters->d));
}


bool is_cERMIT_format(char *pssm_file)
{
 	FILE *fp = fopen(pssm_file, "r");
 	if(fp == NULL)
 		return false;

	char *line = 0; int line_length = 0;
	int num_read = getdelim(&line, &line_length, '\n', fp);
//	printf("num_read = %i/tline: %s\n", num_read, line);
	if(line[(num_read - 1)*sizeof(char)] == '\n')
		line[(num_read - 1)*sizeof(char)] = '\0';
	int row, col;
	sscanf(line, "%i\t%i\n", &row, &col);
	FREE(line);
//	printf("row = %i\tcol = %i\n", row, col);
	fclose(fp);

	return (row == 4) && (col > 0);
}


extern gsl_matrix *U_global;

void subtract_off_covariate_effects(double *y, int n, gsl_matrix *U, double *corrected_y_out)
{
	U_global = U;
//	printf("n = %i\n", n);
	gsl_matrix_view mview_y = gsl_matrix_view_array(y, n, 1);
//	printf("U\n");
//	print_matrix_GSL_m_n(U, 5, 5);
//	getchar();

//	printf("&mview_y.matrix\n");
//	print_matrix_GSL_m_n(&mview_y.matrix, 10, 1);
//	getchar();
	gsl_matrix *U_tr_times_y = gsl_matrix_alloc(U->size2, 1);
	mat_mat_multiply(U, &mview_y.matrix, U_tr_times_y, CblasTrans, CblasNoTrans);
//	printf("U_tr_times_y\n");
//	print_matrix_GSL_m_n(U_tr_times_y, 5, 5);

	gsl_matrix *correction_vector_y = gsl_matrix_alloc(n, 1);

	// correction_vector_y = {Projection of y onto col_space(U)}
	mat_mat_multiply(U, U_tr_times_y, correction_vector_y, CblasNoTrans, CblasNoTrans);

	gsl_matrix_view mview_corrected_y = gsl_matrix_view_array(corrected_y_out, n, 1);
	gsl_matrix_memcpy(&mview_corrected_y.matrix, &mview_y.matrix);

	// y = y - {Projection of y onto col_space(U)}
	gsl_matrix_sub(&mview_corrected_y.matrix, correction_vector_y);

//	printf("&mview_corrected_y.matrix\n");
//	print_matrix_GSL_m_n(&mview_corrected_y.matrix, 10, 1);
	gsl_matrix_free(correction_vector_y);
//	getchar();
}

double mean_before_centering;

void center_data(double *data, int num)
{
	int i;
	double mean = gsl_stats_mean(data, 1, num);
	mean_before_centering = mean;
	for(i=0;i<num;i++)
		data[i] = data[i] - mean;
}


extern bool do_PC_regression;
extern bool add_intercept;


void run_analysis(params *parameters, IO_params *p_IO)
{
	time_t timer;
	int i,j,k;
	char *delim = "\t";

	// score file
	char score_file[strlen(p_IO->output_dir) + strlen(p_IO->score_file) + 1];
	strcpy(score_file, p_IO->output_dir);
	strcat(score_file, p_IO->score_file);
	// seq file
	char seq_file[strlen(p_IO->output_dir) + strlen(p_IO->seq_file) + 1];
	strcpy(seq_file, p_IO->output_dir);
	strcat(seq_file, p_IO->seq_file);

	char null_scores_filename_root[1000];
	if(parameters->significance.evaluate_significance == true)
	{
		MALLOC(top_random_scores, sizeof(double)*parameters->significance.num_random_permutations);
		sprintf(null_scores_filename_root, "%s", p_IO->null_scores_filename);
	}

	random_index = 0;
	parameters->d.num_orth_sequences = 0;
	if(p_IO->cov_file != 0)
	{
		// cov file
		char cov_file[strlen(p_IO->output_dir) + strlen(p_IO->cov_file) + 1];
		strcpy(cov_file, p_IO->output_dir);
		strcat(cov_file, p_IO->cov_file);
		// init regression
		init_regression(cov_file);
	}


	// potentially no values can be provided in which case the default value is '0'
	char *pssm_file = strtok(NULL, delim);
	if(pssm_file != 0)
	{
		if(pssm_file[0] == '/')
		{
			char tmp[100];
			strcpy(tmp, pssm_file+1);
			strcpy(pssm_file, tmp);
		}

		double pseudocount = 1;
		if(file_exist(pssm_file) == true)
		{
			if(is_cERMIT_format(pssm_file) == true)
			{
				parameters->known_pssm = read_mat(pssm_file);
			}
			else
			{
				int num_cols = find_num_cols_TRASNFAC_pssm(pssm_file);
				printf("reading TRANSFAC pssm: num_cols = %i\n", num_cols);
				parameters->known_pssm = gsl_matrix_alloc(no_primary_IUPAC_nucleotides, num_cols);
				read_pssm_TRANSFAC_like_format_counts(parameters->known_pssm, pssm_file, pseudocount);
			}
		}
		else
		{
			parameters->known_pssm = 0;
		}
	}
	else
	{
		parameters->known_pssm = 0;
	}

	// read scores
	read_seq_details_fixed_length(score_file, parameters);

	double copy_seq_level_stats[parameters->d.num_seq_level_stats];
	for(j=0;j<parameters->d.num_seq_level_stats;j++)
		copy_seq_level_stats[j] = parameters->d.seq_level_stats[j];

	// get case name
	char *base = basename(p_IO->seq_file);
	tokenizer t;
	if(strncmp(base, "seq_", strlen("seq_")) == 0)
	{
		t  = create_tokenizer(base,"seq_", true);
		parameters->TF_name = t.tokens[1];
	}
	else
	{
		t  = create_tokenizer(base,".seq", true);
		parameters->TF_name = t.tokens[0];
	}

	// create path with case name
	char base_name[strlen(p_IO->output_dir) + strlen(parameters->TF_name) + 1];
	strcpy(base_name, p_IO->output_dir);
	strcat(base_name, parameters->TF_name);

	// create names for output files
	char *to_append =  "";//"web_logo";
	build_file_name(&p_IO->out_weblogo, base_name, to_append);
	to_append = "motif_summary.txt";
	build_file_name(&p_IO->out_summary, base_name, to_append);
	to_append = "extra_info.txt";
	build_file_name(&p_IO->out_extra, base_name, to_append);

	read_sequence_input(seq_file, &parameters->d.seq, parameters->d.seq.delim);
	if(is_regression_initialized == true)
		covariates = covariates_original;

	if(parameters->significance.evaluate_significance == true)
	{
		gsl_permutation *per = gsl_permutation_alloc(parameters->d.num_seq_level_stats);
		int num_permutation_runs = parameters->significance.num_random_permutations;
		printf("generating randomized scores...\n");
//			permute1(parameters->d.seq_level_stats, parameters->d.num_seq_level_stats, per);
		sprintf(p_IO->null_scores_filename, "%s_%i_draws_%s", null_scores_filename_root, num_permutation_runs, parameters->TF_name);
		FILE *fp = open_file(p_IO->null_scores_filename, "w");
		fclose(fp);

		int percent_step_display = 10;
		int display_interval = (num_permutation_runs / percent_step_display) > 0 ? (num_permutation_runs / percent_step_display) : 1;

		int ii = num_permutation_runs;

		while(ii > 0)
		{
			timer = start_timer();

			permute1(parameters->d.seq_level_stats, parameters->d.num_seq_level_stats, per);
			global_sample_mean = gsl_stats_mean(parameters->d.seq_level_stats, 1, parameters->d.num_seq_level_stats);
			double sd1 = gsl_stats_sd_m(parameters->d.seq_level_stats, 1, parameters->d.num_seq_level_stats, global_sample_mean);
			global_sample_var = sd1*sd1;

			bool success = run_motif_finder(parameters, p_IO);
			if(success == true)
				ii--;

			if( (num_permutation_runs - ii) % display_interval == 0)
			{
				if(parameters->significance.evaluate_significance == true)
					printf("%i percent done...", ((num_permutation_runs - ii) / display_interval) * percent_step_display);
				print_time(timer, stdout);
			}
//				permute1(parameters->d.seq_level_stats, parameters->d.num_seq_level_stats, per);
		}
		gsl_permutation_free(per);
	}

	FILE *fp = open_file(p_IO->out_summary, "w");
//	fprintf(fp, "Score file: %s\n", basename(score_file));
	fflush(fp);
	if(parameters->known_pssm != 0)
		fprintf(fp, "PSSM file    : %s\n\n", basename(pssm_file));
//	else
//		fprintf(fp, "\n");
	fclose(fp);

	for(j=0;j<parameters->d.num_seq_level_stats;j++)
		parameters->d.seq_level_stats[j] = copy_seq_level_stats[j];
	global_sample_mean = gsl_stats_mean(parameters->d.seq_level_stats, 1, parameters->d.num_seq_level_stats);
	double sd1 = gsl_stats_sd_m(parameters->d.seq_level_stats, 1, parameters->d.num_seq_level_stats, global_sample_mean);
	global_sample_var = sd1*sd1;

	bool tmp_store_evaluate_significance = parameters->significance.evaluate_significance;
	parameters->significance.evaluate_significance = false;
	printf("processing %s...", parameters->TF_name);

	// run motif finder
	run_motif_finder(parameters, p_IO);
	parameters->significance.evaluate_significance = tmp_store_evaluate_significance;

	// free memory
	free_sequence_data(parameters);


	destroy_tokenizer(&t);
//		destroy_tokenizer(&t1);
//	printf("done\n");fflush(stdout);

	free_geneset_scoring_params(&parameters->scoring_params);
	FREE(p_IO->out_summary);
	FREE(p_IO->out_weblogo);

/*	for(j=0;j<d_files.d[i].num_seqfile_list;j++)
		FREE(d_files.d[i].seqfile_list[j]);
//		printf("after for(j=0;j<d_files.d[i].num_seqfile_list;j++)\n");fflush(stdout);
	FREE(d_files.d[i].seqfile_list);

	if(p_IO->covariates_list != 0)
	{
		FREE(cov_files.d[i].seqfile_list[0]);
		FREE(cov_files.d[i].seqfile_list);
	}
	FREE(data_files->inputfile);*/

	if(p_IO->cov_file != 0)
		free_regression();

//		printf("after data_files->inputfile\n");fflush(stdout);
	if(parameters->known_pssm != 0)
		gsl_matrix_free(parameters->known_pssm);
	// out of for
/*	if(p_IO->covariates_list != 0)
		FREE(cov_files.d);*/
	if(parameters->significance.evaluate_significance == true)
		FREE(top_random_scores);
}

void test_gene_level_scores()
{
	int i;
//	char *gene_level_scores_file = "/home/sgg4/Desktop/main_project/data/yeast/input_156/inputfile_ABF1_YPD";
	char *gene_level_scores_file = "/nfs/labs/ohlerlab/sata/data/sgeorg/cERMIT_v1.1/source/test_all/input/data/evidence_hsa-miR1";
	gene_level_scores s;
	load_gene_level_scores(&s, gene_level_scores_file);
	for(i=0;i<s.num;i++)
	{	
		if(i < 50 || s.num - i < 50)
			printf("%lf\n", s.scores[i]);fflush(stdout);
	}	

	sort_order so = decreasing;
	sort_gene_level_scores(&s, s.num, so);
	for(i=0;i<s.num;i++)
	{	
		if(i < 50 || s.num - i < 50)
			printf("%lf\n", s.scores[i]);fflush(stdout);
	}

	free_gene_level_scores(&s);
	printf("done\n");
	getchar();
}

