/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#include "eigsvd.h"

//extern gsl_rng *r1;

void SVD_Golub_Reinsch(gsl_matrix *A, eigs_SVD *out)
{
/*
	// This function computes the SVD using the modified Golub-Reinsch algorithm, which is 
	// faster for M>>N. It requires the vector work of length N and the N-by-N matrix X as 
	// additional working space.
	res = gsl_linalg_SV_decomp_mod (gsl_matrix *A, gsl_matrix *X, gsl_matrix *V, gsl_vector *S, gsl_vector *work)
*/
	gsl_matrix *workMat;
	gsl_vector *workVec;
	if(A->size1 < A->size2)
 	{		
		printf("ERROR: 'SVD_Golub_Reinsch', M < N, %i\t%i\n", A->size1, A->size2);exit(1);
 	}
	else
	{
		gsl_matrix_memcpy(out->U, A);
		int res;
		workVec = gsl_vector_alloc(A->size2);
		res = gsl_linalg_SV_decomp(out->U, out->V, out->S, workVec);
		gsl_vector_free(workVec);
		if(res != 0)
			printf("ERROR: 'gsl_linalg_SV_decomp', %i\n", res);
	}
}

void SVD_modified_Golub_Reinsch(gsl_matrix *A, eigs_SVD *out)
{
/*
	// This function computes the SVD using the modified Golub-Reinsch algorithm, which is 
	// faster for M>>N. It requires the vector work of length N and the N-by-N matrix X as 
	// additional working space.
	res = gsl_linalg_SV_decomp_mod (gsl_matrix *A, gsl_matrix *X, gsl_matrix *V, gsl_vector *S, gsl_vector *work)
*/
	gsl_matrix *workMat;
	gsl_vector *workVec;
	if(A->size1 < A->size2)
 	{		
		printf("ERROR: 'SVD_modified_Golub_Reinsch', M < N, %i\t%i\n", A->size1, A->size2);exit(1);
 	}
	else
	{
		gsl_matrix_memcpy(out->U, A);
//		printf("SVD_modified_Golub_Reinsch: out->U\n");
//		print_matrix_GSL_m_n(out->U, 10, 10);

		int res;
		workVec = gsl_vector_alloc(A->size2);
		workMat = gsl_matrix_alloc(A->size2, A->size2);
		res = gsl_linalg_SV_decomp_mod(out->U, workMat, out->V, out->S, workVec);
//		printf("SVD_modified_Golub_Reinsch: out->U\n");
//		print_matrix_GSL_m_n(out->U, 10, 10);
//		getchar();
		gsl_matrix_free(workMat);
		gsl_vector_free(workVec);
		if(res != 0)
			printf("ERROR: 'gsl_linalg_SV_decomp_mod', %i\n", res);
	}
}

void SVD_Jacobi(gsl_matrix *A, eigs_SVD *out)
{
/*
	// This function computes the SVD using the modified Golub-Reinsch algorithm, which is 
	// faster for M>>N. It requires the vector work of length N and the N-by-N matrix X as 
	// additional working space.
	res = gsl_linalg_SV_decomp_mod (gsl_matrix *A, gsl_matrix *X, gsl_matrix *V, gsl_vector *S, gsl_vector *work)
*/
	gsl_matrix *workMat;
	gsl_vector *workVec;
	if(A->size1 < A->size2)
 	{		
		printf("ERROR: 'SVD_Jacobi', M < N, %i\t%i\n", A->size1, A->size2);exit(1);
 	}
	else
	{
		gsl_matrix_memcpy(out->U, A);
		int res;
		res = gsl_linalg_SV_decomp_jacobi(out->U, out->V, out->S);
		if(res != 0)
			printf("ERROR: 'gsl_linalg_SV_decomp_jacobi', %i\n", res);
	}
}


void QR_decomp(gsl_matrix *A, gsl_matrix *Q, gsl_matrix *R)
{
	int signum;
	gsl_vector *tau = gsl_vector_alloc(A->size2);
	gsl_vector *norm = gsl_vector_alloc(A->size2);
	gsl_permutation *p = gsl_permutation_alloc(A->size2);

	// This function factorizes the matrix A (M-by-N) into the decomposition A = Q R P^T 
	// without modifying A itself and storing the output in the separate matrices q and r. 
	// The vector tau must be of length k=\min(M,N)
	// The vector norm is a workspace of length N used for column pivoting
	int res = gsl_linalg_QRPT_decomp2(A, Q, R, tau, p, &signum, norm);
	if(res != 0)
		printf("ERROR: 'gsl_linalg_QRPT_decomp2', %i\n", res);
	gsl_permutation_free(p);
	gsl_vector_free(norm);
	gsl_vector_free(tau);
}

/*

//  DGESVD implements the Golub Kahan bidiagonal QR algorithm


// DGESVD prototype
void dgesvd_(char* jobu, char* jobvt, int* m, int* n, double* a,
                int* lda, double* s, double* u, int* ldu, double* vt, int* ldvt,
                double* work, int* lwork, int* info);

void SVD_Golub_Kahan(gsl_matrix *mat, eigs_SVD *eig, eigs_SVD_Golub_Kahan_options options)
{
//	printf("U = %i, V = %i, S = %i\n", options.estimate_U, options.estimate_V,options.estimate_S);
        // Locals 
        int n = mat->size1, m = mat->size2;
	int lda = m, ldu = m, ldvt = m, info, lwork;
        double wkopt;
        double *work;

        // Local arrays 
        double *s = (double*)malloc(m*sizeof(double));
        double *u = 0, *vt = 0;

	char JOBVT[2], JOBU[2];
	JOBVT[1] = 0;JOBU[1] = 0;

//	printf("U\n");
	if(options.estimate_U == true)
	{
//		ldvt = m;
//		vt = (double*)malloc(m*n*sizeof(double));
//		vt = (double*)malloc(ldvt*n*sizeof(double));
//		printf("size U: %i-by-%i\n", n, m);
//		strcpy(JOBVT, "S");
		strcpy(JOBVT, "O");
	}
	else
		strcpy(JOBVT, "N");
//	printf("V\n");
	if(options.estimate_V == true)
	{
		u = (double*)malloc(ldu*m*sizeof(double));
		strcpy(JOBU, "S");
	}
	else
		strcpy(JOBU, "N");
//	printf("query\t");fflush(stdout);

        // Query and allocate the optimal workspace
        lwork = -1;
        dgesvd_(JOBU, JOBVT, &m, &n, mat->data, &lda, s, u, &ldu, vt, &ldvt, &wkopt, &lwork, &info);

        lwork = (int)wkopt;

//	printf("lwork    = %i\n", lwork);
//	printf("mat size = %i\n", ldu*m*sizeof(double));

//	lwork = 5 * lwork;
        work = (double*)malloc( lwork*sizeof(double) );

 //	printf("Compute SVD\t");fflush(stdout);

        // Compute SVD
        dgesvd_(JOBU, JOBVT, &m, &n, mat->data, &lda, s, u, &ldu, vt, &ldvt, work, &lwork, &info);
 
//	printf("done\n");fflush(stdout);

        // Check for convergence
        if(info > 0) 
	{
                printf("ERROR: (SVD_Golub_Kahan), the algorithm computing SVD failed to converge!\n");
                exit(1);
        }
	free(work);

//	printf("here\n");fflush(stdout);

	if(options.estimate_S == true)
	{
//		printf("eig->S->size = %i\n", eig->S->size);getchar();
		gsl_vector_view v_view_S = gsl_vector_view_array(s, eig->S->size);
//		printf("v_view_S.vector.size= %i\n", v_view_S.vector.size);fflush(stdout);
		gsl_vector_memcpy(eig->S, &v_view_S.vector);
	}
	free(s);
//	printf("lwork = %i\n", lwork);fflush(stdout);

	if(options.estimate_U == true)
	{
//		gsl_matrix_view m_view_U = gsl_matrix_view_array_with_tda(vt, n, m, m);
//		gsl_matrix_view m_view_U = gsl_matrix_view_array_with_tda(mat->data, n, m, m);
		gsl_matrix_view m_view_U = gsl_matrix_view_array_with_tda(mat->data, n, eig->U->size2, m);
		gsl_matrix_memcpy(eig->U, &m_view_U.matrix);
//		free(vt);
//		printf("options.estimate_U == true\n");
//		print_matrix_GSL_m_n(eig->U,4,10);
	}

//	printf("after options.estimate_U == true\n");fflush(stdout);
	if(options.estimate_V == true)
	{
		gsl_matrix_view m_view_V = gsl_matrix_view_array(u, eig->V->size2, m);
//		gsl_matrix_transpose(&m_view_V.matrix);
		gsl_matrix_transpose_memcpy(eig->V, &m_view_V.matrix);
//		gsl_matrix_memcpy(eig->V, &m_view_V.matrix);
		free(u);
//		printf("options.estimate_V == true\n");
//		print_matrix_GSL_m_n(eig->V,4,10);
	}
}
*/

void test_SVD()
{
	int i; time_t t2;
	init_ran_gen_static();
	gsl_rng *r = get_rng();
	pt2svd svd_ptr;

	int m = 10000, n = 2100;
//	gsl_matrix *A = generate_simple_Gaussian(m,n,r);
	gsl_matrix *A = gsl_matrix_alloc(m, n);
	generate_simple_Gaussian_in_place(m, n, A, r);
//	gsl_matrix_set_identity(A);
 	printf("data matrix\n");
 	print_matrix_GSL_m_n(A,5,5);

	eigs_SVD out;
	out.U = gsl_matrix_alloc(m, n);
	out.S = gsl_vector_alloc(n);
	out.V = gsl_matrix_alloc(n, n);

	gsl_matrix *Acopy = gsl_matrix_alloc(A->size1, A->size2);

	gsl_matrix_memcpy(Acopy, A);
 	printf("svd: modified_Golub_Reinsch\n");
	svd_ptr = SVD_modified_Golub_Reinsch;
	t2 = start_timer();
	svd_ptr(A, &out);
//	SVD_modified_Golub_Reinsch(A, &out);
 	printf("modified_Golub_Reinsch: U\n");
 	print_matrix_GSL_m_n(out.U,5,5);
 	printf("modified_Golub_Reinsch: S\n");
 	print_vector_n_GSL(out.S, 5);
	print_time(t2, stdout);	

 	printf("svd: Golub_Reinsch\n");
	svd_ptr = SVD_Golub_Reinsch;
	t2 = start_timer();
	svd_ptr(A, &out);
 	printf("Golub_Reinsch: U\n");
 	print_matrix_GSL_m_n(out.U,5,5);
 	printf("Golub_Reinsch: S\n");
 	print_vector_n_GSL(out.S, 5);
	print_time(t2, stdout);	

 	printf("svd: Jacobi\n");
	svd_ptr = SVD_Jacobi;
	t2 = start_timer();
	svd_ptr(A, &out);
 	printf("Jacobi: U\n");
 	print_matrix_GSL_m_n(out.U,5,5);
 	printf("Jacobi: S\n");
 	print_vector_n_GSL(out.S, 5);
	print_time(t2, stdout);	


	gsl_matrix *A_reconstruct = gsl_matrix_alloc(A->size1, A->size2);
	gsl_matrix *A_reconstruct1 = gsl_matrix_alloc(A->size1, A->size2);
	gsl_matrix *S_diag = gsl_matrix_alloc(out.S->size, out.S->size);
	gsl_matrix_set_zero(S_diag);

	for(i=0;i<out.S->size;i++)
		gsl_matrix_set(S_diag, i, i, out.S->data[i]);
	mat_mat_multiply(out.U, S_diag, A_reconstruct, CblasNoTrans, CblasNoTrans);
	mat_mat_multiply(A_reconstruct, out.V, A_reconstruct1, CblasNoTrans, CblasTrans);
 	printf("reconsctructed: S\n");
	print_matrix_GSL_m_n(A_reconstruct1,5,5);
	printf("done\n");
	exit(0);

}
