/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#ifndef PARAMETERS_H_
#define PARAMETERS_H_

#include <gsl/gsl_matrix.h>
#include <stdbool.h>

#include "geneset_scoring_interface.h"
#include "motif_evolution_interface.h"

typedef enum {Random_Set_ranks, Random_Set_no_ranks, GSEA} score_type;
typedef enum {bayes_factors, posterior_probs} calibration_type;
typedef enum {chip_chip, chip_seq, micro_RNA, no_evidence_preprocessing} dataset_type;
typedef enum {yeast, human, arabidopsis, unspecified} species;

typedef struct
{
	bool calibrate;
	calibration_type type;
	bool apply_log_transform;
}p_values_calibration_param;

typedef struct
{
	unsigned char **oligos;
	char **raw_input;
	int num_oligos;
	bool is_empty;
}oligomers;

typedef struct
{
	char *gene_id;
	double *expr_profile;
	char *seq;
}
gene;

typedef struct
{
	gene *genes; 
	int num_genes, num_samples; char *delim;
	double *pattern_to_correlate_with;
	bool is_initialized;
}expr_data;

typedef struct
{
	char *delim;
	void **index; int num_seq, seq_length;
	char **seq_ids;
	bool is_initialized;
}seq_data;

typedef struct
{
	double *seq_level_stats;
	int num_seq_level_stats;

	double mean_seq_level_stats, var_seq_level_stats;
	double median_seq_level_stats;

	// inter-percentile range (10-th to 90-th percentile)
	double IPR_seq_level_stats;
	double upper_precentile, lower_percentile;

	expr_data expr;
	seq_data seq;
	seq_data *orth_sequences;int num_orth_sequences;
	
	bool is_initialized;
}data;


typedef struct
{
	bool evaluate_significance; 
	int num_random_permutations; 
	int num_random_permutations_remaining;
}signficance_params;

typedef struct
{
	// initialized in 'analysis.c'
	gsl_matrix *known_pssm;
	char *literature_conensus;
	
	gsl_matrix *ensemble_pssm[20];

	// initialized in 'params_reader.c'
	oligomers oligo_seeds;
	score_type scoring;

	gsl_matrix *region_sim_mat, *kNN;
	char *kNN_oligos_file;
	int num_neighbours; // number of neighbors in 'kNN'

	// degeneracy filters
	double degen_threshold;
	double fraction_degen_pos_threshold;
//	double threshold_Harb;
	double cluster_sim_threshold;
	
 	int max_motif_length, min_motif_length;
	int required_core_length; // required overlap for clustering redundant motifs into motif clusters
	double pssm_crop_threshold; // crop all flanking positions in the pssm that have <= pssm_crop_threshold * (# reads of "core")
	double cluster_pssm_threshold_fraction; // (for 'ensemble' pssm construction) consider all cluster members that have score >= cluster_pssm_threshold_fraction * top_scoring_member

	// fraction of top sequences to use for PSSM construction of each geneset
	double fraction_of_top_sequences_to_consider;

	bool microRNA_input;
	double hypegeom_p_val_cutoff;
	bool use_conservation;
	char *TF_name;

	data d;

	bool evaluate_significance; int num_random_permutations; int num_random_permutations_remaining;
	signficance_params significance;

	// number of std relative to the mean of the null scores (used when scoring a motif)
	double significance_threshold_filter;

	// yes -> evolve only "promising" motif seeds (that have positive scores)
	// no  -> evolve all motif seeds (slower option)
	bool fast_mode;

	int num_clusters_to_output_motif_occurrences_for;
	geneset_set_scoring_params scoring_params;
	pt2geneset_ES geneset_ES;
	pt2select_next_motif select_next_greedy;
}params;

void free_parameters(params *p);

void init_geneset_constraints(params *p, char *seq_constraints_file);

typedef struct
{
	// initialized in 'chi_chip.c'
	char *score_file; //*inputfile_list;
	char *cov_file; //*covariates_list;
	char *seq_file; //*input_seq_files;
	char *out_weblogo; // output_dir+params.TF_name
	char *out_summary;
	char *out_extra;
//	char *out_summary_merging;

	char *input_seeds_file;
	
	int num_motif_clusters_to_output;
//	double fraction_of_top_sequences_to_consider;

	// fraction of the counts of the 'core' part required for the outtermost positions 
	// the PSSM is trimmed until the counts are above that threshold
//	double trim_fraction;

	// initialized in 'params_reader.c'
	char *output_dir; 
	char *null_scores_filename;
	char *seed_motif_scores_filename;
	// image file type (eps, png, pdf)
	char *img_type;
}IO_params;


#endif /*PARAMETERS_H_*/
