/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

// Module that implements some similarity metrics to compare 2 motifs

#include "similarity_metrics.h"

double estimate_entropy(gsl_vector *porb_vec, double eps)
{
	double eps_to_use = 1.0e-5;
	if(eps > 0 && eps < 1.0)
		eps_to_use = eps;
	double entropy = 0;
	int i;
	for(i=0;i<porb_vec->size;i++)
	{
		double val = gsl_vector_get(porb_vec, i);
//		printf("%lf, ", val);
		entropy += (val + eps_to_use) * ( log(val + eps_to_use ) / log(2) );
	}
//	printf("\n");
	return -entropy;
}

double ave_entropy(gsl_matrix *pssm, double eps)
{
	int i, j;
	double ave_entropy = 0;
	for(i=0;i<pssm->size2;i++)
	{
		gsl_vector_view vview = gsl_matrix_column(pssm, i);
		double entropy = estimate_entropy(&vview.vector, eps);
		ave_entropy += entropy;
//		printf("col %i = %lf\n", i, entropy);
//		getchar();
	}
//	getchar();
	ave_entropy /= pssm->size2;
	return ave_entropy;
}

bool satisfy_entropy_constraint(gsl_matrix *pssm)
{
	double default_eps = -1;
	double ave_ent = ave_entropy(pssm, default_eps);
	return (ave_ent <= 1.0);
}

void test_ave_entropy()
{
	gsl_matrix *pssm = gsl_matrix_alloc(4,3);
	gsl_matrix_set(pssm, 0, 0, 1);
	gsl_matrix_set(pssm, 1, 0, 0);
	gsl_matrix_set(pssm, 2, 0, 0);
	gsl_matrix_set(pssm, 3, 0, 0);

	gsl_matrix_set(pssm, 0, 1, .5);
	gsl_matrix_set(pssm, 1, 1, .5);
	gsl_matrix_set(pssm, 2, 1, 0);
	gsl_matrix_set(pssm, 3, 1, 0);

	gsl_matrix_set(pssm, 0, 2, .25);
	gsl_matrix_set(pssm, 1, 2, .25);
	gsl_matrix_set(pssm, 2, 2, .25);
	gsl_matrix_set(pssm, 3, 2, .25);

	double res = ave_entropy(pssm, -1);
	print_pssm(pssm);
	printf("ave_entropy = %lf\n", res);
	getchar();
}

//int offset_top_score_global;

double similarity_score_Harbison_equal_length(gsl_matrix *f, gsl_matrix *g)
{
	int i, j;
	double res = 0;
	double res_col;
	for(i=0;i<f->size2;i++)
	{
		res_col = 0;
		for(j=0;j<f->size1;j++)
			res_col += pow((gsl_matrix_get(f, j, i) - gsl_matrix_get(g, j, i)), 2);
		res += sqrt(res_col);
	}
	return 1 - res / (f->size2 * sqrt(2));
}

// Harbison
double similarity_score_Harbison_with_overhang_no_background(gsl_matrix *pssm1, gsl_matrix *pssm2, int required_core_length, int *offset_top_score)
{
	int i;
	gsl_matrix_view g_view_1, g_view_2;
//	init_background_freq();

	// make sure 'pssm1' is longer than 'pssm2'
	if(pssm1->size2 < pssm2->size2)
	{
		// swap pssms
		gsl_matrix *tmp_pssm_ptr = pssm1;
		pssm1 = pssm2;
		pssm2 = tmp_pssm_ptr;
	}

	double sim_score = -10; double tmp_score;
	int overhang;
//	int required_core_length = 6;

//	if(pssm1->size2 == 6 && pssm2->size2 == 6)
//		required_core_length = 5;

	if(pssm2->size2 >= required_core_length)
	{
//		printf("pssm2->size2: %i\tpssm1->size2: %i\n", pssm2->size2, pssm1->size2);
		// shorter motif is at least 6 bases long
		overhang = pssm2->size2 - required_core_length;
		int num_cols, upper_left_col;
		int num_comparisons = 2 * overhang + pssm1->size2 - pssm2->size2 + 1;
		int upper_left_row = 0, num_rows = pssm2->size1;

//		offset_top_score_global = -overhang;
		*offset_top_score = -overhang;
		for(i=0;i<num_comparisons;i++)
		{
	//		printf("i: %i\n", i);
			if(i < overhang)
			{
		//		printf("%i < %i\n", i, overhang);
				//	pssm1: 0 -> pssm2->size2 - (overhang - i)
				//	pssm2: overhang - i -> pssm2->size2
				num_cols = pssm2->size2 - (overhang - i);

				upper_left_col = 0;
				g_view_1 = gsl_matrix_submatrix(pssm1, upper_left_row, upper_left_col, num_rows, num_cols);

				upper_left_col = overhang - i;
				g_view_2 = gsl_matrix_submatrix(pssm2, upper_left_row, upper_left_col, num_rows, num_cols);

			}
			else if(i <= overhang + pssm1->size2 - pssm2->size2)
			{
		//		printf("%i < %i + %i - %i\n", i, overhang, pssm2->size2, pssm1->size2);
				//	pssm1: i -> pssm2->size2 + i
				//  pssm2: 0 -> pssm2->size2
				num_cols = pssm2->size2;

				upper_left_col = i - overhang;
				g_view_1 = gsl_matrix_submatrix(pssm1, upper_left_row, upper_left_col, num_rows, num_cols);

				upper_left_col = 0;
				g_view_2 = gsl_matrix_submatrix(pssm2, upper_left_row, upper_left_col, num_rows, num_cols);
			}
			else
			{
				//  pssm1: pssm1->size2 - pssm2->size2 + i -> pssm2->size2
				//  pssm1: 0 -> pssm1->size2 - i
				num_cols = pssm1->size2 - i + overhang;

				upper_left_col = i - overhang;
			//	printf("%i: num_cols:%i\tupper_left_col: %i\n", i, num_cols, upper_left_col);
				g_view_1 = gsl_matrix_submatrix(pssm1, upper_left_row, upper_left_col, num_rows, num_cols);

				upper_left_col = 0;
				g_view_2 = gsl_matrix_submatrix(pssm2, upper_left_row, upper_left_col, num_rows, num_cols);
			}

			tmp_score = similarity_score_Harbison_equal_length(&g_view_1.matrix, &g_view_2.matrix);
			if(tmp_score > sim_score)
			{
//				offset_top_score_global = i - overhang;
				*offset_top_score = i - overhang;
				sim_score = tmp_score;
			}
		}
	//	printf("sim_score: %lf", sim_score);
	//	getchar();
	}
	else
	{
		overhang = 0;
		int num_comparisons = pssm1->size2 - pssm2->size2 + 1;
		int upper_left_row = 0, num_rows = pssm2->size1, num_cols = pssm2->size2;
		int upper_left_col;
		gsl_matrix_view view;

//		offset_top_score_global = -overhang;
		*offset_top_score = -overhang;
		for(i=0;i<num_comparisons;i++)
		{
			upper_left_col  = i;
			view = gsl_matrix_submatrix(pssm1, upper_left_row, upper_left_col, num_rows, num_cols);
			tmp_score = similarity_score_Harbison_equal_length(&view.matrix, pssm2);
			if(tmp_score > sim_score)
			{
//				offset_top_score_global = i - overhang;
				*offset_top_score = i - overhang;
				sim_score = tmp_score;
			}
		}

	}
	return sim_score;
}


double entropy_optiamal_aligned_core;

// NB: the order of the input pssms is important as the  entropy constraint is estimated 
// from the overlapping core of the FIRST pssm and stored in 'entropy_optiamal_aligned_core'
double similarity_score_Harbison_with_overhang_no_background_with_entropy_constraint
(gsl_matrix *predicted_pssm, gsl_matrix *known_pssm, int required_core_length, int *offset_top_score)
{
//	return similarity_score_Harbison_with_overhang_no_background(predicted_pssm, known_pssm, required_core_length, offset_top_score);

	int i;
	gsl_matrix_view g_view_1, g_view_2;

	gsl_matrix *pssm[2];
	int pred_ind = 0, known_ind = 1;
	pssm[pred_ind] = predicted_pssm;
	pssm[known_ind] = known_pssm;

	int shorter_ind, longer_ind;
	if(pssm[pred_ind]->size2 < pssm[known_ind]->size2)
		shorter_ind = pred_ind;
	else
		shorter_ind = known_ind;
	longer_ind = (shorter_ind + 1) % 2;

	double tmp_score, sim_score = 0.0;
	int overhang;

	if(pssm[shorter_ind]->size2 >= required_core_length)
	{
		// shorter motif is at least 6 bases long
		overhang = pssm[shorter_ind]->size2 - required_core_length;
		int num_cols, upper_left_col;
		int num_comparisons = 2 * overhang + pssm[longer_ind]->size2 - pssm[shorter_ind]->size2 + 1;
		int upper_left_row = 0, num_rows = pssm[shorter_ind]->size1;

		*offset_top_score = -overhang;
		for(i=0;i<num_comparisons;i++)
		{
			if(i < overhang)
			{
				//	pssm[longer_ind]: 0 -> pssm2->size2 - (overhang - i)
				//	pssm2: overhang - i -> pssm2->size2
				num_cols = pssm[shorter_ind]->size2 - (overhang - i);

				upper_left_col = 0;
				g_view_1 = gsl_matrix_submatrix(pssm[longer_ind], upper_left_row, upper_left_col, num_rows, num_cols);

				upper_left_col = overhang - i;
				g_view_2 = gsl_matrix_submatrix(pssm[shorter_ind], upper_left_row, upper_left_col, num_rows, num_cols);

			}
			else if(i <= overhang + pssm[longer_ind]->size2 - pssm[shorter_ind]->size2)
			{
				//  pssm[longer_ind]:  i -> pssm[shorter_ind]->size2 + i
				//  pssm[shorter_ind]: 0 -> pssm[shorter_ind]->size2
				num_cols = pssm[shorter_ind]->size2;

				upper_left_col = i - overhang;
				g_view_1 = gsl_matrix_submatrix(pssm[longer_ind], upper_left_row, upper_left_col, num_rows, num_cols);

				upper_left_col = 0;
				g_view_2 = gsl_matrix_submatrix(pssm[shorter_ind], upper_left_row, upper_left_col, num_rows, num_cols);
			}
			else
			{
				//  pssm[longer_ind]: pssm[longer_ind]->size2 - pssm[shorter_ind]->size2 + i -> pssm[shorter_ind]->size2
				//  pssm[longer_ind]: 0 -> pssm[longer_ind]->size2 - i
				num_cols = pssm[longer_ind]->size2 - i + overhang;

				upper_left_col = i - overhang;
			//	printf("%i: num_cols:%i\tupper_left_col: %i\n", i, num_cols, upper_left_col);
				g_view_1 = gsl_matrix_submatrix(pssm[longer_ind], upper_left_row, upper_left_col, num_rows, num_cols);

				upper_left_col = 0;
				g_view_2 = gsl_matrix_submatrix(pssm[shorter_ind], upper_left_row, upper_left_col, num_rows, num_cols);
			}

			tmp_score = similarity_score_Harbison_equal_length(&g_view_1.matrix, &g_view_2.matrix);
			if(tmp_score > sim_score)
			{
				bool entropy_constraint_flag;
				if(pred_ind == longer_ind)
				{
					entropy_constraint_flag = satisfy_entropy_constraint(&g_view_1.matrix);
					double default_eps = -1;
					entropy_optiamal_aligned_core = ave_entropy(&g_view_1.matrix, default_eps);
				}
				else
				{
					double default_eps = -1;
					entropy_constraint_flag = satisfy_entropy_constraint(&g_view_2.matrix);
					entropy_optiamal_aligned_core = ave_entropy(&g_view_2.matrix, default_eps);
				}
				if(entropy_constraint_flag == true)
				{
					*offset_top_score = i - overhang;
					sim_score = tmp_score;
				}
			}
		}
	}
	else
	{
//		overhang = 0;
		int num_comparisons = pssm[longer_ind]->size2 - pssm[shorter_ind]->size2 + 1;
		int upper_left_row = 0, num_rows = pssm[shorter_ind]->size1, num_cols = pssm[shorter_ind]->size2;
		int upper_left_col;
		gsl_matrix_view view;

//		*offset_top_score = -overhang;
		*offset_top_score = 0;
		for(i=0;i<num_comparisons;i++)
		{
			upper_left_col  = i;
			view = gsl_matrix_submatrix(pssm[longer_ind], upper_left_row, upper_left_col, num_rows, num_cols);
			tmp_score = similarity_score_Harbison_equal_length(&view.matrix, pssm[shorter_ind]);
			if(tmp_score > sim_score)
			{	
				bool entropy_constraint_flag;
				if(pred_ind == longer_ind)
				{
					double default_eps = -1;
					entropy_constraint_flag = satisfy_entropy_constraint(&view.matrix);
					entropy_optiamal_aligned_core = ave_entropy(&view.matrix, default_eps);
				}
				else
				{
					double default_eps = -1;
					entropy_constraint_flag = satisfy_entropy_constraint(pssm[shorter_ind]);
					entropy_optiamal_aligned_core = ave_entropy(pssm[shorter_ind], default_eps);
				}
				if(entropy_constraint_flag == true)
				{
					*offset_top_score = i;
					sim_score = tmp_score;
				}
			}
		}
	}
	return sim_score;
}

double eval_PSSM_similarity(gsl_matrix *pssm1, gsl_matrix *pssm2, int required_core_length, int *offset_top_score)
{
	if (pssm1->size1==8)
	{
		char *struct_logo1;
		char *struct_logo2;
		gsl_matrix *colapsed_pssm1 = colapse_pssm(pssm1, &struct_logo1);
		gsl_matrix *colapsed_pssm2 = colapse_pssm(pssm2, &struct_logo2);
		return similarity_score_Harbison_with_overhang_no_background(colapsed_pssm1, colapsed_pssm2, required_core_length, offset_top_score);
	}
	return similarity_score_Harbison_with_overhang_no_background(pssm1, pssm2, required_core_length, offset_top_score);
}

//////////////////////////////////////
// DRIM similarity metric (E. Eden, D. Lipson, S. Yogev & Z. Yakhini. 
// "Discovering Motifs in Ranked Lists of DNA Sequences", PLoS Computational Biology, 2007)
//////////////////////////////////////


double char_similarity_metric_DRIM(char a, char b)
{
	// dummy if on purpose want to have zero distance to ANY letter
	double w = 0.75; // weight for the structure similarity
	if(a == 'X' || b == 'X')
		return 0;
	else
		return 1 - 2 * ( (get_intersection(a,b) + w * get_intersection(get_struct_complement(a),b)) / (get_degeneracy(a) + get_degeneracy(b)) );
}

double score_equal_length_motifs_DRIM(char *m1, char *m2)
{
	int i;
	int length = strlen(m1);
	double sum = 0;
	for(i=0;i<length;i++)
		sum += char_similarity_metric_DRIM(m1[i], m2[i]);
//	printf("score: %lf\n", sum);
	return sum;
}


int best_start_index_m1, best_start_index_m2, best_end_index_m1, best_end_index_m2, best_u;
double distance_metric_DRIM(char *m1, char *m2)
{
	int i, j;

	// make sure len1 >= len2
	if(strlen(m1) < strlen(m2))
	{
		char *tmp = m1;
		m1 = m2;
		m2 = tmp;
	}
	int len1 = strlen(m1), len2 = strlen(m2);

	char augmented_m1[len1 + 2*(len2-1) + 1];
	augmented_m1[len1 + 2*(len2-1)] = '\0';
	for(i=0;i<len2-1;i++)
	{
		augmented_m1[i] = 'X';
		augmented_m1[len1 + 2*(len2-1) - 1 - i] = 'X';
	}
	for(i=len2-1;i<len1+len2-1;i++)
		augmented_m1[i] = m1[i-len2 + 1];

//	printf("%s\n%s\n%s\n", m1, m2, augmented_m1);

	// weight put on unpaired bases
	double w_1 = 0.6;
	// number of unpaired bases contributed by shorter motif
	int u;

	// number of unpaired bases contributed by longer motif
	int u_longer;

	double distance = 100;

	int start_index_m1, start_index_m2, end_index_m1, end_index_m2;
	int num_iter = len1 + len2 - 1;
	for(i=0;i<num_iter;i++)
	{
		if(i < len2)
		{
			u = (len2 - 1) - i;
			start_index_m1 = 0;
			end_index_m1 = len2 - u;
			start_index_m2 = u;
			end_index_m2 = len2;
		}
		else if(i < len1)
		{
			u = 0;
			start_index_m1 = i - len2 + 1;
			end_index_m1 = start_index_m1 + len2 - u;
			start_index_m2 = 0;
			end_index_m2 = len2;
		}
		else
		{
			u = i-len1 + 1;
			start_index_m1 = i - len2 + 1;
			end_index_m1 = start_index_m1 + len2 - u;
			start_index_m2 = 0;
			end_index_m2 = len2 - u;
		}
		u_longer = len1 - (len2 - u);

		char tmp1[50]; char tmp2[50];
		for(j=start_index_m2;j<end_index_m2;j++)
			tmp2[j-start_index_m2] = m2[j];
		tmp2[j-start_index_m2] = '\0';

		for(j=start_index_m1;j<end_index_m1;j++)
			tmp1[j-start_index_m1] = m1[j];
		tmp1[j-start_index_m1] = '\0';

		double curr = score_equal_length_motifs_DRIM(tmp1, tmp2) + w_1*(u + u_longer);
		curr = curr / ( 0.5 * (strlen(m1) + strlen(m2)) );
		if(distance > curr)
		{
			distance = curr;
			best_start_index_m1 = start_index_m1;
			best_end_index_m1 = end_index_m1;
			best_start_index_m2 = start_index_m2;
			best_end_index_m2 = end_index_m2;
			best_u = u;
		}
	}
/*
	printf("best distance: %lf\n", distance);
	for(j=best_start_index_m1;j<best_end_index_m1;j++)
		printf("%c", m1[j]);
	printf("\n");
	for(j=best_start_index_m2;j<best_end_index_m2;j++)
		printf("%c", m2[j]);
	printf("\n");
*/
	return distance;
}

bool rev_DRIM;
double eval_distance_DRIM(char *m1, char *m2)
{
	double distance1 = distance_metric_DRIM(m1, m2);
	char *rev = rev_compl(m1);
	double distance2 = distance_metric_DRIM(rev, m2);
	FREE(rev);
	double distance;

	if(distance2 < distance1)
	{
//		printf("rev\n");
		rev_DRIM = true;
		distance = distance2;
	}
	else
	{
		distance = distance1;
		rev_DRIM = false;
	}
	return distance;
}

void test_eval_distance_DRIM()
{
	double d1, d2;

	char *m1 = "CCACAG";
	char *m2 = "AAACTGTGG";
//	char *m1 = "AAAAGCCGCGGGCGGGATT";
//	char *m2 = "CARAAR";
	double d = eval_distance_DRIM(m1, m2);

	printf("distance: %lf\n", d);
	getchar();
}

