/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#include "tree.h"

// allocate memory for an array of arbitrary type. If array already
// allocated FREE old array and allocate memory for new array
void initialize_array(void **array, int element_size, int length_array, bool *array_initialized)
{
	if(length_array > 0)
	{
		if(*array_initialized == true)
		{
			FREE(*array);
		}
		else
		{
			*array_initialized = true;
		}
		MALLOC(*array, element_size * length_array);
	}
	else
	{
		printf("ERROR: 'initialize_array', 'length_array' needs to be positive! %i\n", length_array);
		exit(1);
	}
}

int seq_length;
bool seq_length_initialized = false;
void init_seq_length(int sequence_length)
{
	if(seq_length_initialized == false)
	{
		if(sequence_length > 0)
		{
			seq_length = sequence_length;
			seq_length_initialized = true;
		}
		else
		{
			printf("ERROR: 'init_seq_length', 'seq_length' cannot be non-positive! %i", sequence_length);
			exit(1);
		}
	}
}

///////////  1  ///////////

// needs to be allocated from outside function before 'action_initialize_reference' is called
int *reference_positions = NULL;

bool initialize_reference_called = false;
int length_reference_positions;
int length_reference_positions_ = 0;

void action_initialize_reference(const void *nodep, VISIT value, int level)
{
	int *node_value = *((int **)nodep);
	if(value == leaf || value == postorder)
	{
		if(length_reference_positions_ > length_reference_positions)
		{
			printf("ERROR: tree.c\n'action_initialize_reference': more etries were attempted to be stored in 'reference_positions' than pre-allocated memory allows! %i\n", length_reference_positions_);
			exit(1);
		}
		reference_positions[length_reference_positions_] = *node_value;
		length_reference_positions_++;
//		printf("%i\t%i\n", reference_positions[count_-1], *node_value);
//		getchar();
	}
}

// input:
// (1) 'root' -> pointer to root of the binary tree (i.e. double pointer)
// (2) 'n'    -> number of entries in 'reference_positions', acting as storage for the SORTED output of locations
//               the sorting is an artefact the post-order traversal of the binary tree
// output:
// (1) 'reference_positions' is initialized with the values in the binary tree rooted at '**root'
// (2) the length of the 'reference_positions' array is stored in 'length_reference_positions'
void tree_walk_initialize_reference(const void *root, int n, int sequence_length)
{
	init_seq_length(sequence_length);
	initialize_array(&reference_positions, sizeof(int), n, &initialize_reference_called);
//	MALLOC(reference_positions, sizeof(int) * n);

	length_reference_positions_ = 0; length_reference_positions = n;

	twalk(root, action_initialize_reference);

	length_reference_positions_ = 0;
	initialize_reference_called = true;
}

int get_length_reference_positions()
{
	if(initialize_reference_called == true)
		return length_reference_positions;
	else
	{
		printf("ERROR: tree.c\n'tree_walk_initialize_reference' needs to be called to define 'max_ref_index'!\n");
		getchar();
	}
	return -1;
}

int *get_reference_positions()
{
	if(initialize_reference_called == true)
		return reference_positions;
	else
	{
		printf("ERROR: tree.c\n'tree_walk_initialize_reference' needs to be called before 'get_reference_positions'!\n");
		getchar();
	}
	return NULL;
}

void deallocate_reference_positions()
{
	FREE(reference_positions);
	initialize_reference_called = false;
	length_reference_positions = 0;
}

//////////// 2 /////////////

int lower_offset_from_reference = 0;
int upper_offset_from_reference = 0;
void set_offset_from_reference(int lower, int upper)
{
	//if(d < 0)
	//	printf("ERRROR: tree.c\n'offset_from_reference' cannot be negative %i\n", d);
	lower_offset_from_reference = lower;//MAX(0, d);
	upper_offset_from_reference = upper;
}
/*
inline bool overlap(int ref_pos, int query_pos)
{

	return (query_pos == (ref_pos + offset_from_reference));
	//	(query_pos >= ref_pos - offset_from_reference && query_pos <= ref_pos + offset_from_reference);
	//	(fabs(pos1 - pos2) <= offset_from_reference);
}
*/

inline bool within_offset(int ref_pos, int query_pos)
{
//	int lower = query_pos + lower_offset_from_reference;
//	int upper = query_pos + upper_offset_from_reference;
//	return (ref_pos >= lower) && (ref_pos <= upper);
	return (ref_pos >= query_pos - upper_offset_from_reference) && (ref_pos <= query_pos - lower_offset_from_reference);
}

int last_gene;
inline bool is_already_counted(int location)
{
	return (last_gene == (location / seq_length));
}

int ind;
int num_overlaps;
int num_overlaps_;
int num_gene_overlaps;
int num_gene_overlaps_;

void **seq_index;

void action_find_overlap_to_reference(const void* nodep, VISIT value, int level)
{
 	int *node_value = *((int **)nodep);
	if(value == leaf || value == postorder)
	{
		while(ind < length_reference_positions && reference_positions[ind] < *node_value - upper_offset_from_reference )
		{
	//		printf("in %i\t(%i) %i/%i: %i\t%i\n", seq_length, (*node_value) / seq_length, ind, length_reference_positions, reference_positions[ind], *node_value);
	//		getchar();
			ind++;
		}
	//	printf("(%i) %i/%i: %i\t%i\n", *node_value / seq_length, ind, length_reference_positions, reference_positions[ind], *node_value);
	//	getchar();

		if(ind < length_reference_positions && (within_offset(reference_positions[ind], *node_value) == true))
		{
	//		printf("within_offset:\n%i\n%i", reference_positions[ind], *node_value);
	//		getchar();
			num_overlaps_++;
			num_overlaps = num_overlaps_;
			if(is_already_counted(*node_value) == false)
			{
				num_gene_overlaps_++;
				num_gene_overlaps = num_gene_overlaps_;
				last_gene = *node_value / seq_length;
			}
		}
/*		do
		{
			if(ind < length_reference_positions && (within_offset(reference_positions[ind], *node_value) == true)
			{
				num_overlaps_++;
				num_overlaps = num_overlaps_;
				if(is_already_counted(*node_value) == false)
				{
					num_gene_overlaps_++;
					num_gene_overlaps = num_gene_overlaps_;
					last_gene = *node_value / seq_length;
				}
			}
			else
				ind++;
		}
		while(*node_value < reference_positions[ind] + upper_offset_from_reference);
	//	if(ind >= length_reference_positions || (within_offset(reference_positions[ind], *node_value) == false))
		printf("%i\t%i\t%i\n", reference_positions[ind], *node_value, *node_value / seq_length);
		getchar();

		if(ind < length_reference_positions && (within_offset(reference_positions[ind], *node_value) == true))
		{
			num_overlaps_++;
			num_overlaps = num_overlaps_;
			if(is_already_counted(*node_value) == false)
			{
				num_gene_overlaps_++;
				num_gene_overlaps = num_gene_overlaps_;
				last_gene = *node_value / seq_length;
			}
		//	while(ind < length_reference_positions && reference_positions[ind] < *node_value + lower_offset_from_reference)
		//		ind++;
/*
			uchar **snippet;
			MALLOC(snippet, sizeof(uchar *));
			ulong snippet_length;
			int res = extract(*seq_index, *node_value, *node_value+10, snippet, &snippet_length);
*/
	//		printf("snippet: %s\t%i\n", *snippet, res);
	//		printf("*node_value: %i\tlast_gene: %i\tnum_overlaps_: %i\tnum_gene_overlaps: %i\n",
	//					*node_value, last_gene, num_overlaps_, num_gene_overlaps);
	//		getchar();
	//	}
	}
}


bool overlap_to_reference_called = false;
// input:
// (1) 'root' -> pointer to root of the binary tree (i.e. double pointer)
// (2) 'offset_from_reference'-> overlap search offset with respect to the 'reference_positions' entries
// output:
// (1) number of genes where the two motifs have properly offset overlapping occurrences is recorded in 'num_overlaps'
void tree_walk_find_overlap_to_reference(const void *root, int lower_offset_, int upper_offset_, void **index, int length_shorter_motif)
{
	if(initialize_reference_called == true)
	{
		seq_index = index;

		set_offset_from_reference(lower_offset_, upper_offset_);
	//	printf("lower_offset_from_reference: %i\nupper_offset_from_reference:%i\n", lower_offset_from_reference, upper_offset_from_reference);
		ind = 0; last_gene = -1; num_overlaps_ = 0; num_overlaps = 0; num_gene_overlaps_ = 0; num_gene_overlaps = 0;

		twalk(root, action_find_overlap_to_reference);

		ind = 0; last_gene = -1; num_overlaps_ = 0; num_gene_overlaps_ = 0;
		overlap_to_reference_called = true;
	}
	else
	{
		overlap_to_reference_called = false;
		printf("ERROR: tree.c\n'action_initialize_reference' needs to be called before 'tree_walk_find_overlap_to_reference'!\n");
		getchar();
	}
}

int get_num_offset_specific_gene_overlaps()
{
	if(initialize_reference_called == true)
		return num_gene_overlaps;
	else
	{
		printf("ERROR: 'get_num_offset_specific_gene_overlaps', need to call 'tree_walk_find_overlap_to_reference'!\n");
		exit(1);
	}
}

int get_total_num_overlaps()
{
	if(overlap_to_reference_called == true)
		return num_overlaps;
	else
	{
		printf("ERROR: tree.c\n'tree_walk_find_overlap_to_reference' needs to be called to define 'num_matches'!\n");
		getchar();
	}
	return -1;
}

/////////////  3  /////////////

int gene_count_;
int gene_count;
int last_index;
// must to have 'seq_length' initialized prior to calling this function
void action_find_gene_occurrences_count(const void* nodep, VISIT value, int level)
{
 	int *node_value = *((int **)nodep);
	if(value == leaf || value == postorder)
	{
		if(*node_value / seq_length != last_index)
		{
			last_index = *node_value / seq_length;
			gene_count_++;
			gene_count = gene_count_;
		}
	}
}

bool gene_count_done = false;
void tree_walk_count_gene_occurrences(const void *root, void **index, int sequence_length)
{
	init_seq_length(sequence_length);
	gene_count_ = 0; gene_count = 0; last_index = 0;

	twalk(root, action_find_gene_occurrences_count);

	gene_count_ = 0; last_index = 0;
	gene_count_done = true;
}

int get_gene_count()
{
	if(gene_count_done == true)
		return gene_count;
	else
	{
		printf("ERROR: 'get_gene_count', need to call 'tree_walk_count_gene_occurrences'!\n");
		exit(1);
	}
}

//////////  4  ///////////

bool genes_initialized = false;
int *genes = NULL;
int length_genes;
int last_gene_index;
// must to have 'seq_length' initialized prior to calling this function
void action_find_gene_occurrences(const void* nodep, VISIT value, int level)
{
 	int *node_value = *((int **)nodep);
	if(value == leaf || value == postorder)
	{
		if(*node_value / seq_length != last_gene_index)
		{
			last_gene_index = *node_value / seq_length;
			genes[length_genes] = last_gene_index;
			length_genes++;
		}
	}
}

bool genes_array_initialized = false;
void tree_walk_find_gene_occurrences(const void *root, void **index, int sequence_length, int num_genes)
{
	initialize_array(&genes, sizeof(int), num_genes, &genes_initialized);
	init_seq_length(sequence_length);
	length_genes = 0; last_gene_index = 0;

	twalk(root, action_find_gene_occurrences);

	genes_array_initialized = true;
}

int *get_genes()
{
	if(genes_initialized == true)
		return genes;
	else
	{
		printf("ERROR: 'get_genes', need to call 'tree_walk_find_gene_occurrences'!\n");
		exit(1);
	}
}

int get_length_genes()
{
	return length_genes;
}


//////////  5  //////////

int compar_int(const void *s1 ,const void *s2 )
{
	if(*((int *)s1) > *((int *)s2)) return 1;
	else if(*((int *)s1) == *((int *)s2)) return 0;
	else return -1;
}

bool find_element(int element, void **root)
{
	int **ptr = (int **)tfind(&element, root, compar_int);
	return (ptr != NULL);
}

void freefct(){}

void destroy_tree(void **root)
{
	tdestroy(*root, freefct);
}

//////////////////////
//////////////////////


void action_fn(const void* nodep, VISIT value, int level)
{
	int *node_value = *((int **)nodep);
	if(value == leaf || value == postorder)
	{
		printf("%i, level %i\n", *node_value, level);
		getchar();
	}
/*
	switch(value)
	{
		case leaf:
			printf("leaf, ");
			printf("%i, level %i\n", *node_value, level); break;
		case preorder:
			printf("preorder, ");
			printf("%i, level %i\n", *node_value, level); break;
		case postorder:
			printf("postorder, ");
			printf("%i, level %i\n", *node_value, level); break;
		case endorder:
			printf("endorder, ");
			printf("%i, level %i\n", *node_value, level); break;
		default:
			printf("default\n");
	}
	*/
}

void tree_walk_in_sorted_order(const void *root)
{
	twalk(root, action_fn);
}




void test_tree()
{
	int i; int num = 5;
	void *ptr[num];
	for(i=0;i<num;i++)
	{
		ptr[i] = NULL;
		int a = 5, b = 10, c = 7, d = 15;
		printf("%i\n", i);
//		void *rootp = NULL;
		if(ptr[i] == NULL)
			printf("%i\tNULL\n", i);
		else
			printf("%i\tnot\n", i);
		tsearch(&a, &ptr[i], compar_int);
		printf("here\n", i);
		tsearch(&b, &ptr[i], compar_int);
		tsearch(&c, &ptr[i], compar_int);
		tsearch(&d, &ptr[i], compar_int);
		int to_find = 10;
		if(find_element(to_find, &ptr[i]) == true)
			printf("found\n");
		else
			printf("not found\n");
		twalk(ptr[i], action_fn);
		destroy_tree(&ptr[i]);
		getchar();
	}
	printf("done\n");
	getchar();
}

