/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#include "utilities.h"


gsl_matrix *diag(gsl_vector *X)
{
	gsl_matrix *mat = gsl_matrix_alloc(X->size, X->size);
	gsl_vector_view diag = gsl_matrix_diagonal(mat);
	gsl_matrix_set_all(mat, 0.0);
	gsl_vector_memcpy(&diag.vector, X);
	return mat;
}

gsl_matrix *diag_inv(gsl_vector *X)
{
	gsl_matrix *mat = gsl_matrix_alloc(X->size, X->size);
	gsl_vector_view diag = gsl_matrix_diagonal(mat);
	gsl_matrix_set_all(mat, 0.0); 
	gsl_vector_set_all(&diag.vector, 1.0);
	gsl_vector_div(&diag.vector, X);
	return mat;
}



void **create_index(char *text_filename, int *seq_length, int num_sequences_to_read)
{
//	printf("text_filename: %s\n", text_filename);
	char delim = '$';

	int num_seq, seq_len;
//	bool ignore_strand = true;
	get_seq_info(text_filename, delim, &num_seq, &seq_len, true);//ignore_strand);
	num_sequences_to_read = num_seq;

	*seq_length = get_seq_length(text_filename, delim);
//	printf("*seq_length: %i\n", *seq_length);
//	printf("seq_len: %i\n", seq_len);
//	getchar();
	if(access(text_filename, F_OK) == -1)
  	{
   		printf("'create_index': could not open file: %s\n", text_filename);
    	exit(2);
  	}
//	printf("num_sequences_to_read: %i\n", num_sequences_to_read);
  	FILE *fp = open_file(text_filename, "r");
  	unsigned char *text;
  	ulong text_length = 0;

  	if(num_sequences_to_read != -1)
  	{
   		text_length = sizeof(unsigned char) * num_sequences_to_read * (*seq_length) + 1;
	  	MALLOC(text, sizeof(unsigned char) * num_sequences_to_read * (*seq_length) + 1);
  		if(fgets(text, num_sequences_to_read * (*seq_length) + 1, fp) == NULL)
  		{
  			printf("error reading %i number of sequences with given gene level statistics!\n");
  			exit(1);
  		}
  	}
  	else
  	{
  		printf("ERROR: 'create_index', num_sequences_to_read == -1\n");
		getchar();
   	  	getline(&text, &text_length, fp);
   	  	text = 0;
  	}
  	fclose(fp);
//	printf("BEFORE build_index\n");
	// read in text
//  	FILE *fp = open_file(text_filename, "r");
//  	unsigned char *text = 0;
//  	ulong text_length = 0;

//  	getline(&text, &text_length, fp);
//  	fclose(fp);

  	char *build_options = "copy_text;free_text";
  	void **index = malloc(sizeof(TSA_Un_copy *));
  	int e;
  	if((e = build_index(text, strlen(text), build_options, index)) != 0)
  	{
		printf("could not build index: %s\n", error_index(e));
		exit(1);
  	}
// 	printf("index built\n");
  	return index;
}


void destroy_index(void ***index)
{
  	int e;
  	if((e = free_index(**index)) != 0)
  	{
    	printf("could not free index: %s\n", error_index(e));
    	exit(1);
  	}
   	FREE(*index);
}


char *extract_subsequence(void **index, int from, int to, int *length_of_extracted_seq)
{
	char **snippet;
	MALLOC(snippet, sizeof(char *));
	if(extract(*index, from, to, snippet, length_of_extracted_seq) != 0)
	{
		printf("ERROR: 'extract_subsequence' unable to extract sequence!\n");
		exit(1);
	}
	return *snippet;
}

void test_extract_subsequence()
{
	int length_of_extracted_seq;
	char *snippet = extract_subsequence(index, 0, 20, &length_of_extracted_seq);
	printf("%s", snippet);
	getchar();
}


///////////////////////////////////
///////////////////////////////////

gsl_matrix *MAT_times_MAT(gsl_matrix *A, gsl_matrix *B)
{
	if(A->size2 == B->size1)
	{
		gsl_matrix *res = gsl_matrix_alloc(A->size1, A->size1);

		//Compute res = A B
		gsl_blas_dgemm(CblasNoTrans, CblasNoTrans,
		                1.0, A, B,
		                0.0, res);
		return res;
	}
	else
	{
		printf("'MAT_times_MAT': incompatible matrices to multiply!%i\t%i\n", A->size2, B->size1);
		exit(1);
	}
}

gsl_matrix *transpose_MAT(gsl_matrix *m)
{
	gsl_matrix *dest = gsl_matrix_alloc(m->size1, m->size2);
	gsl_matrix_transpose_memcpy(dest, m);
	return dest;
}

gsl_vector *MAT_times_VEC(gsl_matrix *m, gsl_vector *x)
{
//	CBLAS_TRANSPOSE_t TransA = CblasNoTrans;
	double alpha = 1.0, beta = 0.0;
	gsl_vector *y = gsl_vector_alloc(x->size);
	int res = gsl_blas_dgemv(CblasNoTrans, alpha, m, x, beta, y);
	return y;
}

gsl_vector *VEC_times_MAT(gsl_vector *x, gsl_matrix *m)
{
	gsl_matrix *m_transpose = transpose_MAT(m);
	gsl_vector *res = MAT_times_VEC(m_transpose, x);
	gsl_matrix_free(m_transpose);
	return res;
}

gsl_vector *lower_triangular_MAT_times_VEC(gsl_matrix *m, gsl_vector *x)
{
	CBLAS_TRANSPOSE_t TransA = CblasNoTrans;
	CBLAS_UPLO_t Uplo = CblasLower;
	CBLAS_DIAG_t Diag = CblasNonUnit;
	gsl_vector *y = gsl_vector_alloc(x->size);
	gsl_vector_memcpy(y, x);
	int res = gsl_blas_dtrmv (Uplo, TransA, Diag, m, y);
	return y;
}

void test_MAT_times_VEC()
{
	int i,j;
   	int rows = 2, cols = 2;
	gsl_matrix *corr_mat = gsl_matrix_alloc(rows, cols);

	gsl_matrix_set(corr_mat, 0, 0, 1);
	gsl_matrix_set(corr_mat, 0, 1, 2);
	gsl_matrix_set(corr_mat, 1, 0, 3);
	gsl_matrix_set(corr_mat, 1, 1, 1);

//	gsl_matrix_fprintf(stdout, corr_mat, "%.5g\t");

 	for(i=0;i<2;i++)
   	{
   	   	for(j=0;j<2;j++)
   	   	{
   	   		double a = gsl_matrix_get(corr_mat, i, j);
   	   		printf("%lf ", a);
   	   	}
   	   	printf("\n");
   	}

 	printf("\n");

	gsl_vector *x = gsl_vector_alloc(rows);
	gsl_vector_set(x, 0, -1);
	gsl_vector_set(x, 1, 2);

	gsl_vector *y = lower_triangular_MAT_times_VEC(corr_mat, x);
//	gsl_vector *y = MAT_times_VEC(corr_mat, x);
	printf("%lf\n", gsl_vector_get(x, 0));
	printf("%lf\n\n", gsl_vector_get(x, 1));

	printf("%lf\n", gsl_vector_get(y, 0));
	printf("%lf\n", gsl_vector_get(y, 1));
	getchar();
}

void example_GSL_matrix_multiplication()
{
   double a[] = { 0.11, 0.12, 0.13,
                   0.21, 0.22, 0.23 };

    double b[] = { 1011, 1012,
                   1021, 1022,
                   1031, 1032 };

    double c[] = { 0.00, 0.00,
                   0.00, 0.00 };

    gsl_matrix_view A = gsl_matrix_view_array(a, 2, 3);
    gsl_matrix_view B = gsl_matrix_view_array(b, 3, 2);
    gsl_matrix_view C = gsl_matrix_view_array(c, 2, 2);

    /* Compute C = A B */

    gsl_blas_dgemm (CblasNoTrans, CblasNoTrans,
                    1.0, &A.matrix, &B.matrix,
                    0.0, &C.matrix);

    printf ("[ %g, %g\n", c[0], c[1]);
    printf ("  %g, %g ]\n", c[2], c[3]);

    getchar();
}


void test_matrix_op()
{
	time_t timer1 = start_timer();;
	int i,j;
	int dim_matrix = 10;
	printf("dim_matrix: %i\n", dim_matrix);
	int num_uniform1 = dim_matrix * dim_matrix;
	double lower1 = -1, upper1 = 1;
	double *dest_uniform1;
	MALLOC(dest_uniform1, sizeof(double)*num_uniform1);
	ran_uniform(dest_uniform1,num_uniform1,lower1,upper1);

	// create a 'matrix' view of the double[]
	gsl_matrix_view view = gsl_matrix_view_array(dest_uniform1,dim_matrix,dim_matrix);
	gsl_matrix *random_mat = &view.matrix;
	gsl_matrix *res_mat = gsl_matrix_alloc(dim_matrix,dim_matrix);

	print_time(timer1, stdout);
	CBLAS_TRANSPOSE_t TransA = CblasNoTrans;
	CBLAS_TRANSPOSE_t TransB = CblasTrans;
	double alpha = 1,beta = 0;

	// tr(X)X
	gsl_blas_dgemm(TransA,TransB,alpha,random_mat,random_mat,beta,res_mat);
	print_time(timer1, stdout);

	for(i=0;i<dim_matrix;i++)
	{
    	double val = gsl_matrix_get(res_mat,i,i);
    	gsl_matrix_set(res_mat,i,i,val+1);
	}

	int rows, cols;
	rows = cols = dim_matrix;
	for(i=0;i<rows;i++)
	{
		for(j=0;j<cols;j++)
	    {
	       	double res = gsl_matrix_get(res_mat,i,j);
	     	printf("%3.2lf ", res);
	    }
		printf("\n");
	}
//	print_eigenvec_and_eigenvals(res_mat);

    int res = gsl_linalg_cholesky_decomp(res_mat);
   	if(res == GSL_EDOM)
   	{
   		printf("ERROR: 'motif_finder.c', matrix is not positive definite\n");
   		exit(1);
   	}
	for(i=0;i<rows;i++)
	{
		for(j=0;j<cols;j++)
	    {
	       	double res = gsl_matrix_get(res_mat,i,j);
	     	printf("%3.2lf ", res);
	    }
		printf("\n");
	}
//   	print_eigenvec_and_eigenvals(res_mat);
	FREE(dest_uniform1);
	print_time(timer1, stdout);
	printf("done\n");
	getchar();
}


///////////////////////////////////
///////////////////////////////////


double p_value_mean_gaussin_fit(const double data[], int num, double observed_mean)
{
	int stride = 1;
	double hist_mean    = gsl_stats_mean(data, stride, num);
	double hist_sd      = gsl_stats_sd_m(data, stride, num, hist_mean);
	double hist_Z_score = (observed_mean - hist_mean)/hist_sd;
	double hist_gaussian_fit_p_value = gsl_cdf_ugaussian_Q(hist_Z_score);
	return hist_gaussian_fit_p_value;
}

double p_value_permutations(const double data[], int num_data, int *index, int num_geneset, double observed_mean, int num_ran_draws)
{
	int j,k;
//	unsigned long *dest;
	unsigned long num_dest = num_geneset;
//	MALLOC(dest, sizeof(unsigned long) * num_dest);
	unsigned long dest[num_dest];
	unsigned long num_src = num_data;
	unsigned long *src = index;

	int count_extreme = 0;
	for(j=0;j<num_ran_draws;j++)
	{
		sample_without_replacement(dest, num_dest, src, num_src);
		double perm_mean = 0;
		for(k=0;k<num_dest;k++)
			perm_mean += data[dest[k]];
		perm_mean /= num_dest;
		if(perm_mean > observed_mean)
			count_extreme++;
	}
	double perm_p_value = (1.0*count_extreme)/num_ran_draws;
	return perm_p_value;
}

/*
void test_normal_approx()
{
	int i,j,k;
//	char *gene_level_scores_file = "/home/sgg4/Desktop/main_project/data/yeast/input_156/inputfile_ABF1_YPD";
//	char *gene_level_scores_file = "/home/sgg4/Desktop/main_project/data/yeast/input_156/inputfile_DAL81_RAPA";
	char *gene_level_scores_file = "/home/sgg4/Desktop/main_project/data/yeast/input_156/inputfile_SUT1_YPD";

	gene_level_scores s; 
	load_gene_level_scores(&s, gene_level_scores_file);
	for(i=0;i<s.num;i++)
		s.scores[i] = -log(s.scores[i]);

//	FILE *fp = open_file("/home/sgg4/workspace_64bit/GMF_05_10_2008/to_plot", "w");

	// number of the genes in the geneset
	unsigned long num_dest = 10; unsigned long dest[num_dest];
	unsigned long num_src = s.num; unsigned long *src = s.index;
	unsigned long num_dest_random_set = num_dest, dest_random_set[num_dest_random_set];
	unsigned long num_src_random_set = num_src, *src_random_set = s.index;

	double mse_hist_gaussian_fit_p_value = 0, mse_p_value = 0;

	int num_random_motifs = 200, num_ran_draws = 2000;
	for(i=0;i<num_random_motifs;i++)
	{
		if(i % 50 == 0)
			printf("%i\n", i);
		sample_without_replacement(dest_random_set, num_dest_random_set, src_random_set, num_src_random_set);
		int *index_geneset = dest_random_set;int geneset_size = num_dest_random_set;
		double *all_gene_level_stat = s.scores;int num_all_genes = s.num;
		double p_value = random_set_p_value(index_geneset, geneset_size, all_gene_level_stat, num_all_genes);

		double obs_mean = 0;
		for(j=0;j<geneset_size;j++)
			obs_mean += all_gene_level_stat[index_geneset[j]];
		obs_mean /= geneset_size;
//		printf("obs_mean: %lf\n", obs_mean);

		double ran_obs[num_ran_draws];
		for(j=0;j<num_ran_draws;j++)
		{
			sample_without_replacement(dest, num_dest, src, num_src);
			double perm_mean = 0;
			for(k=0;k<num_dest;k++)
				perm_mean += all_gene_level_stat[dest[k]];
			perm_mean /= num_dest;
			ran_obs[j] = perm_mean;
		}
		double hist_gaussian_fit_p_value = p_value_mean_gaussin_fit(ran_obs, num_ran_draws, obs_mean);
		double perm_p_value = p_value_permutations(s.scores, s.num, s.index, geneset_size, obs_mean, num_ran_draws);

		mse_p_value += fabs(perm_p_value - p_value);
		mse_hist_gaussian_fit_p_value += fabs(perm_p_value - hist_gaussian_fit_p_value);

		printf("%i: %lf\t%lf\t%lf\n", i, perm_p_value, hist_gaussian_fit_p_value, p_value);
//		fprintf(fp, "%lf\n", z_score_random_set);
//		getchar();
	}
//	fclose(fp);
	free_gene_level_scores(&s);

	mse_p_value /= num_random_motifs;
	mse_hist_gaussian_fit_p_value /= num_random_motifs;

	printf("mse_p_value: %lf\nmse_hist_gaussian_fit_p_value: %lf\n", mse_p_value, mse_hist_gaussian_fit_p_value);
}

*/

///////////////////////////////////
///////////////////////////////////

int count_num_lines(char *file)
{
	if(access(file, F_OK) == -1)
  	{
   		printf("'get_seq_info': could not open file: %s\n", file);
    	exit(2);
  	}
	FILE *fp = fopen(file, "r");
    char *line = NULL;
    size_t len = 0; ssize_t read;
    int count = 0; char delim = '\n';
	while ((read = getdelim(&line, &len, delim, fp)) != -1)
	{
		count++;
		FREE(line);
	}
	FREE(line);
  	fclose(fp);
  	return count;
}

int count_num_lines_with_double_entry(char *file)
{
	if(access(file, F_OK) == -1)
  	{
   		printf("'get_seq_info': could not open file: %s\n", file);
    	exit(2);
  	}
	FILE *fp = fopen(file, "r");
    char *line = NULL;
    size_t len = 0; ssize_t read;
    int count = 0; char delim = '\n';double dummy;
	while ((read = getdelim(&line, &len, delim, fp)) != -1)
	{
		if(sscanf(line, "%lf", &dummy) == 1)
			count++;
		FREE(line);
	}
	FREE(line);
  	fclose(fp);
  	return count;
}

#include <stdlib.h>
#include <string.h>

double_arr read_array_from_file(char *file)
{
	int total_num_lines = count_num_lines(file);
	int num_lines_with_double_entry = count_num_lines_with_double_entry(file);
	if(total_num_lines != num_lines_with_double_entry)
	{
		printf("potential ERROR: num_lines_with_double_entry(%i) != total_num_lines(%i)\n", num_lines_with_double_entry, total_num_lines);
	}
	FILE *fp = open_file(file, "r");
	char *raw_input;
	double_arr a;
	MALLOC(a.array, sizeof(double)*num_lines_with_double_entry);
	a.num = num_lines_with_double_entry; int curr_index = 0;
	while(fscanf(fp, "%as", &raw_input) != EOF)
	{
		if(sscanf(raw_input, "%lf", &a.array[curr_index]) == 1)
		{
//			printf("%lf\n", a.array[curr_index]);
			curr_index++;
		}
	}
	return a;
}

void test_read_array_from_file()
{
	char *file = "/home/sgg4/workspace_64bit/GMF_05_10_2008/test";
	double_arr arr = read_array_from_file(file);
}

///////////////////////////////////
///////////////////////////////////

char *extract_TF_name(char *seq_file_path, params *p)
{
	return p->TF_name;
	int i;
	// basename(char *string) returns a pointer to the existing string, doesn't allocate memory!
	char *base_name = basename(seq_file_path);
	char TF_tmp[20];

	bool flag = false;
	int ind = 0, len = strlen(base_name);
	for(i=0;i<len;i++)
	{
		if(flag == true)
		{
			if(base_name[i]== '.')
			{
				TF_tmp[ind] = '\0';
				break;
			}
			else
				TF_tmp[ind++] = base_name[i];
		}
		else
		{
			if(base_name[i] == '_')
				flag = true;
		}
	}
	char *TF_name;
	MALLOC( TF_name, sizeof(char) * (strlen(TF_tmp)+1) );
	strcpy(TF_name, TF_tmp);
	return TF_name;
}

/////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////



/////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////

extern species species_type_global;
extern bool is_regression_initialized;
//extern char *covariates_file;
extern bool consider_rev_compl_flag;
extern int num_PCs;

void output_param_settings(char *summary_file_name, params *p, IO_params *p_IO)
{
//	p->scoring = Random_Set_no_ranks;

	FILE *fp = open_file(summary_file_name, "w");
	fprintf(fp,"==============\n");
	fprintf(fp,"I/O parameters\n");
	fprintf(fp,"==============\n");
	fprintf(fp,"input_seeds_file: %s\n", p_IO->input_seeds_file);
	fprintf(fp,"score_file: %s\n", p_IO->score_file);
	fprintf(fp,"seq_file: %s\n", p_IO->seq_file);
	fprintf(fp,"output_dir: %s\n", p_IO->output_dir);
	fprintf(fp,"\n");

	fprintf(fp,"==========================\n");
	fprintf(fp,"geneset scoring parameters\n");
	fprintf(fp,"==========================\n");
/*
	if(p->geneset_ES == geneset_random_set_exact_null_ES && p->scoring_params.pre_calculate_location_and_scale == 0)
		fprintf(fp,"geneset_random_set_exact_null_ES\n");
	else if(p->geneset_ES == geneset_random_set_empirical_null_ES && p->scoring_params.pre_calculate_location_and_scale == estimate_global_MEAN_and_VAR_random_set)
		fprintf(fp,"geneset_random_set_empirical_null_ES\n");
	else if(p->geneset_ES == geneset_ks_ES && p->scoring_params.pre_calculate_location_and_scale == 0)
		fprintf(fp,"geneset_ks_ES + exact null\n");
	else if(p->geneset_ES == geneset_ks_ES && p->scoring_params.pre_calculate_location_and_scale == estimate_global_MEAN_and_VAR_KS_enrichment_score)
		fprintf(fp,"geneset_ks_ES + empirical null\n");
	else if(p->geneset_ES == geneset_adaptive_to_noise_level_ES)
		fprintf(fp,"geneset_adaptive_to_noise_level_ES\n");
	else if(p->geneset_ES == geneset_stable_ES)
		fprintf(fp,"geneset_stable_ES\n");
	else if(p->geneset_ES == geneset_bootstrap_ES)
		fprintf(fp,"geneset_bootstrap_ES\n");
	else
		fprintf(fp,"p->geneset_ES: unkown type\n");
*/
	if(p->fast_mode == true)
		fprintf(fp,"fast_mode: yes\n");
	else
		fprintf(fp,"fast_mode: no\n");

	if(is_regression_initialized == true)
	{
		fprintf(fp,"use regression scoring: yes\n");
		fprintf(fp,"covariates file: %s\n", p_IO->cov_file);
//		fprintf(fp,"num_PCs = %i\n", num_PCs);		
	}
	else
		fprintf(fp,"using regression scoring: no\n");

	if(p->scoring_params.bootstrap_motif_score == true)
	{
		fprintf(fp,"bootstrap_motif_score: yes\n");
		fprintf(fp,"num_bootstrap_draws: %i\n", p->scoring_params.num_bootstrap_draws);
		fprintf(fp,"bootstrap_Std_threshold: %5.2lf\n", p->scoring_params.bootstrap_Std_threshold);
	}
	else
		fprintf(fp,"bootstrap_motif_score: no\n");

	if(consider_rev_compl_flag == true)
	{
		fprintf(fp,"strand-specific: no\n");
	}
	else
	{
		fprintf(fp,"strand-specific: yes\n");
	}
	fprintf(fp,"\n");

//	if(p->geneset_ES == geneset_random_set_ES && p->scoring_params.pre_calculate_location_and_scale == 0)
//		fprintf(fp,"geneset_random_set_ES + exact null\n");
//	else if(p->geneset_ES == geneset_random_set_ES && p->scoring_params.pre_calculate_location_and_scale == estimate_global_MEAN_and_VAR_random_set)
/*

	if(p->select_next == select_next_motif_top_scoring)
		fprintf(fp,"select_next: select_next_motif_top_scoring\n");
	else if(p->select_next == select_next_motif_sample)
		fprintf(fp,"select_next: select_next_motif_sample\n");
	else
		fprintf(fp,"p->select_next: unkown type\n");
*/
	fprintf(fp,"degen_threshold: %5.2lf\n", p->degen_threshold);
	fprintf(fp,"fraction_degen_pos_threshold: %5.2lf\n", p->fraction_degen_pos_threshold);
	fprintf(fp,"scoring_params.max_gene_set_size_percentage_threshold: %5.2lf\n", p->scoring_params.max_gene_set_size_percentage_threshold);
	fprintf(fp,"scoring_params.min_gene_set_size_absolute_threshold: %i\n", p->scoring_params.min_gene_set_size_absolute_threshold);
	fprintf(fp,"scoring_params.min_gene_set_size_percentage_threshold: %5.2lf\n", p->scoring_params.min_gene_set_size_percentage_threshold);
//	fprintf(fp,"p->scoring_params.num_draws: %i\n", p->scoring_params.num_draws);
	fprintf(fp,"min_motif_length: %i\n", p->min_motif_length);
	fprintf(fp,"max_motif_length: %i\n", p->max_motif_length);
//	fprintf(fp,"p->num_neighbours: %i\n", p->num_neighbours);
	fprintf(fp,"\n");

	fprintf(fp,"========================\n");
	fprintf(fp,"ensemble PSSM parameters\n");
	fprintf(fp,"========================\n");
	fprintf(fp,"required_core_length: %i\n", p->required_core_length);
	fprintf(fp,"pssm_crop_threshold: %5.2lf\n", p->pssm_crop_threshold);
	fprintf(fp,"cluster_pssm_threshold_fraction: %5.2lf\n", p->cluster_pssm_threshold_fraction);
	fprintf(fp,"fraction_of_top_sequences_to_consider: %5.2lf\n", p->fraction_of_top_sequences_to_consider);
//	fprintf(fp,"p_IO->num_motif_clusters_to_output: %i\n", p_IO->num_motif_clusters_to_output);
	if(species_type_global == yeast)
		fprintf(fp,"species_type: yeast\n");
	else if(species_type_global == arabidopsis)
		fprintf(fp,"species_type: arabidopsis\n");
	else if(species_type_global == human)
		fprintf(fp,"species_type: human\n");
	else
		fprintf(fp,"species_type: unspecified (default: uniform) \n");
	fprintf(fp,"\n");

	fprintf(fp,"===============================================\n");
	fprintf(fp,"clustering similar motif predictions parameters\n");
	fprintf(fp,"===============================================\n");
	fprintf(fp,"cluster_sim_threshold: %5.2lf\n", p->cluster_sim_threshold);
	fprintf(fp,"hypegeom_p_val_cutoff: %e\n", p->hypegeom_p_val_cutoff);
	fprintf(fp,"\n");

	fprintf(fp,"===============\n");
	fprintf(fp,"MISC parameters\n");
	fprintf(fp,"===============\n");
	if(p->significance.evaluate_significance == true)
	{
		fprintf(fp,"estimate p-value: yes (%i iterations)\n", p->significance.num_random_permutations);
		fprintf(fp,"random scores file: %s\n", p_IO->null_scores_filename);
	}
	else
		fprintf(fp,"estimate p-value: no\n");

//	if(p->scoring_params.pre_calculate_global_datasets_location_and_scale == estimate_global_MEAN_and_VAR)
//		fprintf(fp,"p->scoring_params.pre_calculate_global_datasets_location_and_scale: estimate_global_MEAN_and_VAR\n");
//	else
//		fprintf(fp,"p->scoring_params.pre_calculate_global_datasets_location_and_scale: estimate_global_MEDIAN_and_IQR\n");

	fclose(fp);
//	printf("done outputing params\n");
//	getchar();
}

void create_parameters_log(params *p, IO_params *p_IO)
{
	char out_params[200];
	strcpy(out_params, p_IO->output_dir);strcat(out_params, "params_used");

//	if(p->scoring == Random_Set_no_ranks)
//		fprintf(stdout,"**scoring type: Random_Set_no_ranks\n");
//	else if(p->scoring == Random_Set_ranks)
//		fprintf(stdout,"**scoring type: Random_Set_ranks\n");
//	else
//		fprintf(stdout,"**scoring type: GSEA\n");

	output_param_settings(out_params, p, p_IO);

//	if(p->scoring == Random_Set_no_ranks)
//		fprintf(stdout,"**scoring type: Random_Set_no_ranks\n");
//	else if(p->scoring == Random_Set_ranks)
//		fprintf(stdout,"**scoring type: Random_Set_ranks\n");
//	else
//		fprintf(stdout,"**scoring type: GSEA\n");
}


/////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////


char *construct_file_name(char **name_parts_arr, int num_name_parts, char *dir_name)
{
	if(num_name_parts <= 0)
	{
		printf("ERROR: 'construct_file_name' illegal value for 'num_name_parts' = %i\n ", num_name_parts);
		exit(1);
	}
	char *file_name;
	int i; int num_bytes = 0;
	for(i=0;i<num_name_parts;i++)
		num_bytes += strlen(name_parts_arr[i]);
	num_bytes += strlen(dir_name) + (dir_name[strlen(dir_name)/sizeof(char) - 1] == '/')?0:1;
	MALLOC(file_name, sizeof(char)*num_bytes + 1);
	strcpy(file_name, dir_name);
	if(dir_name[strlen(dir_name)/sizeof(char) - 1] != '/')
		strcat(file_name, "/");
	for(i=0;i<num_name_parts;i++)
		strcat(file_name, name_parts_arr[i]);
	return file_name;
}


/////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////

// BINARY SEARCH
// finds the index of the BIGGEST entry in 'A' that is LESS THAN OR EQUAL to 'value'
// can be use to find index of a sequence which contains a motif match given in absolute coordinates
// in which case the array 'A' should contain the sequences' start positions
int search_location(int *A, int num, int value)
{
	int low = 0, high = num - 1;
	bool flag = false;
	int mid;
	while (low <= high)
	{
		mid = (low + high) / 2;
		if(A[mid] > value)
   			high = mid - 1;
		else if(A[mid] < value)
    		low = mid + 1;
    	else
      	{
//    		printf("exact match: %i\n", A[mid]);
        	return mid;
        }
	}
// 	printf("in range: [%i,%i]\n", A[high], A[high+1]);
	return high;
}

// call: test_search(strtol(argv[1], 0, 0));
void test_search(int value)
{
	int num = 5;
	int A[5] = {1,5,8,10,15};
	int res = search_location(A, num, value);
	printf("value: %i\tres: %i\n", value, res);
    getchar();
}

/////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////

/*
void free_gene_set_profile(gene_set_profile *gp)
{
	if(gp == 0)
		return;
	FREE(gp->oligo_counts);
}

void free_oligo_count_pairs(oligo_count_pair **oc, int num)
{
	if(oc == 0)
		return;
	int i;
	for(i=0;i<num;i++)
	{
		FREE(oc[i]);
	}
	FREE(oc);
}

oligo_count_pair **build_gene_set_profile(void **index, int seq_length, int num_sequences, gene_set_profile *gs_profile)
{
	int i,j,k;
	unsigned char *pattern;
	unsigned long pattern_length = strlen(gs_profile->oligos[0]);

	 // including overlaps!
	unsigned long numocc, **occ;
	MALLOC(occ, sizeof(unsigned long *));

	int pattern_in_seq[num_sequences];
	for(j=0;j<num_sequences;j++)
	{
		pattern_in_seq[j]=0;
	}

	oligo_count_pair **oc;
	MALLOC(oc, sizeof(oligo_count_pair *)*gs_profile->num_oligos);

	int seq_index;
	for(k=0;k<gs_profile->num_oligos;k++)
	{
		pattern=gs_profile->oligos[k];
		if(locate(*index, pattern, pattern_length, occ, &numocc) != 0)
		{
			printf("couldn't find %s in HUMAN!\n", pattern);
			continue;
		}
//		printf("numocc: %i\n", numocc);
//		getchar();
		for(i=0;i<numocc;i++)
		{
			seq_index = occ[0][i]/seq_length;
			if(seq_index < 0 || seq_index >= num_sequences)
			{
				printf("ERROR: array out of bounds!\npattern: %s\nseq_index: %i\nnum_sequences: %i\n",
																	pattern, seq_index, num_sequences);
				printf("occ[0][%i]=%i\tseq_length: %i\tnum_occ: %i\n", i, occ[0][i], seq_length, numocc);
				continue;
			}
			pattern_in_seq[seq_index] = 1;
		}
		FREE(*occ);

		MALLOC(oc[k], sizeof(oligo_count_pair));
		int ind = 0, count = 0; oc[k]->oligo = pattern; oc[k]->count = 0;
		gs_profile->oligo_counts[k] = 0;
		for(i=0;i<num_sequences;i++)
		{
			if(ind < gs_profile->num_genes && i == gs_profile->genes[ind])
			{
				ind++;
				if(pattern_in_seq[i] == 1)
				{
					gs_profile->oligo_counts[k] += 1;
					oc[k]->count +=1;
				}
			}
			pattern_in_seq[i]=0;
		}
	}
	FREE(occ);
	return oc;
}


void init_gene_set_profile(gene_set_profile *gp, int *genes, int num_genes, char **oligos, int num_o)
{
	gp->genes = genes; gp->num_genes = num_genes;
	gp->oligos = oligos; gp->num_oligos = num_o;
	MALLOC(gp->oligo_counts, sizeof(int)*gp->num_oligos);
}


// 'build_gene_set_profile' in 'degen_gene_set.c'
oligo_count_pair **build_oligo_counts(void **index, int seq_length, int num_sequences, int *genes,
									   int num_genes, char **oligos, int num_o)
{
	gene_set_profile gp;
	init_gene_set_profile(&gp, genes, num_genes, oligos, num_o);
	return build_gene_set_profile(index, seq_length, num_sequences, &gp);
}

//run_motif_finder(double *seq_level_stat, char **seq_ids, char *seq_file, int num_seq, char *expected_motif, int seq_length)
// returns an array of 'oligo_count_pair *' containing pairs of (oligo, counts)
// for all instantiations of a given 'motif' found in 'index'
oligo_count_pair **get_oligo_counts(char *motif, int *num_instantiations, void **index, int num_seq,
									 int seq_length, int *genes, int num_genes)
{
	if(genes == 0)
	{
		int i;
		num_genes = num_seq;
		MALLOC(genes, sizeof(int)*num_genes);
		for(i=0;i<num_genes;i++)
			genes[i] = i;
	}
	// initializes '*num_instantiations' with the number of instantiations of 'motif'
	char **target_oligos = decode_IUPAC(motif, strlen(motif), num_instantiations);
//	getchar();
	oligo_count_pair **oc = build_oligo_counts(index, seq_length, num_seq, genes, num_genes, target_oligos, *num_instantiations);
	FREE(genes);
	return oc;
}


// generates and outputs to a file the counts of a given motif
// creates two files: 'gene set profile' and 'outside gene set profile'
void get_oligo_profile(d_gs *motif, void **index, int num_seq, int seq_length, int *genes, int num_genes, char *seq_file, params *p)
{
	int i, num_inst;
	char **target_oligos = decode_IUPAC(motif->degen_oligo, motif->oligo_length, &num_inst);
	for(i=0;i<num_inst;i++)
	{
		printf("%s\n", target_oligos[i]);
	}

	oligo_count_pair **oc = build_oligo_counts(index, seq_length, num_seq, genes, num_genes, target_oligos, num_inst);

	score_type st = p->scoring;

	char filename[400];
	strcpy(filename, seq_file);
	strcat(filename, "_gs_profile");
	FILE *f_counts = open_file(filename, "w");
	fprintf(f_counts, "total num genes: %i\n", num_genes);
	for(i=0;i<num_inst;i++)
	{
		char *oligo;
		MALLOC(oligo, sizeof(char) * (strlen(oc[i]->oligo) + 1));
  		strcpy(oligo, oc[i]->oligo);

	//	printf("%s\t%i\n", oc[i]->oligo, oc[i]->count);
		d_gs *oligo_motif = create_d_gs(oligo, index, num_seq, motif->gene_level_stat, seq_length, st, NULL, ZERO, 0, p);
		fprintf(f_counts, "%s\t%i\t%5.3lf\n", oc[i]->oligo, oc[i]->count, oligo_motif->score);
		free_d_gs(oligo_motif);
	}
	fclose(f_counts);
	free_oligo_count_pairs(oc, num_inst);
}
*/

/////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////


species get_species_type(char *species)
{
	if(strcmp(species, "-y") == 0)
		return yeast;
	else if(strcmp(species, "-h") == 0)
		return human;
	else if(strcmp(species, "-a") == 0)
		return arabidopsis;
	else
	{
		printf("Sorry, but currently only -y 'yeast', -h 'human', or -a 'arabidopsis' option is supported!\n");
		exit(1);
	}
}


/////////////////////
/////////////////////

char *clone_string(char *source)
{
	char *str;
	MALLOC(str, sizeof(char *) * strlen(source) + 1);
	strcpy(str, source);
	return str;
}

void read_oligos(char *filename, oligomers *oligo_seeds)
{
	int i;
//	printf("filename: %s\n", filename);
	read_file_lines(filename, &oligo_seeds->raw_input, &oligo_seeds->num_oligos);

	// removed for submission
	
 //   printf("oligo_seeds->num_oligos: %i\n", oligo_seeds->num_oligos);

	
	int delim = "\n";

	if(oligo_seeds->num_oligos == 0)
	{
		printf("ERROR: 'read_oligos', oligos file empty! oligo_seeds->num_oligos = %i\n", oligo_seeds->num_oligos);
		exit(1);
	}
	MALLOC(oligo_seeds->oligos, sizeof(char *) * oligo_seeds->num_oligos);
	oligo_seeds->is_empty = false;
	oligo_seeds->oligos[0] = strtok(oligo_seeds->raw_input[0], delim);
	for(i=1;i<oligo_seeds->num_oligos;i++)
	{
		oligo_seeds->oligos[i] = strtok(oligo_seeds->raw_input[i], delim);
//		printf("%i: %s\n", i, oligo_seeds->oligos[i]);
	}
//	getchar();
//	return oligo_seeds;
}

void free_oligos(oligomers *oligo_seeds)//uchar **oligos, int num)
{
	int i;
	if(oligo_seeds->is_empty == false)
	{
		free_file_lines(&oligo_seeds->raw_input, oligo_seeds->num_oligos);
		FREE(oligo_seeds->oligos);
	}
}

/////////////////////
/////////////////////

void create_and_set_output_directory(char *name, char **output_directory)
{
	char new_dir[200];
	strcpy(new_dir, name);
	create_dir(new_dir);
	MALLOC(*output_directory, strlen(new_dir) + 3);
	strcpy(*output_directory, ".");
	strcat(*output_directory, new_dir);
	strcat(*output_directory, "/");
}


/////////////////////////
/////////////////////////

// need to preallocate 'retbuf' with sufficiently many characters to hold the number 'n'
char *itoa(int n, char *retbuf)
{
	sprintf(retbuf, "%d", n);
	return retbuf;
}

bool eq_ignore_case(char *str1, char *str2)
{
	int len1 = strlen(str1), len2 = strlen(str2);
	if(len1 != len2)
		return false;
	else
	{
		int i;
		for(i=0;i<len1;i++)
		{
			if(toupper(str1[i]) != toupper(str2[i]))
				return false;
		}
		return true;
	}
}

void test_eq_ignore_case()
{

	char *str1 = "adfDc";
	char *str2 = "AdfdC";
	if(eq_ignore_case(str1, str2) == true)
		printf("true\n");
	else
		printf("false\n");
	getchar();
}

bool eq(char *str1, char *str2)
{
//	eq_ignore_case(str1, str2);
	if(strcmp(str1, str2) == 0)
		return true;
	else
		return false;
}


// invert the order of elements in an integer array
void invert_order(double *arr, int num)
{
	int i; double tmp;
	for(i=0;i<num/2;i++)
	{
		tmp = arr[i];
		arr[i] = arr[num - i - 1];
		arr[num - i - 1] = tmp;
	}
}


void call_prog(char *command)
{
//	call_prog("R <\"/home/sgg4/R/my_GSEA/test.Run.Leukemia_C1.R\" --no-save");
/* First use the system() call. Output will go to stdout. */
 	system(command);
}


////////////////////////////
////////////////////////////

// reverses the order of sequences and p-values in an input dataset
// (e.g. making them in increasing instead of decreasing order or vise versa)
void reverse_sequece_order(char *seq_file, char *seq_statistics_file)
{
	int i;
//	int num_seq, num_samples_phen1, num_samples_phen2;
//	gene_stat_seq_pair **g_s;
//	char *in_file = "/home/sgg4/Desktop/main_project/data/human/E2F3/unmasked_E2F3_seq";
//	char *in_file = "/home/sgg4/workspace1/motif_finder/test_seq";

	// reading sequence
	FILE *fp = open_file(seq_file, "r");
	char *lineptr = 0; int n = 0; int delim = '$';
	int num_sequences = 0;
	while(getdelim(&lineptr, &n, delim, fp) > 0)
	{
		num_sequences++;
		FREE(lineptr);
	}
	fclose(fp);

	num_sequences;
	printf("num_sequences: %i\n", num_sequences/2);
	getchar();

	char *seq[num_sequences];
	fp = open_file(seq_file, "r");
	for(i=0;i<num_sequences;i++)
	{
		seq[i] = 0;
		getdelim(&(seq[i]), &n, delim, fp);
	}
	fclose(fp);

	// writing sequence

	FILE *fp_seq = open_file("/home/sgg4/workspace1/motif_finder/test_seq_out", "w");
	for(i=num_sequences-1; i>0; i -= 2)
	{
		fprintf(fp_seq, "%s%s", seq[i-1], seq[i]);
		FREE(seq[i-1]);
		FREE(seq[i]);
	}
	fclose(fp_seq);
//	printf("done\n");
//	getchar();

	// reading seq_level_stats
//	FILE *fp_p_vals_in = open_file("/home/sgg4/workspace1/motif_finder/test_p_vals", "r");
	FILE *fp_p_vals_in = open_file(seq_statistics_file, "r");

	int num_p_vals = 0;
	char delim1 = '\n';
	while(getdelim(&lineptr, &n, delim1, fp_p_vals_in) > 0)
	{
//		printf("%s\n", lineptr);
		num_p_vals++;
		FREE(lineptr);
	}
	fclose(fp_p_vals_in);
	printf("num_p_vals: %i\n", num_p_vals);
	getchar();

	if(num_p_vals != num_sequences/2)
	{
		printf("pvals: %i, seq: %i\n", num_p_vals, num_sequences/2);
		exit(1);
	}
	fflush(stdout);

	char *seq_ids[num_p_vals];
	double p_vals[num_p_vals];

//	fp_p_vals_in = open_file("/home/sgg4/workspace1/motif_finder/test_p_vals", "r");
	fp_p_vals_in = open_file(seq_statistics_file, "r");

	for(i=0;i<num_p_vals;i++)
	{
		if(fscanf(fp_p_vals_in, "%as\t%lf", &(seq_ids[i]), &(p_vals[i]) ) == 0)
		{
			printf("ERROR reading p_vals: %i\n", i);
		}
	//	else
		//	printf("%s\t%lf\n", seq_ids[i], p_vals[i]);
	}

	// writing seq_level_stats
	FILE *fp_p_vals_out = open_file("/home/sgg4/workspace1/motif_finder/test_p_vals_out", "w");
	for(i=num_p_vals-1;i>=0;i--)
	{
		fprintf(fp_p_vals_out, "%s\t%lf\n", seq_ids[i], p_vals[i]);
		FREE(seq_ids[i]);
	}
	fclose(fp_p_vals_out);

	printf("done\n");
	fflush(stdout);
	getchar();
}

void test_reverse_sequece_order()
{
//	char *seq_file = "/home/sgg4/Desktop/main_project/data/human/E2F3/unmasked_E2F3_seq";
//	char *seq_statistics_file = "/home/sgg4/Desktop/main_project/data/human/E2F3/inputfile_E2F3";

	char *seq_file = 	"/home/sgg4/workspace1/motif_finder/test_seq";
	char *seq_statistics_file = "/home/sgg4/workspace1/motif_finder/test_p_vals";

	reverse_sequece_order(seq_file, seq_statistics_file);
}

/////////////////////////////
/////////////////////////////

int seq_length_global;

// 'extract' does not add '\0' at the end of the char* it extracts
void extract_from_locations()//int *locations, int num_locations, void **index)
{
	int i;
	printf("inside extract_from_locations\n");
	char *seq_file = "/home/sgg4/workspace1/motif_finder/test_seq";//"/home/sgg4/c_workspace/GMF_new/source_code_new/input/yeast/unmaskedseqfile_ABF1_YPD.fasta";//"test_index";
	void **index = create_index(seq_file, &seq_length_global, 3);
	printf("done\n");
	printf("seq_length_global: %i\n",seq_length_global);

	unsigned long numocc;
	unsigned long **occ;
	MALLOC(occ, sizeof(unsigned long *));
	const char *str = "T";
	locate(*index, str, strlen(str), occ, &numocc);
	printf("occ[0][0]: %i\n", occ[0][0]);

	char **snippet;
	MALLOC(snippet, sizeof(char *));
	int from = occ[0][0], to = occ[0][0] + 2;
//	MALLOC(*snippet, sizeof(char) * (to-from + 2));
//	(*snippet)[to-from+1] = 0;

 //   ulong from=1; ulong to=2;
	int snippet_length;
	int res = extract(*index, from, to, snippet, &snippet_length);
	printf("snippet_length: %i\n", snippet_length);
//	(*snippet)[snippet_length] = 0;
//	printf("snippet: %s\n", *snippet);
	for(i=0;i<snippet_length;i++)
		printf("%c", (*snippet)[i]);
	getchar();
}

/*
void test_index_search()
{
	int i;
	char *seq_file = "/home/sgg4/workspace1/motif_finder/test_seq";//"test_index";
	void **index = create_index(seq_file, &seq_length_global, -1);
	int num_seq = 3;
	char *motif = "TGA";int num_instantiations;
	char **target_oligos = decode_IUPAC(motif, strlen(motif), &num_instantiations);
	motif_instances all_motif_instances;
	find_degen_oligo_set_occurrence_locations_fixed_length
					(index, target_oligos, strlen(motif), num_instantiations,
					seq_length_global, num_seq, &all_motif_instances);
	printf("done\n");
	for(i=0;i<num_seq;i++)
	{
		printf("sequence %i:\n", i);
		tree_walk_in_sorted_order(all_motif_instances.locations_tree);
	}
	getchar();
}
*/

////////////////////
////////////////////

//typedef struct
//{
//	char *raw_input;
//	char **tokens; int num_tokens;
//	char *delim_str;
//	int curr;
//}
//tokenizer;

int count_tokens(const char *string_to_parse, const char *delim_str)
{
	int curr = 0;int num_tokens = 0;char *res;
	while( (res = strstr(string_to_parse + curr, delim_str)) != NULL)
	{
		curr = curr + (res-(string_to_parse+curr)) + strlen(delim_str);
		num_tokens++;
	}
	if(num_tokens > 0 && curr < strlen(string_to_parse) || num_tokens == 0 && strlen(string_to_parse) > 0)
		num_tokens++;
	return num_tokens;
}

tokenizer create_tokenizer(const char *string_to_parse, const char *delim_str, bool remove_newlines)
{
	tokenizer t;
	MALLOC(t.raw_input, sizeof(char) * strlen(string_to_parse) + 1);
	strcpy(t.raw_input, string_to_parse);
	t.num_tokens = count_tokens(t.raw_input, delim_str);//t.num_tokens++;
//	printf("num_tokens: %i\n", t.num_tokens);
	t.curr = 0;
	MALLOC(t.tokens, sizeof(char *) * t.num_tokens);
	MALLOC(t.delim_str, sizeof(char) * strlen(delim_str) + 1);
	strcpy(t.delim_str, delim_str);

	int i; int curr = 0, ind = 0;
	char *res;
	for(i=0;i<t.num_tokens;i++)
	{
		res = strstr(t.raw_input + curr, delim_str);
		int len;

		if(res != NULL)
		{
			len = (res-(t.raw_input+curr));
			if(res[0] == '\n') {t.num_tokens--; break;}
		}
		else
			len = strlen(t.raw_input) - curr;
		MALLOC(t.tokens[ind], sizeof(char)*len+1);
		t.tokens[ind][len] = '\0';
		strncpy(t.tokens[ind], t.raw_input+curr, len);
//		printf("%i: %s(%i)\n", ind, t.tokens[ind], strlen(t.tokens[ind]));
		ind++; curr = curr + (res-(t.raw_input+curr)) + strlen(delim_str);
	}
	if(t.tokens[t.num_tokens - 1][0] == '\n')
	{
	//	printf("freeing\n");
		FREE(t.tokens[t.num_tokens - 1]);
		t.num_tokens--;
	}
	if(remove_newlines == true && t.tokens[t.num_tokens - 1][strlen(t.tokens[t.num_tokens - 1]) - 1] == '\n')
		t.tokens[t.num_tokens - 1][strlen(t.tokens[t.num_tokens - 1]) - 1] = '\0';
	return t;
}

void destroy_tokenizer(tokenizer *t)
{
	int i;
	for(i=0;i<t->num_tokens;i++)
		FREE(t->tokens[i]);
	FREE(t->tokens);
	FREE(t->raw_input);
	FREE(t->delim_str);
}

char *reset_tokenizer_curr_index(tokenizer *t)
{
	t->curr = 0;
}

char *get_next_token(tokenizer *t)
{
	if(t->curr < t->num_tokens)
	{
		t->curr++;
		return t->tokens[t->curr-1];
	}
	else
		return 0;
}

void display_tokens(tokenizer *t)
{
//	printf("%s\n", t->raw_input);
	char *res;
	while(res = get_next_token(&t))
		printf("%s\n", res);
}

void test_tokenizer()
{
	char *string_to_parse = "||aa";
	char *delim_str = "||";
	bool remove_newlines = true;
	tokenizer t = create_tokenizer(string_to_parse,delim_str,remove_newlines);
	char *res;
	printf("%s\n", string_to_parse);
	printf("num_tokens: %i\n", t.num_tokens);
	while(res = get_next_token(&t))
		printf("%s\n", res);
//	reset_tokenizer_curr_index(&t);
//	while(res = get_next_token(&t))
//		printf("%s\n", res);
	destroy_tokenizer(&t);
	printf("done\n");
//	getchar();
}


////////////////////
////////////////////



#include <stdio.h>


/*
* Copy prefix of string pointed by pbuf before first occurance of delim
* to dst. Increment pbuf so that it points to part of string right after
* occurance of delim. Return 0 if operation succeeds, error code (to specify)
* otherwise.
* @param dst_size size of buffer dst
*/

//#include <string.h>
//char *strstr(const char *haystack, const char *needle);

GET_DELIM_RESULT get_to_delim(const char** pbuf, const char* delim, char* dst, int dst_size)
{
	const char* sub = strstr(*pbuf,delim);
	if(sub==NULL)
		return GET_DELIM_NOT_FOUND_DELIM;
	else
	{
		int prefix_len = sub-(*pbuf);

		if(prefix_len+1>dst_size)
			return GET_DELIM_PREFIX_TOO_LONG;
		// strncat requires dst to be null-terminated
		*dst=0;
		strncat(dst,*pbuf,prefix_len);
		*pbuf+=prefix_len+strlen(delim);
		return GET_DELIM_OK;
	}
}

int test_get_to_delim()
{
	const char buf[] = "abc||def&&ghi";
	const char* pbuf = buf;
	char abc[10];
	char def[10];
	get_to_delim(&pbuf,"||",abc,10);
	get_to_delim(&pbuf,"&&",def,10);
	printf("%s\n",abc);
	printf("%s\n",def);
	return 0;
}



double sum_elements(double *vec, int num)
{
	int i;
	double sum = 0;	
	for(i=0;i<num;i++)
		sum += vec[i];
//	printf("sum = %lf\n", sum);//getchar();
	return sum;
}

double sum_square_elements(double *vec, int num)
{
	int i;
	double sum_square = 0;	
	for(i=0;i<num;i++)
		sum_square += vec[i]*vec[i];
	return sum_square;
}


void center_and_scale_data_matrix(gsl_matrix *mat)
{
	int i;
	for(i=0;i<mat->size2;i++)
	{
		gsl_vector_view col_view = gsl_matrix_column(mat, i);
		double mean = sum_elements(col_view.vector.data, col_view.vector.size) / mat->size1;
//		printf("mean = %lf\n", mean);getchar();
		double sum_square  = sum_square_elements(col_view.vector.data, col_view.vector.size);
//		printf("sum_square = %lf\n", sum_square);getchar();

		double sd = sqrt(sum_square / mat->size1 - mean * mean); 
		gsl_vector_add_constant(&col_view.vector, -mean);
		if(sd == 0)
			printf("WARNING: 'utilities.c', 'center_and_scale_data_matrix', sd == 0! column = %i\n", i);
		else
			gsl_vector_scale(&col_view.vector, 1.0/sd);
//		gsl_vector_scale(&col_view.vector, 1.0/mat->size1);
	}
}

void center_data_matrix(gsl_matrix *mat)
{
	int i;
	for(i=0;i<mat->size2;i++)
	{
		gsl_vector_view col_view = gsl_matrix_column(mat, i);
		double mean = sum_elements(col_view.vector.data, col_view.vector.size) / mat->size1;
		gsl_vector_add_constant(&col_view.vector, -mean);
	}
}

void print_matrix_GSL_m_n(gsl_matrix *dat, int m, int n)
{
	printf("%i-by-%i\n", dat->size1, dat->size2);
	int i,j;
	int display_index1 = m > dat->size1 ? dat->size1 : m;
	int display_index2 = n > dat->size2 ? dat->size2 : n;
	for(i=0;i<display_index1;i++)
	{
		for(j=0;j<display_index2;j++)
		{
			printf("%10.5lf ", gsl_matrix_get(dat,i,j));
//			printf("%20.15lf ", gsl_matrix_get(dat,i,j));
		}
		printf("\n");
	}	
	printf("\n");
}


void print_vector_n_GSL(gsl_vector *dat, int n)
{
//	printf("length = %i\n", dat->size);
	int i;
	int display_index = n > dat->size ? dat->size : n;
	for(i=0;i<display_index;i++)
	{
		printf("%10.5lf ", dat->data[i]);
	}	
	printf("\n");
}



////////////////////
////////////////////

/* implementation of comparison between two Position Specific Frequencey
 * Matrices using Pearson Correlation/Kulback-Libler divergence. The master function
 * is 'main1' and it can be made into the 'main' module that accpets as
 * input PFWMs and generates the output similarity scores. Laplace smoothing not implemented  */
// assume rows are in following order: A, G, C, T
//gsl_matrix *read_mat(char *file)
/*
void compare_PCC_divergence_batch_pssm_files(char *filename_list_file, char *out_file)
{
	int i;
	char **file_names; int num_files;
	read_file_name_list(filename_list_file, &file_names, &num_files);

	gsl_matrix *pssm1, *pssm2;
	char *delim = "\t";
	char *curr_pssm1_file, *curr_pssm2_file;
	FILE *fp = open_file(out_file, "w");
	for(i=0;i<num_files;i++)
	{
		curr_pssm1_file = strtok(file_names[i], delim);
		pssm1 = read_mat(curr_pssm1_file);

		curr_pssm2_file = strtok(NULL, delim);
		pssm2 = read_mat(curr_pssm2_file);

		gsl_matrix *rev_compl_planted_pssm1 = get_rev_compl_pssm(pssm1);

		double similarity1 = similarity_score_pearson_correlation_with_overhang(pssm1, pssm2);
		double similarity2 = similarity_score_pearson_correlation_with_overhang(rev_compl_planted_pssm1, pssm2);
		double similarity_PCC = MAX(similarity1, similarity2);

//		double similarity1 = similarity_score_KL(pssm1, pssm2);
//		double similarity2 = similarity_score_KL(rev_compl_planted_pssm1, pssm2);
// 		double similarity_PCC = MIN(similarity1, similarity2);

//		printf("%lf\n", similarity_PCC); getchar();
		fprintf(fp, "%lf\n", similarity_PCC);
		gsl_matrix_free(pssm1);
		gsl_matrix_free(pssm2);
		gsl_matrix_free(rev_compl_planted_pssm1);

		FREE(file_names[i]);
	}
	fclose(fp);
	FREE(file_names);
}

int main_PCC(int argc, char*argv[])
{
//	input_params **params;
	int size_params_arr, i;
	if(argc == 3)
	{
		compare_PCC_divergence_batch_pssm_files(argv[1], argv[2]);
	}
	else
	{
		printf("Usage: main_out <list_input_pssm_files> <output_file>\n");
		printf("Output: Pearson Correlation Coefficient for each pair of pssms from the <list_input_pssm_files>\n");
      	exit(0);
	}
}

*/

// Print functions
/*
void print_d_gs_oligo(const d_gs *dg)
{
	printf("%s\n",dg->degen_oligo);
}

void print_d_gs_genes(const d_gs *dg)
{
	int i;
	for(i=0;i<dg->num_genes;i++)
		printf("%i\t",dg->genes[i]);
	printf("\n");
}

void print_d_gs_gene_set(const d_gs *dg)
{
	int i;
	printf("%s\n",dg->degen_oligo);
	for(i=0;i<dg->num_genes; i++)
		printf("\t%i",dg->genes[i]);
	printf("\n");
}

// need to add code to manage 'seq_level_satat'
void print_seq_level_stat(d_gs *dg)
{
	printf("seq_level_statistics: %s\n", dg->degen_oligo);
	if(dg->num_genes == 0)
		printf("empty gene set!\n");
	else
	{
		int i;
		for(i=0;i<dg->num_genes;i++)
			printf("%i: %lf\n", i, dg->gene_level_stat[dg->genes[i]]);
	}
}

*/


