/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#ifndef UTILITIES_H_
#define UTILITIES_H_

// version that works also  if a == b
#define SWAP(a, b) (((a) == (b)) || (((a) ^= (b)), ((b) ^= (a)), ((a) ^= (b))))

#include <unistd.h>
#include <sys/types.h>
#include <stdbool.h>

#include "random_gen.h"
#include "parameters.h"
#include "degen_gene_set.h"
#include "interface.h"
#include "geneset_scoring_interface.h"
#include "motif_evolution_interface.h"
#include "geneset_scoring.h"


#include <gsl/gsl_cblas.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_permutation.h>

// needed for hash table implementation
// int hcreate(size_t nel);
// void hdestroy(void);
// ENTRY *hsearch(ENTRY item, ACTION action);
// also binary tree search
#define _GNU_SOURCE
#include <search.h>

typedef struct
{
	double *array;
	int num;
}
double_arr;


gsl_matrix *diag(gsl_vector *X);
gsl_matrix *diag_inv(gsl_vector *X);


void center_and_scale_data_matrix(gsl_matrix *mat);
void center_data_matrix(gsl_matrix *mat);
void print_matrix_GSL_m_n(gsl_matrix *dat, int m, int n);
void print_vector_n_GSL(gsl_vector *dat, int n);

char *clone_string(char *source);

void create_out_dir(IO_params *p_IO);
bool should_use_conservation(IO_params *p_IO);

void read_file_name_list(char *filename_list_file, char ***file_names, int *num_files);


void create_and_set_output_directory(char *name, char **output_directory);
void read_oligos(char *filename, oligomers *oligos);

void free_oligos(oligomers *oligos);

void **create_index(char *text_filename, int *seq_length, int num_sequences_to_read);
void destroy_index(void ***index);


char *extract_subsequence(void **index, int from, int to, int *length_of_extracted_seq);

gsl_matrix *MAT_times_MAT(gsl_matrix *A, gsl_matrix *B);
gsl_vector *MAT_times_VEC(gsl_matrix *m, gsl_vector *x);
gsl_vector *VEC_times_MAT(gsl_vector *x, gsl_matrix *m);

gsl_vector *lower_triangular_MAT_times_VEC(gsl_matrix *m, gsl_vector *x);
void test_MAT_times_VEC();

int count_num_lines(char *file);
int count_num_lines_with_double_entry(char *file);
double_arr read_array_from_file(char *file);

char *extract_TF_name(char *seq_file_path, params *p);

typedef enum
{
	GET_DELIM_OK=0,
	GET_DELIM_NOT_FOUND_DELIM,
	GET_DELIM_PREFIX_TOO_LONG
} GET_DELIM_RESULT;

GET_DELIM_RESULT get_to_delim(const char** pbuf, const char* delim, char* dst, int dst_size);
int test_get_to_delim();

typedef struct 
{
	char *raw_input;
	char **tokens; int num_tokens;
	char *delim_str;
	int curr;
}
tokenizer;

tokenizer create_tokenizer(const char *string_to_parse, const char *delim_str, bool remove_newlines);
char* get_next_token(tokenizer *t);
void destroy_tokenizer(tokenizer *t);

// returns teh set of uniquer motif indexes in '*indexes_of_unique_motifs' fomr the given array of 'd_gs **arr'
//void extract_unique_motifs(d_gs **arr, int num, int **indexes_of_unique_motifs, int *num_unique);

void create_parameters_log(params *p, IO_params *p_IO);
//void output_param_settings(char *summary_file_name, params *p);

char *construct_file_name(char **name_parts_arr, int num_name_parts, char *dir_name);

// BINARY SEARCH
// finds the index of the BIGGEST entry in 'A' that is LESS THAN OR EQUAL to 'value'
// can be use to find index of a sequence which contains a motif match given in absolute coordinates
// in which case the array 'A' should contain the sequences' start positions
int search_location(int *A, int num, int value);
void test_search(int value);

species get_species_type(char *species);

// convert an integer to a character string
// need to preallocate 'retbuf' with sufficiently many characters to hold the number 'n'
char *itoa(int n, char *retbuf);

// test for equality of strings (case insensitive)
bool eq_ignore_case(char *str1, char *str2);
void test_eq_ignore_case();

// test for equality of strings (case insensitive)
bool eq(char *str1, char *str2);

// invert the order of elements in an integer array
void invert_order(double *arr, int num);

//	call_prog("R <\"/home/sgg4/R/my_GSEA/test.Run.Leukemia_C1.R\" --no-save");
void call_prog(char *command);

#endif /*UTILITIES_H_*/
