#!/usr/bin/env python

###############################################################################
#                                                                             #
#    SSMART: Sequence-structure motif identification for RNA-binding proteins #
#    Version 1.2, 20 Jul 2017                                                 #
#    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           #
#                                                                             #
#    SSMART is free software: you can redistribute it and/or modify           #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.    #
#                                                                             #
#    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   #
#                                                                             #
###############################################################################

from argparse import ArgumentParser,FileType, SUPPRESS
from operator import itemgetter
import os, sys

PARAMS= ['img','species', 'max_seq_len', 'seed_len', 'seed_file', 'min_motif_len', 'max_motif_len', 'core_length', 'motif_sim', 'motif_hypergeom_pval', 'motif_degeneracy', 'motif_degeneracy_frac', 
         'pwm_crop', 'pwm_cluster', 'pwm_seq_frac', 'out_motifs', 'max_seq_frac', 'min_seq', 'min_seq_frac', 'bootstrap', 'bootstrap_draws', 'bootstrap_Std_threshold', 'random_runs', 'fast_mode']
PARAMS_INTERVAL = ['min_motif_len', 'max_motif_len', 'core_length', 'motif_sim', 'motif_degeneracy', 'motif_degeneracy_frac', 'pwm_crop', 'pwm_cluster', 'pwm_seq_frac', 'max_seq_frac', 'min_seq_frac']
ALPHABET = 'ACGTUacgtu'

ACCEPTED_COV = {0: 'length',
                1: 'nucleotide_composition',
                2: 'di-nucleotide_composition',
                3: 'num_uniqe_3mers', 
                4: 'num_uniqe_4mers', 
                5: 'num_uniqe_5mers'}
ACCEPTED_COV_INTERVAL = [0,5]
ACCEPTED_SEEDS = [3,4,5]
ACCEPTED_MIN_INTERVAL = [3,10]
ACCEPTED_MAX_INTERVAL = [5,15]
ACCEPTED_CORE_INTERVAL = [1,5]
ACCEPTED_FRAC_INTERVAL = [0.0,1.0]

oligo_folder= 'seeds'
r_script = 'plot_similarity.R'
separator = ' ' # not \t in this case


def run_ssmart(ssmart_path, input_file, output_name, covariates, params):
    # transform paths into absolute paths
    output_path = os.path.abspath(output_name[:output_name.rfind('/')])
    # get ssmart_path
    if not ssmart_path: 
        ssmart_path = os.getcwd() # get current path
    else:
        ssmart_path = os.path.abspath(ssmart_path)
    # create the result_path if it do not exists
    if not os.path.isdir(output_path):
        os.makedirs(output_path)    
    # pre-processing steps
    if output_name[-1]=='/':
        # if no case name is specified in output, use the name from input file
        case_name = input_file.name.split('/')[-1].split('.fa')[0]
    else:
        case_name = output_name.split('/')[-1]
    # convert fasta sequence format to SSMART internal sequence format
    [sequences, seq_file, score_file] = convert_fasta(case_name, input_file, output_path, params)
    # extract covariates from SSMART format sequence files
    cov_file = 'no_cov_file'
    if covariates: 
        cov_file = extract_covariates(case_name, sequences, covariates, output_path)
    # create SSMART command
    # $executable $evidence_file_list $sequence_file_list $out_dir $in_seeds_file $min_motif_length $max_motif_length $required_core_length $pssm_crop_threshold $cluster_pssm_threshold_fraction $cluster_sim_threshold $fraction_of_top_sequences_to_consider $hypegeom_p_val_cutoff $num_motif_clusters_to_output $degen_threshold $fraction_degen_pos_threshold $max_gene_set_size_percentage_threshold $min_gene_set_size_absolute_threshold $min_gene_set_size_percentage_threshold $num_clusters_to_output_motif_occurrences_for $species $random_runs_filename $num_random_runs $significance_estimates_file $strand_specific_analysis $bootstrap_Std_threshold $bootstrap_motif_score $num_bootstrap_draws $use_regression_scoring $covariates_file_list $fast_mode
    # ./cERMIT input-test/evidence_file_list/evidence_file_list_fold_output input-test/seq_file_list/sequence_file_list_fold_output out_seed_length=4:use_regression_scoring=no:bootstrap_motif_score=no:min_gene_set_size_percentage_threshold=0.10_tmp ./oligos_size_4_ul 4 10 3 0.25 0.5 0.75 0.5 1.0e-20 3 0.75 0.75 0.95 20 0.10 3 human random_scores 0 significance_estimates yes 2.0 no 20 no input-test/cov_file_list/covariates_file_list_fold_output no
    cmd = os.path.join(ssmart_path,'SSMART')+' '
    cmd += score_file+' '                           # argv[1] - score file # or os.path.join(output_path,score_file)
    cmd += seq_file+' '                             # argv[2] - seq file
    cmd += output_path+' '                          # argv[3] - output path
    # file with oligo seeds
    if params['seed_file']:
        cmd += os.path.abspath(params['seed_file'])+' ' # argv[4] - seed file
    else:
        cmd += os.path.join(ssmart_path,oligo_folder,'oligos_size_'+str(params['seed_len'])+'_sel')+' '
    cmd += str(params['min_motif_len'])+' '         # argv[5] - min_motif_len
    cmd += str(params['max_motif_len'])+' '         # argv[6] - max_motif_len
    cmd += str(params['core_length'])+' '           # argv[7] - core_length
    cmd += str(params['pwm_crop'])+' '              # argv[8] - pwm_crop
    cmd += str(params['pwm_cluster'])+' '           # argv[9] - pwm_cluster
    cmd += str(params['motif_sim'])+' '             # argv[10] - motif_sim
    cmd += str(params['pwm_seq_frac'])+' '          # argv[11] - pwm_seq_frac
    cmd += str(params['motif_hypergeom_pval'])+' '  # argv[12] - motif_hypergeom_pval
    cmd += str(params['out_motifs'])+' '            # argv[13] - out_motifs
    cmd += str(params['motif_degeneracy'])+' '      # argv[14] - motif_degeneracy
    cmd += str(params['motif_degeneracy_frac'])+' ' # argv[15] - motif_degeneracy_frac
    cmd += str(params['max_seq_frac'])+' '          # argv[16] - max_seq_frac
    cmd += str(params['min_seq'])+' '               # argv[17] - min_seq
    cmd += str(params['min_seq_frac'])+' '          # argv[18] - min_seq_frac
    cmd += str(params['out_motifs'])+' '            # argv[19] - out_motifs
    cmd += str(params['species'])+' '               # argv[20] - species
    cmd += 'random_scores '                         # argv[21] - random_runs_filename
    cmd += str(params['random_runs'])+' '           # argv[22] - num_random_runs
    cmd += 'significance_estimates '                # argv[23] - significance_estimates_file
    cmd += 'yes '                                   # argv[24] - strand_specific_analysis
    cmd += str(params['bootstrap_Std_threshold'])+' ' # argv[25] - bootstrap_Std_threshold
    cmd += str(params['bootstrap'])+' '             # argv[26] - bootstrap
    cmd += str(params['bootstrap_draws'])+' '       # argv[27] - bootstrap_draws
    if covariates:
        cmd += 'yes '                               # argv[28] - use_regression_scoring
    else:
        cmd += 'no '
    cmd += str(cov_file)+' '                        # argv[29] - cov_file
    cmd += str(params['fast_mode'])+' '             # argv[30] - fast_mode
    cmd += str(params['img'])                       # argv[31] - img_type
    # ssmart run
    print 'COMMAND: '+cmd
    os.system(cmd)
    # plot similarity between unique evolved motifs
    plot_similarity(output_path, case_name, params['img'], params['out_motifs'])
    # remove score & seq files
    remove_file(output_path, score_file)
    remove_file(output_path, seq_file)
    # final output
    print 'DONE!'
    return True

def remove_file(path, name):
    myfile = os.path.join(path, name)
    if os.path.isfile(myfile):
        os.remove(myfile)
    return True
    
def plot_similarity(path, case_name, img_type, out_motifs):
    # R < plot_similarity.R --slave --vanilla --args ../results/parclipHEK/similarities2/ZFP36_SRR1046759_sim.txt eps "ZFP similar. blabla"
    # read file with extra_info - (lower triangular matrix) and scores   
    input_file = open(os.path.join(path, case_name+'_extra_info.txt'),'r')
    [motifs, sim, scores, clusters] = read_matrix(input_file)
    # write a tmp file with similarity in the accepted format
    tmp_file = os.path.join(path, case_name+'_sim')
    write_matrix(motifs, sim, scores, clusters, tmp_file)
    # plot
    cmd = 'R < '+r_script+' --slave --vanilla --args '+os.path.join(path, tmp_file)+' '+img_type+' "'+case_name+' similarities" '+str(out_motifs)
    os.system(cmd)
    # print cmd
    # remove sim txt file    
    remove_file(path, tmp_file)
    return

def write_matrix(motifs, sim, scores, clusters, out_file):
    # write the matrix in a simpler txt file, filling the upper diagonal
    dim = len(sim)
    assert dim==len(sim[-1])
    # put -1 on the upper triangular part (?)    
    for idx in range(dim):
        for blank in range(idx+1,dim):
            sim[idx].append(-1)
    # add the scores as the second column
    str_sim = 'Sim\tScores\tClusters\t'
    for motif in motifs:
        str_sim += motif + '\t'
    str_sim += '\n'
    for i in range(dim):
        str_sim += motifs[i]+'\t'
        str_sim += str(scores[i])+'\t'
        str_sim += str(clusters[i])+'\t'
        for j in range(dim): 
            if sim[i][j]>-1:
                str_sim += str(sim[i][j])+'\t'
            else:
                str_sim += str(sim[j][i])+'\t'
        str_sim += '\n'
    f = open(out_file,'w')
    f.write(str_sim)
    f.close()
    
def read_matrix(input_file):
    # read data
    lines = input_file.readlines()
    input_file.close()
    # construct list
    matrix = False
    scores = []
    clusters = []
    sim = []
    for line in lines:
        if line.startswith("Similarity"):
            # labels
            labels = line.strip().split(separator)[1:]
            # start of the matrix
            matrix = True
        else:
            if matrix: # read similarity
                line = line.strip().split(separator)
                if len(line)>0:
                    sim.append([float(item) for item in line[1:]])
            else: # read scores
                line = line.strip().split('\t')
                l = len(line[0])
                if (l>3) and (l<18):
                    scores.append(float(line[1]))
                    clusters.append(line[-1])
    assert len(labels)==len(scores)
    assert len(labels)==len(sim)
    return [labels,sim, scores, clusters]    

def convert_fasta(case_name, input_file, output_path, params):
    # read a fasta file, convert to SSMART format and write 2 new file: .scores & .seq
    # .scores file has content like:
# G3206.1    2.703
# G788.1    2.386
# G146.1    2.292
    # .seq file has a single line with all sequences: seq+'X'*n+'$'
# tatctggttGatcctgctAGTagXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX$AGTAGgaGaaagtagtattTCTAAATTaatattattttGtcATATATTTCATTTTAAATGTACGTGTAGTAACCAAGTGAATGTATTTACTAGTATAAATTGTAAATTAGatatttgacgggt$tGATGATGATGatGatgatgatgAgtGAXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX    
    # read input file
    print 'Parsing sequences in '+case_name+' ...'
    fasta = SimpleFastaParser(input_file)
    input_file.close()    
    # parse sequences, get scores and max length
    data = [] # input list [id, seq, score]
    max_seq_len = 0
    for seq in fasta:        
        # get seq length
        seq_len = len(seq[1])
        if seq_len<=params['max_seq_len']:
            # test sequence
            if test_seq(seq[1], ALPHABET):
                # replace all Us with Ts and add it to input list
                sequence = seq[1].replace('U','T').replace('u','t')
                # build scores from header
                header = seq[0].split('|')
                data.append([header[0], sequence, float(header[-1])])
                if max_seq_len<seq_len:
                    max_seq_len = seq_len    
    # sort sequences by scores
    data.sort(key=itemgetter(2), reverse=True)
    # write scores
    scores = ''
    for [sid, seq, score] in data:
        scores += sid+'\t'+str(score)+'\n'
    score_file = case_name+'.scores'
    f=open(os.path.join(output_path, score_file),'w')
    f.write(scores) 
    f.close()
    # convert sequences
    output_sequences = ''
    sequences = []
    for [sid, seq, score] in data:          
        sequences.append(seq)
        # padd seq with X and add it to file
        output_sequences += '{message:{fill}{align}{width}}'.format(message=seq, fill='X', align='<', width=max_seq_len) + '$'    
    # write scores
    seq_file = case_name+'.seq'
    f=open(os.path.join(output_path, seq_file),'w')
    f.write(output_sequences) 
    f.close()
    # some verbose
    print "{:<33}".format('Number of read sequences:')+"{:>6}".format(str(len(fasta)))
    print "{:<33}".format('Number of considered sequences:')+"{:>6}".format(str(len(sequences)))+' (with length<'+str(params['max_seq_len'])+' and nucleotides in {'+ALPHABET+'})'
    return [sequences, seq_file, score_file]

def test_seq(seq, alphabet):
    # test that the characters are in the given alphabet
    for letter in seq:
        if letter not in alphabet:
            return False
    return True

def extract_covariates(case_name, sequences, covariates, output_path):
    # compute all covariates values and write 2 files: .cov & .cov.names
    lseq = len(sequences)
    cov_names = [] # names of columns
    cov_values = [[] for idx in range(lseq)] # list of list with values by seq, not by cov
    # compute values  
    # 0 - length
    if 0 in covariates:
        cov_names.append(ACCEPTED_COV[0])
        for idx in range(lseq):
            cov_values[idx].append(len(sequences[idx]))
    # 1 - nucleotide_composition & 2 - di-nucleotide_composition
    for cov in [1,2]:
        if cov in covariates:
            if cov==1:
                names = [letter for letter in ALPHABET[:4]]
            else:
                names = [l1+l2 for l1 in ALPHABET[:4] for l2 in ALPHABET[:4]]
            for name in names:
                cov_names.append(name)
            for idx in range(lseq):
                kmers = count_kmers(sequences[idx], cov)
                for name in names:
                    cov_values[idx].append(kmers.get(name,0))
    # 3, 4, 5 - num_uniqe_3mers, num_uniqe_4mers, num_uniqe_5mers
    for cov in [3,4,5]:
        if cov in covariates:
            cov_names.append(ACCEPTED_COV[cov])
            for idx in range(lseq):
                cov_values[idx].append(num_unique_kmers(sequences[idx], cov))
    # write values
    cov_file = case_name+'.cov'
    f=open(os.path.join(output_path, cov_file),'w')
    f.write(print_matrix(cov_values)) 
    f.close()
    # write names
    cov_names_file = case_name+'.cov.names'
    f=open(os.path.join(output_path, cov_names_file),'w')
    f.write(" ".join(cov_names)) 
    f.close()
    print "Extracted covariates: "+" ".join(cov_names)
    return cov_file

def print_matrix(matrix):
    # convert values from a list of list to a print-ready string (tab separated)
    result = str(len(matrix))+'\t'+str(len(matrix[0]))+'\n'
    for row in matrix:
        for item in row:
            result += str(item)+'\t'
        result +='\n'
    return result

def count_kmers(seq, k):
    # returns the number of all kmers in seq
    kmers = {}
    for idx in range(len(seq)-k+1):
        kmer = seq[idx:idx+k]
        kmers[kmer] = kmers.get(kmer, 0) + 1
    return kmers
  
def num_unique_kmers(seq, k):
    # returns the number of unique kmers in seq
    kmers = count_kmers(seq, k)
    return len(kmers.keys())

def SimpleFastaParser(handle):
    """Generator function to iterator over Fasta records (as string tuples).

    For each record a tuple of two strings is returned, the FASTA title
    line (without the leading '>' character), and the sequence (with any
    whitespace removed). The title line is not divided up into an
    identifier (the first word) and comment or description.

    >>> for values in SimpleFastaParser(open("Fasta/dups.fasta")):
    ...     print(values)
    ('alpha', 'ACGTA')
    ('beta', 'CGTC')
    ('gamma', 'CCGCC')
    ('alpha (again - this is a duplicate entry to test the indexing code)', 'ACGTA')
    ('delta', 'CGCGC')

    """
    #Skip any text before the first record (e.g. blank lines, comments)
    while True:
        line = handle.readline()
        if line == "":
            return  # Premature end of file, or just empty?
        if line[0] == ">":
            break

    result = []
    while True:
        if line[0] != ">":
            raise ValueError(
                "Records in Fasta files should start with '>' character")
        title = line[1:].rstrip()
        lines = []
        line = handle.readline()
        while True:
            if not line:
                break
            if line[0] == ">":
                break
            lines.append(line.rstrip())
            line = handle.readline()

        #Remove trailing whitespace, and any internal spaces
        #(and any embedded \r which are possible in mangled files
        #when not opened in universal read lines mode)
        result.append([title, "".join(lines).replace(" ", "").replace("\r", "")])

        if not line:
            return result # StopIteration

    assert False, "Should not reach this line"

def extract_params(args):
    # Test parameter values and return dictionary with all secondary param values
    params = {}
    # params with no further restriction
    params['img'] = args.img
    params['species'] = args.species
    params['max_seq_len'] = args.max_seq_len
    params['seed_file'] = args.seed_file
    params['motif_hypergeom_pval'] = args.motif_hypergeom_pval
    params['out_motifs'] = args.out_motifs
    params['min_seq'] = args.min_seq
    if args.fast_mode:
        params['fast_mode'] = 'yes'
    else:
        params['fast_mode'] = 'no'
    if args.bootstrap:
        params['bootstrap'] = 'yes'
    else:
        params['bootstrap'] = 'no'
    params['bootstrap_draws'] = args.bootstrap_draws
    params['bootstrap_Std_threshold'] = args.bootstrap_Std_threshold
    params['random_runs'] = args.random_runs
    # seed length
    if args.seed_len in ACCEPTED_SEEDS:
        params['seed_len'] = args.seed_len
    else:
        print "Wrong value for '--seed' option. Accepted values: "+", ".join(str(seed) for seed in ACCEPTED_SEEDS)
        return False
    # params with interval restrictions
    params['min_motif_len'] = test_param('min', args.min_motif_len, ACCEPTED_MIN_INTERVAL)
    params['max_motif_len'] = test_param('max', args.max_motif_len, ACCEPTED_MAX_INTERVAL)
    params['core_length'] = test_param('core', args.core_length, ACCEPTED_CORE_INTERVAL)
    params['motif_sim'] = test_param('sim', args.motif_sim, ACCEPTED_FRAC_INTERVAL)
    params['motif_degeneracy'] = test_param('degen', args.motif_degeneracy, ACCEPTED_FRAC_INTERVAL)
    params['motif_degeneracy_frac'] = test_param('degen_frac', args.motif_degeneracy_frac, ACCEPTED_FRAC_INTERVAL)
    params['pwm_crop'] = test_param('pwm_crop', args.pwm_crop, ACCEPTED_FRAC_INTERVAL)
    params['pwm_cluster'] = test_param('pwm_cluster', args.pwm_cluster, ACCEPTED_FRAC_INTERVAL)
    params['pwm_seq_frac'] = test_param('pwm_seq_frac', args.pwm_seq_frac, ACCEPTED_FRAC_INTERVAL)
    params['max_seq_frac'] = test_param('max_seq_frac', args.max_seq_frac, ACCEPTED_FRAC_INTERVAL)
    params['min_seq_frac'] = test_param('min_seq_frac', args.min_seq_frac, ACCEPTED_FRAC_INTERVAL)
    # if one of the params with interval restrictions is False, return False
    for param in PARAMS_INTERVAL:
        if not param:
            return False
    return params

def test_param(name, value, interval):
    if value>=interval[0] and value<=interval[1]:
        return value
    else:
        print "Wrong value for '--"+name+"' option. Accepted values: "+str(interval)
        return False
    
def test_covariates(cov_list):
    # return true if all covariates are from the accepted list
    unique_cov = list(set(cov_list)) #removes duplicate elements, do not preserve order
    for cov in unique_cov:
        if cov<ACCEPTED_COV_INTERVAL[0] or cov>ACCEPTED_COV_INTERVAL[1]:
            return False
    return unique_cov

def hide_args(arglist):
    # if '-h' is used, display selected options and hide secondary options
    for action in arglist:
        action.help=SUPPRESS

########## Argument parser ##########
hidelist = [] # list of arguments to be hidded from normal help
parser = ArgumentParser(description='SSMART - Sequence-structure motif identification for RNA-binding proteins')
parser.add_argument('-f', '--file', action='store', dest="input_file", required=True, type=FileType('r'), help="Input fasta file with header lines containing ID|score and the sequence encoded with the 8 letter alphabet: A, C, G, T, a, c, g, t (uppercase = unpaired base, lowercase = paired base).")
parser.add_argument('-o', '--out', action='store', dest="output_name", required=True, help="Path and prefix for the output files.")
parser.add_argument('-c', '--cov', action='store', dest='covariates', nargs='+', type=int, default=False, help="The list of covariates to use in the motif analysis, separated by space. If any covariate is specified, SSMART will be run in regression setting. Accepted values: "+', '.join("%r (%s)" % (key,val) for (key,val) in ACCEPTED_COV.items())+". Default: False.")
parser.add_argument('-p','--path', action='store', dest="ssmart_path", default=False, help="Path for SSMART. Use only if you call this from someplace else. Default: current folder.")
parser.add_argument('--species', action='store', dest="species", default='human', help="Supported options: 'human', 'yeast', 'arabidopsis'; any other value will result in uniform background frequencies. Default: 'human'.")
parser.add_argument('--img', action='store', dest="img", choices=['png', 'eps', 'pdf'], default='png', help="The type of the image files. Supported options: 'png', 'eps', 'pdf'. Default: 'png'.")
parser.add_argument('--options', action = 'help', dest="options", help="Show extended list of program options and exit.")

hidelist.append(parser.add_argument('--max_seq', action='store', dest="max_seq_len", type=int, default=150, help="Maximum sequence length (skip longer sequence entries). Default: 150."))
hidelist.append(parser.add_argument('--out_motifs', action='store', dest="out_motifs", type=int, default=3, help="Number of top motif cluster PWM to be output. Default: 3."))

seed_group = parser.add_mutually_exclusive_group(required=False)
hidelist.append(seed_group.add_argument('--seed', action='store', dest="seed_len", type=int, default=4, help="Starting oligomer seed length. Accepted values: "+", ".join(str(seed) for seed in ACCEPTED_SEEDS)+". Default: 4."))
hidelist.append(seed_group.add_argument('--seed_file', action='store', dest="seed_file", type=FileType('r'), default=False, help="Path for a file with starting oligomer seeds, one on each line. If used, the --seed option is not available."))
hidelist.append(seed_group)

hidelist.append(parser.add_argument('--min', action='store', dest="min_motif_len", type=int, default=4, help="Minimum motif length. Accepted values: "+str(ACCEPTED_MIN_INTERVAL)+". Default: 4."))
hidelist.append(parser.add_argument('--max', action='store', dest="max_motif_len", type=int, default=10, help="Maximum motif length. Accepted values: "+str(ACCEPTED_MAX_INTERVAL)+". Default: 10."))
hidelist.append(parser.add_argument('--core', action='store', dest="core_length", type=int, default=3, help="Required core (overlapping) length for aligning motifs during clustering. Accepted values: "+str(ACCEPTED_CORE_INTERVAL)+". Default: 3."))
hidelist.append(parser.add_argument('--sim', action='store', dest="motif_sim", type=float, default=0.75, help="Required sequence similarity to assign motifs to the same cluster. Accepted values: "+str(ACCEPTED_FRAC_INTERVAL)+" (1 <=> identical). Default: 0.75."))
hidelist.append(parser.add_argument('--pval', action='store', dest="motif_hypergeom_pval", type=float, default=1.0e-20, help="Required hypergeometric p-value for co-occurrence to assign motifs to the same cluster. Default: 1.0e-20."))
hidelist.append(parser.add_argument('--degen', action='store', dest="motif_degeneracy", type=float, default=0.75, help="Allowable motif 'degeneracy score'. Range [0.0, 1.0], 0.0 => no degeneracy, 1.0 => all 'N's. Default: 0.75."))
hidelist.append(parser.add_argument('--degen_frac', action='store', dest="motif_degeneracy_frac", type=float, default=0.75, help="Allowable percentage of degenerate bases in a motif. Default: 0.75."))
hidelist.append(parser.add_argument('--pwm_crop', action='store', dest="pwm_crop", type=float, default=0.25, help="Crop all flanking positions in the PWM that have <= pwm_crop * (# reads of 'core'). Default: 0.25."))
hidelist.append(parser.add_argument('--pwm_cluster', action='store', dest="pwm_cluster", type=float, default=0.5, help="Consider all cluster members that have score >= pwm_cluster * top_cluster_score. Default: 0.5."))
hidelist.append(parser.add_argument('--pwm_seq_frac', action='store', dest="pwm_seq_frac", type=float, default=0.5, help="Fraction of top sequences to use in the construction of the PWM. Default: 0.5."))
hidelist.append(parser.add_argument('--max_seq_frac', action='store', dest="max_seq_frac", type=float, default=0.95, help="Maximum target size for a motif cluster (as a percentage of total number of input sequence regions). Default: 0.95."))
hidelist.append(parser.add_argument('--min_seq_frac', action='store', dest="min_seq_frac", type=float, default=0.1, help="Minimum target size for a motif cluster (as a percentage of total number of input sequence regions). Default: 0.1."))
hidelist.append(parser.add_argument('--min_seq', action='store', dest="min_seq", type=int, default=20, help="Minimum target size for a motif cluster (as an absolute number of input sequence regions). Default: 20."))
hidelist.append(parser.add_argument('--bootstrap', action='store_true', dest="bootstrap", default=False, help="If used, use a bootstrap filter during the motif 'evolution' step, when scoring the set of motif candidates. Default: False."))                                    
hidelist.append(parser.add_argument('--bootstrap_draws', action='store', dest="bootstrap_draws", type=int, default=20, help="Number of bootstrap draws. Default: 20."))
hidelist.append(parser.add_argument('--bootstrap_Std_threshold', action='store', dest="bootstrap_Std_threshold", type=float, default=2.0, help="Value used in the motif enrichment score 'stability' criterion: bootstrap_mean - bootstrap_Std_threshold * bootstrap_sd > 0, where 'bootstrap_mean' -> mean of bootstrapped samples from the set of motif target regions,  'bootstrap_sd'   -> standard deviation of bootstrapped samples from the set of motif target regions. Default: 2.0."))
hidelist.append(parser.add_argument('--random_runs', action='store', dest="random_runs", type=int, default=0, help="Number of random scores generated based on randomly assigned score values (used for a p-value estimate). Default: 0."))

hidelist.append(parser.add_argument('--fast_mode', action='store_true', dest="fast_mode", default=False, help="Fast mode option. If used, evolve only 'promising' motif seeds (that have positive scores). Default: evolve all motif seeds (slower option).")) 


if '-h' in sys.argv[1:] or '--help' in sys.argv[1:]:
    hide_args(hidelist)
args = parser.parse_args()

# Test & get parameter values 
params = extract_params(args) # dictionary with all secondary param values

if params:
    # check the covariates list
    if args.covariates:
        covariates = test_covariates(args.covariates)
        if covariates:
            run_ssmart(args.ssmart_path, args.input_file, args.output_name, covariates, params) # linear regression scoring of target set, using intercept, center + scale covariates (using sample StdDev)
        else:
            print "Wrong value for '-c' option. Accepted values:\n"+'\n'.join("%r (%s)" % (key,val) for (key,val) in ACCEPTED_COV.items())+"\nor any combination of them separated by space."
    else:
        run_ssmart(args.ssmart_path, args.input_file, args.output_name, args.covariates, params) # Random Set scoring of target set     
         
