/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#include "degen_gene_set.h"
#include "sort.h"

extern const char primary_IUPAC_nucleotides[8];
extern const int no_primary_IUPAC_nucleotides;

// human upstream 700bp of TSS nucleotide counts:
// A, T, G, C	0.224627689666608 0.224627689666608 0.275372310333392 0.275372310333392
// cumulative:  0.224627689666608 0.449255379333216 0.724627689666608 1.0

// yeast
// p(A) = 0.3098, p(C) = 0.1909, p(G) = 0.1906, p(T) = 0.3087
// p(A) = 0.32444181395159
// p(C) = 0.184170782480918
// p(G) = 0.176762316829258
// p(T) = 0.314625086738234
gsl_vector *background_freq_vec_AGCT;

bool is_valid_score(double score){ return score > 0;}

// creating gene sets
void create_empty_d_gs(params *p, d_gs **dg)
{
	MALLOC(*dg, sizeof(d_gs));
	(*dg)->degen_oligo=0; (*dg)->oligo_length=0;
	(*dg)->pssm=0;
	(*dg)->genes=0; (*dg)->num_genes=0;
	(*dg)->instantiations=0; (*dg)->instantiations_length=0;
	(*dg)->gene_level_stat=0;
	(*dg)->score=0;//HUGE;
	(*dg)->is_valid=true;
	(*dg)->num_evolution_steps=0;
}

// creates a single 'd_gs' structure and returns a pointer to it
// assigns its pointers to pre-allocated and pre-initialized: 'oligo', 'seq_level_stat', 'gene_set->genes'
d_gs *create_d_gs(const char *oligo, params *p, int num_evolution_steps)
{
	void **index = p->d.seq.index;
	int total_num_genes = p->d.seq.num_seq;
	int seq_length = p->d.seq.seq_length;
	score_type st = p->scoring;

	d_gs *dg;
	create_empty_d_gs(p, &dg);
	dg->oligo_length = strlen(oligo);

	MALLOC(dg->degen_oligo, sizeof(char)*(dg->oligo_length + 1));
	strcpy(dg->degen_oligo, oligo);
	
	// allocates memory for 'dg->rev_compl_degen_oligo'
	dg->gene_level_stat = p->d.seq_level_stats;
	dg->p_value = -1.0;
	// generate all exact oligos matching the degenerate oligo pattern

	init_decode_IUPAC(dg->degen_oligo, &dg->instantiations, &dg->instantiations_counts, &dg->instantiations_length);
	decode_IUPAC(dg->degen_oligo, &dg->instantiations);

	int how_many_top_sequences = (int)(p->fraction_of_top_sequences_to_consider * total_num_genes);
	init_gene_set_occurrences_fixed_length(index, dg, seq_length, total_num_genes, how_many_top_sequences);

	dg->total_num_genes = p->d.seq.num_seq;
	dg->score = p->geneset_ES(dg->genes, dg->num_genes, p->d.seq_level_stats, p->d.num_seq_level_stats, &p->scoring_params);
	dg->num_evolution_steps = num_evolution_steps + 1;
//	printf("degen_gene_set.c Line 65\n");
	return dg;
}


// 'genes' and 'gene_level_stat' are initialized and destroyed somewhere else
void free_d_gs(d_gs *dg)
{
	if(dg == 0)
		return;
	int i;
	free_decode_IUPAC(&dg->instantiations, dg->instantiations_length, &dg->instantiations_counts);
//	for(i=0;i<dg->instantiations_length;i++)
//		FREE(dg->instantiations[i]);
//	FREE(dg->instantiations);
//	FREE(dg->instantiations_counts);
	FREE(dg->genes);
	gsl_matrix_free(dg->pssm);
	FREE(dg->degen_oligo);
	FREE(dg);
}


void free_d_gs_batch(d_gs **dg_arr, int num)
{
	if(dg_arr == 0)
		return;
	int i;
	for(i=0;i<num;i++)
		free_d_gs(dg_arr[i]);
}

void estimate_pssm_freq_MLE(double *theta_MLE, int num, double *target, int n, double *theta_shrinkage_out)
{
	int i;
	double lambda = 0;
	for(i=0;i<num;i++)
		theta_shrinkage_out[i] = theta_MLE[i];
}


// "Entropy Inference and the James-Stein estimator" by Jean Hausser and Korbinian Strimmer
void estimate_pssm_freq_using_James_Stein_shrinakage(double *theta_MLE, int num, double *target, int n, double *theta_shrinkage_out)
{
	int i;
	double sum_theta_MLE_squared = 0.0, mse = 0.0;
	for(i=0;i<num;i++)
	{
		sum_theta_MLE_squared += theta_MLE[i] * theta_MLE[i];
		mse += (target[i] - theta_MLE[i]) * (target[i] - theta_MLE[i]);
	}

	double lambda;
	if(n == 1 || (n - 1) * mse < 10e-10 * (1 - sum_theta_MLE_squared))
		lambda = 0;
	else
		lambda = (1 - sum_theta_MLE_squared) / ( (n - 1)*mse );

	for(i=0;i<num;i++)
	{
		theta_shrinkage_out[i] = lambda*target[i] + (1 - lambda)*theta_MLE[i];
//		printf("theta_MLE[%i] =%-10.3lftheta_shrinkage_out[%i] =%-10.3lf\n", i, theta_MLE[i], i, theta_shrinkage_out[i]);
	}
//	getchar();
}




void init_gene_set_occurrences_fixed_length(void **index, d_gs *dg, int seq_length, int num_sequences, int how_many_top_sequences)
// function that finds each occurrence of the motif in the input sequence
{
	int i,j,k;
	unsigned char *pattern;
	unsigned long pattern_length = dg->oligo_length;
	 // including overlaps!
	unsigned long numocc, *occ;
//	MALLOC(occ, sizeof(unsigned long *));

	int pattern_in_seq[num_sequences];
	int pattern_in_seq_counts[num_sequences];
	for(j=0;j<num_sequences;j++)
	{
		pattern_in_seq[j]=0;
//		occurrences[j]=0;
		pattern_in_seq_counts[j] = 0;
	}

	int seq_index;
	for(k=0;k<dg->instantiations_length;k++)
	{
		pattern=dg->instantiations[k];
//		printf("\n----- instantiations-index: %i\t\t instantiations: %s\n", k, pattern);
		dg->instantiations_counts[k] = 0;
		if(locate(*index, pattern, pattern_length, &occ, &numocc) != 0)
		{
			printf("'init_gene_set_occurrences_fixed_length': couldn't find %s!\n", pattern);
			FREE(occ);
			continue;
		}
//		printf("\n----- numocc: %i\tof %s -----\n", numocc, pattern);
		for(i=0;i<numocc;i++)
		{
			seq_index = occ[i] / seq_length;
			if(seq_index < 0 || seq_index >= num_sequences)
			{
				printf("ERROR (init_gene_set_occurrences_fixed_length): array out of bounds!\npattern: %s\nseq_index: %i\nnum_sequences: %i\n",
																	pattern, seq_index, num_sequences);
				printf("occ[0][%i]=%i\tseq_length: %i\tnum_occ: %i\n", i, occ[i], seq_length, numocc);
				FREE(occ);
				continue;
			}
			pattern_in_seq[seq_index] = 1;
			pattern_in_seq_counts[seq_index] = 1;
		}
		FREE(occ);

		for(i=0;i<num_sequences;i++)
		{
			if(pattern_in_seq_counts[i] == 1)
			{
				pattern_in_seq_counts[i] = 0;
				if(i < how_many_top_sequences)
					dg->instantiations_counts[k] += 1;
			}
		}
	}
//	FREE(occ);

	// initialize 'int dg->num_genes' and 'int *dg->genes'
	int occurrences[num_sequences];
	dg->num_genes = 0;
	for(i=0;i<num_sequences;i++)
	{
		if(pattern_in_seq[i] == 1)
		{
			occurrences[dg->num_genes] = i;
			dg->num_genes++;
		}
	}
	if(dg->num_genes == 0)
		dg->genes = 0;
	else
		MALLOC(dg->genes, sizeof(int)*dg->num_genes);

	for(i=0;i<dg->num_genes;i++)
		dg->genes[i] = occurrences[i];

	// initialize 'gsl_matrix *dg->pssm'
	char map_mat[257];
	int num = no_primary_IUPAC_nucleotides;
	for(i=0;i<num;i++)
		map_mat[primary_IUPAC_nucleotides[i]]=i;

	dg->pssm = gsl_matrix_alloc(num, dg->oligo_length);
	int q;
	double total;
	for(i=0;i<dg->oligo_length;i++)
	{
		int count[8]      = {0,0,0,0,0,0,0,0};
		for(j=0;j<dg->instantiations_length;j++)
		{
			count[map_mat[dg->instantiations[j][i]]] += dg->instantiations_counts[j];
		}
		total = 0;
		for(q=0;q<num;q++)
		{
			total += count[q];
		}

		gsl_vector *backgnd = background_freq_vec_AGCT;//get_background_freq_vec_AGCT();



		double theta_MLE[num];
		double pseudo_count;
		if (total>0.0)
			pseudo_count = (total / (num*100));
		else
			pseudo_count = (1.0 / (num*100));
//		printf("total = %f\n", total);
//		printf("pseudo_count = %f\n", pseudo_count);
//		printf("dg->PSSM: size1=%i, size2=%i\n",dg->pssm->size1, dg->pssm->size2);
		for(j=0;j<num;j++)
		{
			double bp_psudo_count = gsl_vector_get(backgnd, j) * pseudo_count;	
			gsl_matrix_set(dg->pssm, j, i, (count[j] + bp_psudo_count)/ (total + pseudo_count));
			theta_MLE[j] = (count[j] + bp_psudo_count)/ (total + pseudo_count);
//			theta_MLE[j] = count[j] / total;
		}


		double *target = backgnd->data;
		double theta_shrinkage_out[num];
		int n = total;
//		estimate_pssm_freq_using_James_Stein_shrinakage(theta_MLE, num, target, n, theta_shrinkage_out);
		estimate_pssm_freq_MLE(theta_MLE, num, target, n, theta_shrinkage_out);

		for(j=0;j<num;j++)
			gsl_matrix_set(dg->pssm, j, i, theta_shrinkage_out[j]);
	}
	//////////////////////////********************************
	//		print_pssm(dg->pssm);
	//////////////////////////********************************
}



// build pssm from sequence region indexes in 'dg->genes' 
void build_pssm_from_top_sequences_only(void **index, int seq_length, int num_sequences, d_gs *dg, int how_many_top_sequences)
{
	int i,j,k;
	unsigned char *pattern;
	unsigned long pattern_length = dg->oligo_length;
	 // including overlaps!
	unsigned long numocc;
	unsigned long **occ;
	MALLOC(occ, sizeof(unsigned long *));

	int pattern_in_seq[num_sequences];
	int occurrences[num_sequences];
	for(j=0;j<num_sequences;j++)
	{
		pattern_in_seq[j]=0;
		occurrences[j]=0;
	}


	for(k=0;k<dg->instantiations_length;k++)
	{
		pattern=dg->instantiations[k];
		if(locate(*index, pattern, pattern_length, occ, &numocc) != 0)
		{
			printf("'build_pssm_from_top_sequences_only': couldn't find %s!\n", pattern);
			FREE(*occ);
			continue;
		}
		for(i=0;i<numocc;i++)
		{
			if(occ[0][i]/seq_length < 0 || occ[0][i]/seq_length >= num_sequences)
			{
				printf("ERROR (build_pssm_from_top_sequences_only): array out of bounds! %li \n", occ[0][i]/seq_length);
				printf("occ[0][%i]=%i\t%i\n", i, occ[0][i], seq_length);
				FREE(*occ);
				continue;
			}
			pattern_in_seq[occ[0][i]/seq_length] = 1;
		}
		FREE(*occ);

		dg->instantiations_counts[k] = 0;

		// make sure the PSSM is built only using the genes in the '*genes'
		j = 0;
		for(i=0;i<dg->num_genes;i++)
		{
			while(j < dg->genes[i])
			{
				pattern_in_seq[j] = 0;
				j++;
			}
			if(pattern_in_seq[dg->genes[i]] == 1)
			{
				pattern_in_seq[dg->genes[i]] = 0;
				if(dg->genes[i] < how_many_top_sequences)
					dg->instantiations_counts[k] += 1;
			}
		}
	}
	FREE(occ);

	int num = no_primary_IUPAC_nucleotides;
	dg->pssm = gsl_matrix_alloc(num, dg->oligo_length);
	double total;
	int q;
	for(i=0;i<dg->oligo_length;i++)
	{
		int count[8] = {0,0,0,0,0,0,0,0};
		for(j=0;j<dg->instantiations_length;j++)
		{
			int idx = find_index(primary_IUPAC_nucleotides,no_primary_IUPAC_nucleotides,dg->instantiations[j][i]);
			if (idx>=0)
				count[idx]+=dg->instantiations_counts[j];
			else
				break;
		}
		total = 0;
		for(q=0;q<num;q++)
			total += count[q];

		double theta_MLE[num];

		gsl_vector *backgnd = background_freq_vec_AGCT;//get_background_freq_vec_AGCT();
		double pseudo_count = (total / (num*100));
		for(j=0;j<num;j++)
		{
			double bp_psudo_count = gsl_vector_get(backgnd, j) * pseudo_count; // A
			gsl_matrix_set(dg->pssm, j, i, (count[j] + bp_psudo_count)/ (total + pseudo_count));
			theta_MLE[j] = (count[j] + bp_psudo_count)/ (total + pseudo_count);
//			theta_MLE[j] = count[j] / total;
		}

		double *target = backgnd->data;
		double theta_shrinkage_out[num];
		int n = total;
//		estimate_pssm_freq_using_James_Stein_shrinakage(theta_MLE, num, target, n, theta_shrinkage_out);
		estimate_pssm_freq_MLE(theta_MLE, num, target, n, theta_shrinkage_out);

		for(j=0;j<num;j++)
			gsl_matrix_set(dg->pssm, j, i, theta_shrinkage_out[j]);

//		printf("done\n");getchar();
//		printf("after\n");
	}
}



// searching sequences for oligomer occurrences and return 
// set of seqeunce region indexes with at least on occurrence
int *find_degen_oligo_set_occurrences_fixed_length(void **index, unsigned char **oligos, int oligo_length, int num_oligos,
												   int seq_length, int num_sequences, int *num_occurrences)
{
	int i,j,k;
	unsigned char *pattern;
	unsigned long pattern_length = oligo_length;
	 // including overlaps!
	unsigned long numocc;
	unsigned long **occ;
	MALLOC(occ, sizeof(unsigned long *));

	int pattern_in_seq[num_sequences];
	int occurrences[num_sequences];
	for(j=0;j<num_sequences;j++)
	{
		pattern_in_seq[j]=0;
		occurrences[j]=0;
	}

	int seq_index;
	for(k=0;k<num_oligos;k++)
	{
		pattern=oligos[k];
		if(locate(*index, pattern, pattern_length, occ, &numocc) != 0)
		{
			printf("'find_degen_oligo_set_occurrences_fixed_length': couldn't find %s!\n", oligos[k]);
			FREE(*occ);
			continue;
		}
		for(i=0;i<numocc;i++)
		{
			seq_index = occ[0][i]/seq_length;
			if(seq_index < 0 || seq_index >= num_sequences)
			{
				printf("ERROR: array out of bounds! %li \n", seq_index);
				printf("occ[0][%i]=%i\t%i\n", i, occ[0][i], seq_length);
				FREE(*occ);
				continue;
			}
			pattern_in_seq[seq_index] = 1;
		}
		FREE(*occ);
	}
	FREE(occ);

	*num_occurrences = 0;
	for(i=0;i<num_sequences;i++)
	{
		if(pattern_in_seq[i] == 1)
		{
			occurrences[*num_occurrences] = i;
			(*num_occurrences)++;
		}
	}
	int *genes;
	if(*num_occurrences > 0)
	{
		MALLOC(genes, sizeof(int)*(*num_occurrences));
		for(i=0;i<*num_occurrences;i++)
		{
			genes[i] = occurrences[i];
//			printf("genes[%i]: %i\n", i, genes[i]);
		}
	}
	else
		genes = 0;
	return genes;
}





// Reverse complement

static char compl_mat[257];
bool rev_compl_init_first = true;
void init_compl_mat()
{
	if(rev_compl_init_first == true)
	{
		int i;
		for(i=0;i<257;i++)
			compl_mat[i] = get_complement(compl_mat[i]);
		rev_compl_init_first = false;
	}
}

char *rev_compl(char *oligo)
{
	init_compl_mat();
	int i;
	int len = strlen(oligo);
	char *rev_compl_o;
	MALLOC(rev_compl_o, sizeof(char)*(len+1))
	rev_compl_o[len] = '\0';
	for(i=0;i<len;i++)
		rev_compl_o[i] = compl_mat[oligo[len-i-1]];
	return rev_compl_o;
}

void test_rev_compl()
{
	char *o = "AGTC";
	char *rev_compl_o = rev_compl(o);
	printf("%s\n%s", o, rev_compl_o);
	FREE(rev_compl_o);
	getchar();
}

// Compare motifs

static double compare_score_mat[257][257];
bool compare_score_mat_init_first = true;
void init_compare_score_mat()
{
	if(compare_score_mat_init_first == true)
	{
		int i,j;
		for(i=0;i<257;i++)
			for(j=0;j<257;j++)
				// get the relationship score for encoding between two symbols: 1 if identic, 1/degeneracy if one is encoded by the other, 0 otherwise
				compare_score_mat[i][j] = get_relation_score(i,j);
		compare_score_mat_init_first = false;
	}
}


bool increase_degen_first = true;
static char increase_degen_mat[257][4];
void init_increase_degen_mat()
{
	if(increase_degen_first == true)
	{
		increase_degen_first = false;
		int i;
		for(i=0;i<257;i++)
			// get all possible new values for an increase in degeneracy (works for degen 1,2,3)
			strcpy(increase_degen_mat[i], increase_degeneracy(i));
	}
}

bool decrease_degen_first = true;
static char decrease_degen_mat[257][11];
void init_decrease_degen_mat()
{
	if(decrease_degen_first == true)
	{
		decrease_degen_first = false;
		int i;
		for(i=0;i<257;i++)
			// get all possible new values for a decrease in degeneracy (works for degen 2 and 4)
			strcpy(decrease_degen_mat[i], decrease_degeneracy(i));
	}
}

char new_str[100];

// appends a character to a string and returns the new string
char *append(char *str, unsigned str_length, char char_to_append)
{
	if(memcpy(new_str, str, str_length*sizeof(char))==NULL)
		printf("couldn't append: memcpy failed!\n");
	new_str[str_length]=char_to_append;
	new_str[str_length+1]='\0';
	return new_str;
}

// prepends a character to a string and returns the new string
char *prepend(char *str, int str_length, char char_to_prepend)
{
	if(memcpy(new_str+1, str, str_length*sizeof(char))==NULL)
		printf("couldn't prepend: memcpy failed!\n");
	new_str[0]=char_to_prepend;
	new_str[str_length+1]='\0';
	return new_str;
}

// return a substring defined by: start index=start, num characters = num_chars
char *substr(char *str, int start, int num_chars)
{
	if(memcpy(new_str, str + start, (num_chars)*sizeof(char))==NULL)
		printf("couldn't extract sustring: memcpy failed!\n");
	new_str[num_chars]='\0';
	return new_str;
}



/*
	 	(1) append  A, C, G, T, M(A,C), R(A,G), W(A,T), S(C,G), Y(G,T), K(G,T) & lowercase
	 	(2) prepend A, C, G, T, M(A,C), R(A,G), W(A,T), S(C,G), Y(G,T), K(G,T) & lowercase
	 	(3) exact -> double degenrate
		(4) double degenerate -> full degenerate (excluding the ends)
*/

int count_num_candidates(char *degen_oligo, bool bidirectional_evolution, params *p)
{
	int i;
	int count_new = 0;

	int degen_oligo_length = strlen(degen_oligo);


	// add degeneracy inside the starting motif
	for(i=0;i<degen_oligo_length;i++)
	{
		int degeneracy = get_degeneracy(degen_oligo[i]);
		if(degeneracy == 1)
		   	count_new += 4;
	// triple degeneracy code...
	   	else if(degeneracy == 2 && i != 0 && i != (degen_oligo_length-1))
			count_new += 2;
	}
	// append + prepend A,C,G,T,a,c,g,t * 2 = 16 and M,R,W,S,Y,K,m,r,w,s,y,k * 2 = 24  => TOTAL = 16 + 24 = 40
	count_new += (40 + 1); // add 1 for the original oligo gene set

	if(bidirectional_evolution == true)
	{
		// remove degeneracy inside the starting motif and reduce length
		for(i=0;i<degen_oligo_length;i++)
		{
			int degeneracy = get_degeneracy(degen_oligo[i]);
			if(degeneracy == 2)
				count_new += 3;
			else if(degeneracy == 4)
				count_new += 11;
		}
		if(degen_oligo_length > p->min_motif_length)
			count_new += 2; // add 2 for removing left and right end nucleotides
	}
	return count_new;
}

// has_score_improved(top_dg->score, evolved->score)
inline bool has_score_improved(double old_score, double new_score)
{
	return new_score < old_score;
}


d_gs *create_sub_geneset(char *oligo, void **index, int *subset_genes, int num_subset_genes, 
			int total_num_genes, int seq_length, score_type st, params *p, int num_evolution_steps)
{
//	printf("END create_sub_geneset\n");fflush(stdout);

	d_gs *tmp = create_d_gs(oligo, p, num_evolution_steps);
//	if(tmp == 0)
//		return 0;
//	else
//		FREE(tmp->genes);
	
	double *subset_genes_copy;
	MALLOC(subset_genes_copy, sizeof(int) * num_subset_genes);
	memcpy(subset_genes_copy, subset_genes, sizeof(int) * num_subset_genes);

	FREE(tmp->genes);
	tmp->genes = subset_genes_copy;//subset_genes;
	tmp->num_genes = num_subset_genes;

	gsl_matrix_free(tmp->pssm);

//	tmp->pssm = build_pssm_from_counts_fixed_length(index, seq_length, total_num_genes, tmp);
//	build_pssm_from_counts_fixed_length(index, seq_length, total_num_genes, tmp);
	int how_many_top_sequences = total_num_genes;
	build_pssm_from_top_sequences_only(index, seq_length, total_num_genes, tmp, how_many_top_sequences);


	tmp->score = p->geneset_ES(tmp->genes, tmp->num_genes, p->d.seq_level_stats, p->d.num_seq_level_stats, &p->scoring_params);

//	printf("START create_sub_geneset\n");fflush(stdout);

//	score_d_gs(tmp, p);
	return tmp;
}


d_gs *create_geneset_applying_orth_filter(char *degen_oligo, params *p, int num_evolution_steps)
{
//	printf("START create_geneset_applying_orth_filter\n");fflush(stdout);

	void **index = p->d.seq.index;
	int total_num_genes = p->d.seq.num_seq;
	int seq_length = p->d.seq.seq_length;
	score_type st = p->scoring;

	int i, j;

	// create geneset in original species
	d_gs *gs_original_species = create_d_gs(degen_oligo, p, num_evolution_steps);
	if(gs_original_species->num_genes == 0)
	{
//		printf("gs_original_species->num_genes == 0\n");fflush(stdout);
		free_d_gs(gs_original_species);
		return 0;
	}

	if(p->d.num_orth_sequences < 1)
	{
//		printf("p->d.num_orth_sequences < 1\n");fflush(stdout);

//		local_copy = clone_oligo(degen_oligo);
//		free_d_gs(gs_original_species);
		return gs_original_species;
//		return create_d_gs(degen_oligo, index, total_num_genes, p->d.seq_level_stats, seq_length, st, NULL, 0, 0, p);
 	}


//	printf("111create_geneset_applying_orth_filter\n");fflush(stdout);

	// create subset using ortholog filter storing output in 'overlapped_genes' and 'num_overlaped_genes'
	int num_overlaped_genes, num_overlaped_genes_tmp;
	int overlapped_genes[total_num_genes], overlapped_genes_tmp[total_num_genes];

	// intersection of first ortholgous species with the "original"
	void **index_orth = p->d.orth_sequences[0].index;;
	int seq_length_orth = p->d.orth_sequences[0].seq_length;
	d_gs *gs_orth = create_d_gs(degen_oligo, p, num_evolution_steps);
	if(gs_orth->num_genes == 0)
	{
			free_d_gs(gs_original_species);
			free_d_gs(gs_orth);
			return 0;
	}
	
	num_overlaped_genes = gs_orth->num_genes;
	for(j=0;j<num_overlaped_genes;j++)
		overlapped_genes[j] = gs_orth->genes[j];
	
	get_overlapping_genes(overlapped_genes, num_overlaped_genes,
							gs_original_species->genes, gs_original_species->num_genes, 
							total_num_genes, overlapped_genes_tmp, &num_overlaped_genes_tmp);
	free_d_gs(gs_original_species);
	
	if(num_overlaped_genes_tmp == 0)
	{
		free_d_gs(gs_orth);
		return 0;
	}
		
	num_overlaped_genes = num_overlaped_genes_tmp;
	for(j=0;j<num_overlaped_genes;j++)
		overlapped_genes[j] = overlapped_genes_tmp[j];//genes_orth[i];
	
	int num_in_orth;
	
	// intersection of all remaining ortholgous species
	for(i = 1;i < p->d.num_orth_sequences; i++)
	{
		index_orth = p->d.orth_sequences[i].index;//p->d.orth_sequences.orth_index;
		seq_length_orth = p->d.orth_sequences[i].seq_length;//p->d.seq.orth_seq_length;
		
		// construct the geneset in the orthologous species (only the array of gene ideces)
		int *genes_orth = find_degen_oligo_set_occurrences_fixed_length(index_orth, gs_orth->instantiations, gs_orth->oligo_length,
				gs_orth->instantiations_length, seq_length_orth, total_num_genes, &num_in_orth);
		if(num_in_orth == 0)
		{
			free_d_gs(gs_orth);
			return 0;
		}
		
		get_overlapping_genes(overlapped_genes, num_overlaped_genes, genes_orth, num_in_orth, total_num_genes, overlapped_genes_tmp, &num_overlaped_genes_tmp);
		FREE(genes_orth);
		
		if(num_overlaped_genes_tmp == 0)
		{
			free_d_gs(gs_orth);
			return 0;
		}
		
		num_overlaped_genes = num_overlaped_genes_tmp;
		for(j=0;j<num_overlaped_genes;j++)
			overlapped_genes[j] = overlapped_genes_tmp[j];//genes_orth[i];
	}
	free_d_gs(gs_orth);

//	int *overlapped_genes_in_all_orthologs;
	int num_overlapped_genes_orth = num_overlaped_genes;
//	MALLOC(overlapped_genes_in_all_orthologs, sizeof(int)*num_overlapped_genes_orth);
	int overlapped_genes_in_all_orthologs[num_overlapped_genes_orth];
	for(i = 0;i < num_overlapped_genes_orth; i++)
		overlapped_genes_in_all_orthologs[i] = overlapped_genes[i];

	// create final candidate geneset using ortholog filter
	d_gs *filtered_geneset =
		create_sub_geneset(degen_oligo, index, overlapped_genes_in_all_orthologs, 
					num_overlapped_genes_orth, total_num_genes, seq_length, st, p, num_evolution_steps);

//	printf("END create_geneset_applying_orth_filter\n");fflush(stdout);
	return filtered_geneset;
}



d_gs *evolve_motif_report_top_bidirectional(char *degen_oligo, params *p, create_geneset_fptr create_d_gs_fptr, pt2select_next_motif select_next, int num_evolution_steps)
{
	void **index = p->d.seq.index;
	int total_num_genes = p->d.seq.num_seq;
	int seq_length = p->d.seq.seq_length;
	score_type st = p->scoring;
	double *gene_level_stat = p->d.seq_level_stats;
	int oligo_length = strlen(degen_oligo);

//	p->parent_geneset_size_switch = dg->geneset_size_switch;
	/*
	 	(1) append  A, C, G, T, M(A,C), R(A,G), W(A,T), S(C,G), Y(G,T), K(G,T)
	 	(2) prepend A, C, G, T, M(A,C), R(A,G), W(A,T), S(C,G), Y(G,T), K(G,T)
	 	(3) exact -> double degenerate
		(4) double degenerate -> full degenerate (excluding the ends)
	*/
	init_decrease_degen_mat();
	init_increase_degen_mat();
	int i,j,k;
	int count_new = 0;
	bool bidirectional_evolution = true;
	count_new = count_num_candidates(degen_oligo, bidirectional_evolution, p);
	count_new -= 24; // I don't know why, but error without it !!!!??????? (in original version was -12)

	d_gs *dg_evolved[count_new];

	// append + prepend
	if(oligo_length < p->max_motif_length)
	{
		for(i=0;i<8;i++) {
			// append A,G,C,T,a,g,c,t
			dg_evolved[i] = create_d_gs_fptr(append(degen_oligo, oligo_length, primary_IUPAC_nucleotides[i]), p, num_evolution_steps);
			// prepend A,G,C,T,a,g,c,t
			dg_evolved[i+8] = create_d_gs_fptr(prepend(degen_oligo, oligo_length, primary_IUPAC_nucleotides[i]), p, num_evolution_steps);
		}
	}
	else
	{
		// if MAX length has been reached, just make copies of the motifs with no extension
		for(i=0;i<16;i++)
			dg_evolved[i] = create_d_gs_fptr(degen_oligo, p, num_evolution_steps);
	}

	// include the original in the evolved list
	char tmp_oligo[100];
	strncpy(tmp_oligo, degen_oligo, oligo_length);
	tmp_oligo[oligo_length]='\0';
	int curr = 16;
	dg_evolved[curr] = create_d_gs_fptr(tmp_oligo, p, num_evolution_steps);

	// add degeneracy
	int ind = curr + 1;//21;
	for(i=0;i<oligo_length;i++)
	{
		// AGCT
		int degeneracy = get_degeneracy(degen_oligo[i]);
		if(degeneracy == 1)
		{
			for(j = 0; j < 4;j++)
		   	{
				strncpy(tmp_oligo, degen_oligo, oligo_length);
				tmp_oligo[oligo_length]='\0';
		   		tmp_oligo[i] = increase_degen_mat[tmp_oligo[i]][j];
		   		dg_evolved[ind + j] =
		   		create_d_gs_fptr(tmp_oligo, p, num_evolution_steps);
		   	}
			ind += 4; //new_oligo = 0;
		}
		else if(degeneracy == 2 && i != 0 && i != (oligo_length-1))
		{
			for(j = 0; j < 2;j++)
			{
				strncpy(tmp_oligo, degen_oligo, oligo_length);
				tmp_oligo[oligo_length]='\0';
				tmp_oligo[i] = increase_degen_mat[tmp_oligo[i]][j];
				dg_evolved[ind + j] =
				create_d_gs_fptr(tmp_oligo, p, num_evolution_steps);
			}
			ind += 2;
		}
	}

	// remove degeneracy
	for(i=0;i<oligo_length;i++)
	{
		int degeneracy = get_degeneracy(degen_oligo[i]);
		if(degeneracy == 2)
		{
			for(j=0;j<3;j++)
			{
				strncpy(tmp_oligo, degen_oligo, oligo_length);
				tmp_oligo[oligo_length]='\0';
			   	tmp_oligo[i] = decrease_degen_mat[tmp_oligo[i]][j];
			   	dg_evolved[ind] = create_d_gs_fptr(tmp_oligo, p, num_evolution_steps);
			   	ind++;
			}
		}
		else if(degeneracy == 4)
		{
			for(j=0;j<11;j++)
			{
				strncpy(tmp_oligo, degen_oligo, oligo_length);
				tmp_oligo[oligo_length]='\0';
				tmp_oligo[i] = decrease_degen_mat[tmp_oligo[i]][j];
				dg_evolved[ind] = create_d_gs_fptr(tmp_oligo, p, num_evolution_steps);
				ind++; //new_oligo = 0;
			}
		}
	}

	// remove leftmost and rightmost basepair
	if(oligo_length > p->min_motif_length)
	{
		int ind2 = oligo_length-2;
		while(ind2 > p->min_motif_length && get_degeneracy(degen_oligo[ind2])==4)
			ind2--;
		
		for(i=0;i<ind2;i++)
			tmp_oligo[i] = degen_oligo[i];
		tmp_oligo[ind2] = '\0';
		dg_evolved[ind] = create_d_gs_fptr(tmp_oligo, p, num_evolution_steps);
		ind++;

		int ind1 = 1;
		while((oligo_length - ind1) > p->min_motif_length && get_degeneracy(degen_oligo[ind1])==4)
			ind1++;
		for(i=ind1;i<oligo_length;i++)
			tmp_oligo[i-ind1] = degen_oligo[i];
		tmp_oligo[i-ind1] = '\0';

		dg_evolved[ind] = create_d_gs_fptr(tmp_oligo, p, num_evolution_steps);
		ind++; //new_oligo = 0;
	}

	d_gs *current = create_d_gs_fptr(degen_oligo, p, num_evolution_steps);
	int num_neighbours = count_new;
	double neighbours[num_neighbours];
	for(i=0;i<num_neighbours;i++)
	{
		if(dg_evolved[i] != 0)
			neighbours[i] = dg_evolved[i]->score;
		else
			neighbours[i] = 0;
	}
	int next_index;
	select_next(current->score, neighbours, num_neighbours, &next_index);

	for(i=0;i<num_neighbours;i++)
		if(i != next_index)
			free_d_gs(dg_evolved[i]);
	if(next_index == -1)
	{
		// current is selected
		return current;
	}
	else
	{	
		free_d_gs(current);
		return dg_evolved[next_index];
	}
}

/*

d_gs *evolve_motif_report_top_bidirectional_orth_filter(char *degen_oligo, params *p, create_geneset_fptr create_d_gs_fptr, pt2select_next_motif select_next, int num_evolution_steps)
{

	void **index = p->d.seq.index;
	int total_num_genes = p->d.seq.num_seq;
	int seq_length = p->d.seq.seq_length;
	score_type st = p->scoring;
	double *gene_level_stat = p->d.seq_level_stats;
	int degen_oligo_length = strlen(degen_oligo);

	init_decrease_degen_mat();
	init_increase_degen_mat();
	int i,j,k;
	int count_new = 0;
	bool bidirectional_evolution = true;
	count_new = count_num_candidates(degen_oligo, bidirectional_evolution, p);
	//count_new -= 12; why???????

	d_gs *dg_evolved[count_new];

	// append + prepend
	if(degen_oligo_length < p->max_motif_length)
	{
		for(i=0;i<8;i++) {
			// append A,G,C,T,a,g,c,t
			dg_evolved[i] = create_geneset_applying_orth_filter(append(degen_oligo, degen_oligo_length, primary_IUPAC_nucleotides[i]), p, num_evolution_steps);
			// prepend A,G,C,T,a,g,c,t
			dg_evolved[i+8] = create_geneset_applying_orth_filter(prepend(degen_oligo, degen_oligo_length, primary_IUPAC_nucleotides[i]), p, num_evolution_steps);
		}
	}
	else
	{
		// if MAX length has been reached, just make copies of the motifs with no extension
		for(i=0;i<16;i++)
			dg_evolved[i] = create_geneset_applying_orth_filter(substr(degen_oligo, 0, strlen(degen_oligo)), p, num_evolution_steps);
	}

	// include the original in the evolved list
	char tmp_oligo[100];
	strncpy(tmp_oligo, degen_oligo, degen_oligo_length);
	tmp_oligo[degen_oligo_length]='\0';
	int curr = 16;
	dg_evolved[curr] = create_geneset_applying_orth_filter(tmp_oligo, p, num_evolution_steps);

	// add degeneracy
	int ind = curr + 1;//21;
	for(i=0;i<degen_oligo_length;i++)
	{
		// AGCT
		int degeneracy = get_degeneracy(degen_oligo[i]);
		if(degeneracy == 1)
		{
			for(j = 0; j < 3;j++)
			{
				strncpy(tmp_oligo, degen_oligo, degen_oligo_length);
				tmp_oligo[degen_oligo_length]='\0';

				tmp_oligo[i] = increase_degen_mat[tmp_oligo[i]][j];
				dg_evolved[ind + j] = create_geneset_applying_orth_filter(tmp_oligo, p, num_evolution_steps);
			}
			ind += 3; //new_oligo = 0;
		}
		else if(degeneracy == 4 && i != 0 && i != (degen_oligo_length-1))
		{
			strncpy(tmp_oligo, degen_oligo, degen_oligo_length);
			tmp_oligo[degen_oligo_length]='\0';
			if (degen_oligo[i] >= 65 && degen_oligo[i] <= 90) {// is upper
				tmp_oligo[i] = 'N';
			} else {
				tmp_oligo[i] = 'n';
			}
			dg_evolved[ind] = create_geneset_applying_orth_filter(tmp_oligo, p, num_evolution_steps);
			ind += 1;// new_oligo = 0;
		}
	}

	// remove degeneracy
	for(i=0;i<degen_oligo_length;i++)
	{
		int degeneracy = get_degeneracy(degen_oligo[i]);
		if(degeneracy == 2)
		{
			for(j=0;j<2;j++)
			{
				strncpy(tmp_oligo, degen_oligo, degen_oligo_length);
				tmp_oligo[degen_oligo_length]='\0';
				tmp_oligo[i] = decrease_degen_mat[tmp_oligo[i]][j];
				dg_evolved[ind] = create_geneset_applying_orth_filter(tmp_oligo, p, num_evolution_steps);
				ind++;
			}
		}
		else if(degeneracy == 4)
		{
			for(j=0;j<10;j++)
			{
				strncpy(tmp_oligo, degen_oligo, degen_oligo_length);
				tmp_oligo[degen_oligo_length]='\0';
				tmp_oligo[i] = decrease_degen_mat[tmp_oligo[i]][j];
				dg_evolved[ind] = create_geneset_applying_orth_filter(tmp_oligo, p, num_evolution_steps);
				ind++; //new_oligo = 0;
			}
		}
	}

	// remove leftmost and rightmost basepair
	if(degen_oligo_length > p->min_motif_length)
	{
		int ind2 = degen_oligo_length-2;
		while(ind2 > p->min_motif_length && get_degeneracy(degen_oligo[ind2])==4)
			ind2--;

		for(i=0;i<ind2;i++)
			tmp_oligo[i] = degen_oligo[i];
		tmp_oligo[ind2] = '\0';
		dg_evolved[ind] = create_geneset_applying_orth_filter(tmp_oligo, p, num_evolution_steps);
		ind++;

		int ind1 = 1;
		while((degen_oligo_length - ind1) > p->min_motif_length && get_degeneracy(degen_oligo[ind1])==4)
			ind1++;
		for(i=ind1;i<degen_oligo_length;i++)
			tmp_oligo[i-ind1] = degen_oligo[i];
		tmp_oligo[i-ind1] = '\0';

		dg_evolved[ind] = create_geneset_applying_orth_filter(tmp_oligo, p, num_evolution_steps);
		ind++; //new_oligo = 0;
	}

	int first_non_zero = 0;
	while(first_non_zero < count_new && dg_evolved[first_non_zero] == 0) first_non_zero++;
	if(first_non_zero == count_new)
		return 0;
	d_gs *top = dg_evolved[first_non_zero];
	int start_index = first_non_zero + 1;

	d_gs *current = create_geneset_applying_orth_filter(degen_oligo, p, num_evolution_steps);

	int num_neighbours = count_new - first_non_zero;
	double neighbours[num_neighbours];
	for(i=0;i<num_neighbours;i++)
	{
		if(dg_evolved[i+first_non_zero] != 0)
			neighbours[i] = dg_evolved[i+first_non_zero]->score;
		else
			neighbours[i] = 0;
	}
	int next_index;
	select_next(current->score, neighbours, num_neighbours, &next_index);
	for(i=0;i<num_neighbours;i++)
		if(i != next_index && dg_evolved[i+first_non_zero] != 0)
			free_d_gs(dg_evolved[i+first_non_zero]);
	if(next_index == -1)
	{
		// current is selected
		return current;
	}
	else
	{	
		free_d_gs(current);
		return dg_evolved[next_index+first_non_zero];
	}
}
*/

void select_next_motif_top_scoring(double current, double *neighbours, int num_neighbours, int *next_index)
{
	int i;
	*next_index = 0;
	for(i=1;i<num_neighbours;i++)
	{
		if(has_score_improved(neighbours[*next_index], neighbours[i]) == true)
			*next_index = i;
	}
}

void eval_degen_stats(char *oligo, double *degen_score, double *fraction_degen_pos)
{
	*fraction_degen_pos = 0; *degen_score = 0;
	int i;
	int len = strlen(oligo);
	for(i=0;i<len;i++)
	{
		switch(get_degeneracy(oligo[i]))
		{
			case 1:
				break;
			case 2:
				*degen_score += 0.5;
				*fraction_degen_pos += 1;
				break;
			case 4:
				*degen_score += 1;
				*fraction_degen_pos += 1;
				break;
			default:
				break;
		}
	}
	*degen_score = *degen_score / len;
	*fraction_degen_pos = *fraction_degen_pos / len;
}

inline void sort_evolved(d_gs **evolved, int num)
{
	sort_order so = increasing;
	sort_d_gs(evolved, num, so);
}

bool pass_degen_filter(double degen_score, double fraction_degen_pos, double degen_score_threshold, double fraction_degen_pos_threshold)
{
	return (degen_score < degen_score_threshold) && (fraction_degen_pos < fraction_degen_pos_threshold);
}

extern int count_draws;


void get_overlapping_genes(int *genes1, int num1, int *genes2, int num2, int total_num_genes, int *overlapped_genes, int *num_overlaped_genes)
{
	int i;
	*num_overlaped_genes = 0;

	int ind1 = 0;//int overlap_count = 0,
	for(i=0;i<num2;i++)
	{
		while(ind1 < num1 && genes1[ind1] < genes2[i])
			ind1++;
		if(genes1[ind1] == genes2[i])
		{
			overlapped_genes[*num_overlaped_genes] = genes1[ind1];
			(*num_overlaped_genes)++;
		}
	}
}

int *get_union_genes(int *genes1, int num1, int *genes2, int num2, int total_num_genes, int *num_union_genes)
{
	int i;
	int all_gene_indexes[total_num_genes];
	for(i=0;i<total_num_genes;i++)
		all_gene_indexes[i] = 0;
	for(i=0;i<num1;i++)
		all_gene_indexes[genes1[i]]++;
	for(i=0;i<num2;i++)
		all_gene_indexes[genes2[i]]++;
	// count genes
	*num_union_genes = 0;
	for(i=0;i<total_num_genes;i++)
	{
		if(all_gene_indexes[i] > 0)
		{
			(*num_union_genes)++;	
		}
	}
	// get actual genes
	int *union_genes = malloc(sizeof(int) * (*num_union_genes));
	int count = 0;
	for(i=0;i<total_num_genes;i++)
	{
		if(all_gene_indexes[i] > 0)
		{
			union_genes[count] = i;
			count++;
		}
	}
	return union_genes;
}

int *get_intersect_genes(int *genes1, int num1, int *genes2, int num2, int total_num_genes, int *num_intersect_genes)
{
	int i;
	int all_gene_indexes[total_num_genes];
	for(i=0;i<total_num_genes;i++)
		all_gene_indexes[i] = 0;
	for(i=0;i<num1;i++)
		all_gene_indexes[genes1[i]]++;
	for(i=0;i<num2;i++)
		all_gene_indexes[genes2[i]]++;
	// count intersection
	*num_intersect_genes = 0;
	for(i=0;i<total_num_genes;i++)
	{
		if(all_gene_indexes[i] == 2)
		{
			(*num_intersect_genes)++;
		}
	}
	// get intersecting genes
	int *intersect_genes = malloc(sizeof(int) * (*num_intersect_genes));
	int count = 0;
	for(i=0;i<total_num_genes;i++)
		{
			if(all_gene_indexes[i] == 2)
			{
				intersect_genes[count] = i;
				count ++;
			}
		}
	return intersect_genes;
}


d_gs *evolve_motif(d_gs *dg, params *p)
{
	evolve_motif_report_top_fptr evolve_fptr = evolve_motif_report_top_bidirectional;
	create_geneset_fptr create_d_gs_fptr;
	if(p->use_conservation == true)
	{
//		evolve_fptr = evolve_motif_report_top_bidirectional_orth_filter;
		create_d_gs_fptr = create_geneset_applying_orth_filter;
	}
	else
	{
//		evolve_fptr = evolve_motif_report_top_bidirectional;.
		create_d_gs_fptr = create_d_gs;
	}

	d_gs *top_dg = dg;
	double degen_score_orth, fraction_degen_pos_orth;
	while(true)
	{
		// create a new candidate motif
		d_gs *evolved = evolve_fptr(top_dg->degen_oligo, p, create_d_gs_fptr, p->select_next_greedy, top_dg->num_evolution_steps);
		if(evolved == 0)
			break;
		// check filters for ortholog case
		eval_degen_stats(evolved->degen_oligo, &degen_score_orth, &fraction_degen_pos_orth);
		if(
		    strlen(evolved->degen_oligo) <= p->max_motif_length 									&&
		    strlen(evolved->degen_oligo) >= p->min_motif_length 									&&
		    pass_degen_filter(degen_score_orth, fraction_degen_pos_orth, p->degen_threshold, p->fraction_degen_pos_threshold) == true 	&&
		    has_score_improved(top_dg->score, evolved->score) == true
		  )
		{
			// continue evolving 'evolved_orth' as conservation has been enabled and score improves
			free_d_gs(top_dg);
			top_dg = evolved;
		}
		else
		{
			// stop evolving because 'evolved' doesn't improve score
			free_d_gs(evolved);
			break;
		}
	}
	return top_dg;
}


//extern species species_type;
bool init_background_freq_first = true;
void init_background_freq(species species_type)
{
	int i;
	float freq_AG[2] = {0.25,0.25}; // frequencies for A and G nucleotides - must sum up to 0.5. the rest are dependent: freq(C)=freq(G) and freq(T)=freq(A)
	float freq_UP[2] = {0.6, 0.4}; // frequencies for being Paired or Unpaired - must sum up to 1
	if(init_background_freq_first == true)
	{
		background_freq_vec_AGCT = gsl_vector_alloc(no_primary_IUPAC_nucleotides);
		switch(species_type)
		{
			case yeast:
				freq_AG[0] = 0.11694; // A-T
				freq_AG[1] = 0.38306; // G-C
				break;
			case human://printf("init human\n");
				freq_AG[0] = 0.224627689666608; // A-T
				freq_AG[1] = 0.275372310333392; // G-C
				break;
			case arabidopsis:
				freq_AG[0] = 0.14026; // A-T
				freq_AG[1] = 0.35974; // G-C
				break;
			default: // uniform distribution (implicit initialization)
				break;
		}
		if (no_primary_IUPAC_nucleotides==4)
		{
			gsl_vector_set(background_freq_vec_AGCT, 0, freq_AG[0]); // A
			gsl_vector_set(background_freq_vec_AGCT, 1, freq_AG[1]); // G
			gsl_vector_set(background_freq_vec_AGCT, 2, freq_AG[1]); // C
			gsl_vector_set(background_freq_vec_AGCT, 3, freq_AG[0]); // T
		}
		else // no_primary_IUPAC_nucleotides==8
			// we consider derived from data frequencies for being paired or unpaired
		{
			gsl_vector_set(background_freq_vec_AGCT, 0, freq_AG[0]*freq_UP[0]); // A
			gsl_vector_set(background_freq_vec_AGCT, 1, freq_AG[1]*freq_UP[0]); // G
			gsl_vector_set(background_freq_vec_AGCT, 2, freq_AG[1]*freq_UP[0]); // C
			gsl_vector_set(background_freq_vec_AGCT, 3, freq_AG[0]*freq_UP[0]); // T
			gsl_vector_set(background_freq_vec_AGCT, 4, freq_AG[0]*freq_UP[1]); // a
			gsl_vector_set(background_freq_vec_AGCT, 5, freq_AG[1]*freq_UP[1]); // g
			gsl_vector_set(background_freq_vec_AGCT, 6, freq_AG[1]*freq_UP[1]); // c
			gsl_vector_set(background_freq_vec_AGCT, 7, freq_AG[0]*freq_UP[1]); // t
		}
		init_background_freq_first = false;
	}
}




//////////////////////////


void test_gsl_sort_index()
{
	int i,j,k;
	size_t *p; size_t stride = 1; int n = 5;
	MALLOC(p, sizeof(size_t)*n);
	double data[5] = {0,10,-1,4,7};
	gsl_sort_index(p, data, stride, n);
	for(i=0;i<n;i++)
		printf("%i:\t%i:\t%lf\t%lf\n", i, p[i], data[i], data[p[i]]);
	getchar();
}


void test_eval_degen_stats()
{
	double degen_score, fraction_degen_pos;
	char *oligo = "AGTNR";
	eval_degen_stats(oligo, &degen_score, &fraction_degen_pos);
	printf("degen score: %s\t: %-2.1lf\t: %-2.1lf\n", oligo, degen_score, fraction_degen_pos);
	oligo = "AGTTA";
	eval_degen_stats(oligo, &degen_score, &fraction_degen_pos);
	printf("degen score: %s\t: %-2.1lf\t: %-2.1lf\n", oligo, degen_score, fraction_degen_pos);
	oligo = "WGTNR";
	eval_degen_stats(oligo, &degen_score, &fraction_degen_pos);
	printf("degen score: %s\t: %-2.1lf\t: %-2.1lf\n", oligo, degen_score, fraction_degen_pos);
	oligo = "WRNNW";
	eval_degen_stats(oligo, &degen_score, &fraction_degen_pos);
	printf("degen score: %s\t: %-2.1lf\t: %-2.1lf\n", oligo, degen_score, fraction_degen_pos);
	getchar();
}


//////////////////
